#!/bin/bash
# lensing_slurm_20221021 - script for the lensing project, to be called by meta_lensing_slurm_20221021.sh
# (C) 2022 B. Roukema GPL-3+

# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
# 
# This project is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along
# with this project.  If not, see <https://www.gnu.org/licenses/>.

# Please read through all the parameters and comments: you *must*
# modify some of the parameters for your system.

LANG=C
set -x  # this is to print out the shell commands to stdout

## Initial stdout and stderr will be fed to
## ${HOME_PROJECT}/slurm-${SLURM_JOB_ID}.out
##
## COMMIT should be provided by the meta_ script that calls this one.

# source package name and commit ID
SRCPKG_COMMIT=lensing-${COMMIT}
# final pdf file produced by the maneage project
OUTPUTFILE=lensing.pdf
# name of the final pdf file that includes the commit ID and the slurm job ID
OUTPUTFILE_HOME=lensing-${COMMIT}-${SLURM_JOB_ID}.pdf

## Set up a directory TMP in which to create a temporary build
## directory - the directory ${TMP} may *not* be a subdirectory of
## your package directory.
TMP=${BUILD_PROJECT}/${SLURM_JOB_ID}
mkdir -pv ${TMP}

DATE=$(date +"%Y-%m-%d-%H-%M-%S")
## This will be the name of your main maneage log file.\n"
LOG=${TMP}/${DATE}.log
printf "The log file will be ${LOG} .\n"
touch ${LOG}
ls -l ${LOG}


printf "######################################################################" >> ${LOG}
printf "######################################################################" >> ${LOG}
# Start logging, and run several tests that can be used to debug in
# case something goes wrong. The outputs will be stored in the log
# file.

printf "Log of running ${SRCPKG_COMMIT} on ${DATE}.\n\n" 2>&1 >> ${LOG}
cd ${TMP}  2>&1 >> ${LOG}

uname -a  2>&1 >> ${LOG}
for i in "debian_" "redhat-"; do 
    printf "Try ${i}...\n" >> ${LOG}
    (ls -l ${i}* || true) >> ${LOG}
    cat /etc/${i}* >> ${LOG}
done

printf "Environment variable check:\n" 2>&1 >> ${LOG}
printf "  HOME=${HOME_PROJECT}\n" 2>&1 >> ${LOG}
printf "  PATH=${PATH}\n" 2>&1 >> ${LOG}
printf "  TMP=${TMP}\n" 2>&1 >> ${LOG}
printf "  CCFLAGS=${CCFLAGS}\n" 2>&1 >> ${LOG}
printf "  CFLAGS=${CFLAGS}\n" 2>&1 >> ${LOG}
printf "  LDFLAGS=${LDFLAGS}\n" 2>&1 >> ${LOG}
printf "  CPUS_PER_TASK=${CPUS_PER_TASK}\n" 2>&1 >> ${LOG}
printf "  COMMIT=${COMMIT}\n\n" 2>&1 >> ${LOG}

printf "NEED_MODULE=${NEED_MODULE}\n" >> ${LOG}
printf "HOST_OS_GCC_VERSION=${HOST_OS_GCC_VERSION}\n" >> ${LOG}

## Make sure that 'cc' is interpreted as 'gcc', and 'c++' as 'g++'.
alias cc='gcc'
alias c++='g++'

if [ "x${NEED_MODULE}" = "xYES" ]; then
    printf "Use of modules needs some hacks. Sorry.\n"
    exit 1
fi

printf "\n\n" >> ${LOG}
printf "PATH=${PATH}\n" 2>&1 >> ${LOG}

SRCARCHIVE_FILE=${SRCPKG_COMMIT}-snapshot.tar.gz
printf "SRCARCHIVE_FILE=${SRCARCHIVE_FILE}\n" 2>&1 >> ${LOG}

SRCARCHIVE_FULLPATH=${HOME_PROJECT}/${SRCARCHIVE_FILE}
printf "SRCARCHIVE_FULLPATH=${SRCARCHIVE_FULLPATH}\n" 2>&1 >> ${LOG}
ls -l ${SRCARCHIVE_FULLPATH}  2>&1 >> ${LOG}

printf "DATE=${DATE}\n" >> ${LOG}
printf "TMP=${TMP}; LOG=${LOG}\n" >> ${LOG}
printf "meta-level slurm log file=${SLURM_LOG_DIR}/slurm-${SLURM_JOB_ID}.out\n" >> ${LOG}

printf "Node parameters:\n" 2>&1 >> ${LOG}
uname -a  2>&1 >> ${LOG} || true
nproc     2>&1 >> ${LOG} || true
free -h     2>&1 >> ${LOG} || true
w     2>&1 >> ${LOG} || true
bash --version 2>&1 >> ${LOG} || true
whereis bash   2>&1 >> ${LOG} || true
which bash   2>&1 >> ${LOG} || true
printf "\n" 2>&1 >> ${LOG}

gcc --version  2>&1 >> ${LOG}
printf "Test a gcc compile of a 1-line program:\n" >> ${LOG}
cat <<EOF > dummy.c 
int main(void){ int i = 314; i+=5; return 0; }
EOF
(gcc -Wall --verbose dummy.c 2>&1) >> ${LOG}
printf "\n" 2>&1 >> ${LOG}
ls -l a.out 2>&1 >> ${LOG}
printf "\n" 2>&1 >> ${LOG}


g++ --version  2>&1 >> ${LOG}
printf "Test a g++ compile of a 1-line program:\n" >> ${LOG}
cat <<EOF > dummy.c 
int main(void){ int i = 314; i+=5; return 0; }
EOF
(g++ -Wall --verbose dummy.c 2>&1) >> ${LOG}
printf "\n" 2>&1 >> ${LOG}
ls -l a.out 2>&1 >> ${LOG}
printf "\n" 2>&1 >> ${LOG}

whereis grep   2>&1 >> ${LOG} || true
which grep   2>&1 >> ${LOG} || true
grep --version  2>&1 >> ${LOG} || true
printf "\n" 2>&1 >> ${LOG}
whereis awk   2>&1 >> ${LOG} || true
which awk   2>&1 >> ${LOG} || true
awk --version  2>&1 >> ${LOG}
printf "\n" 2>&1 >> ${LOG}
whereis sed   2>&1 >> ${LOG} || true
which sed   2>&1 >> ${LOG} || true
sed --version  2>&1 >> ${LOG}
printf "\n" 2>&1 >> ${LOG}

/bin/pwd 2>&1 >> ${LOG}
cp -pv ${SRCARCHIVE_FULLPATH} . 2>&1 >> ${LOG}

# Make the build directory
printf "We should now be inside the top directory ready to create the build directory.\n"
/bin/pwd 2>&1 >> ${LOG}
mkdir -pv build 2>&1 >> ${LOG}
cd build 2>&1 >> ${LOG}
/bin/pwd 2>&1 >> ${LOG}

BUILD=$(pwd)
printf "The build directory is BUILD=${BUILD}.\n" 2>&1 >> ${LOG}
cd .. 2>&1 >> ${LOG}
/bin/pwd 2>&1 >> ${LOG}

# Extract the files and directories from the source .tar.gz file.
printf "Ready to extract ${SRCARCHIVE_FILE}...\n" 2>&1 >> ${LOG}
tar -xv -f "${SRCARCHIVE_FILE}" 2>&1 >> ${LOG}
cd ${SRCPKG_COMMIT} 2>&1 >> ${LOG}

/bin/pwd 2>&1 >> ${LOG}

# Check the initial conditions file.
printf "\n" >> ${LOG}
cat reproduce/analysis/config/init-conditions-params.conf >> ${LOG}
printf "\n" >> ${LOG}



printf "######################################################################" >> ${LOG}
printf "######################################################################" >> ${LOG}
# Configure stage of 'maneage', including some checks.

# Check that ./project help info is available.
(cat /dev/null | ./project configure --help 2>&1) >> ${LOG}

# Do the main build of the project:
(cat /dev/null | ./project configure --build-dir=${BUILD} -j${CPUS_PER_TASK} 2>&1) >> ${LOG}

## Double and triple check the configure stage - some parallel
## installs may fail if the order of compiling and installing is not
## sufficiently constrained, since maneage is still very much in the
## early development stage. For example, as of 2020-07-26, openblas-0.3.10 is
## suspected to have a compile order bug that sometimes appears when
## trying to create libopenblas_haswellp-r0.3.10.so .
sleep 60 >> ${LOG}
printf "\nDouble-checking:::: ./project configure -e:\n" >> ${LOG}
(./project configure --existing-conf -j${CPUS_PER_TASK} 2>&1) >> ${LOG}

sleep 120 >> ${LOG}
printf "\nTriple-checking:::: ./project configure -e:\n" >> ${LOG}
(./project configure --existing-conf -j${CPUS_PER_TASK} 2>&1) >> ${LOG}


## Sometimes a few attempts might be needed to build 'texlive':
TEXLIVE_IS_READY_FILE=${BUILD}/software/installed/version-info/tex/texlive-ready-tlmgr

count_texlive=0
texlive_delay=300

TEXLIVE_IS_READY=$(cat ${TEXLIVE_IS_READY_FILE})
printf "TEXLIVE_IS_READY=${TEXLIVE_IS_READY}\n" 
if ! ( printf " ${TEXLIVE_IS_READY}" | grep "TeX Live is ready\." ) ; then
    rm -fv ${TEXLIVE_IS_READY_FILE} >> ${LOG}
fi

# Try up to 12 times.
while( ( ! [ -e ${TEXLIVE_IS_READY_FILE} ] ) && ( [ ${count_texlive} -le 12 ] ) ); do
    count_texlive=$((${count_texlive}+1))
    printf "Tex-live is not yet installed. Will try again after ${texlive_delay} seconds...\n"  >> ${LOG}
    sleep ${texlive_delay} # wait ${texlive_delay} seconds before trying again
    
    (./project configure --existing-conf -j${CPUS_PER_TASK} 2>&1) >> ${LOG}
        
    TEXLIVE_IS_READY=$(cat ${TEXLIVE_IS_READY_FILE}) >> ${LOG}
    printf "TEXLIVE_IS_READY=${TEXLIVE_IS_READY}\n" 
    if ! ( printf " ${TEXLIVE_IS_READY}" | grep "TeX Live is ready\." ) ; then
        rm -fv ${TEXLIVE_IS_READY_FILE} >> ${LOG}
    fi
done

## Do one more check, just to be sure.
(./project configure --existing-conf -j${CPUS_PER_TASK} 2>&1) >> ${LOG}


printf "######################################################################" >> ${LOG}
printf "######################################################################" >> ${LOG}
# Run the project itself:
(./project make 2>&1) >> ${LOG}

# Copy the final pdf file to ${HOME_PROJECT}/ .
cp -pv ${OUTPUTFILE} ${HOME_PROJECT}/${OUTPUTFILE_HOME} 2>&1 >> ${LOG}
