#!/bin/bash
# meta_lensing_slurm_20221021 - meta-level script for calling 'sbatch' for the lensing project
# (C) 2022 B. Roukema GPL-3+

# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
# 
# This project is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along
# with this project.  If not, see <https://www.gnu.org/licenses/>.

# Please read through all the parameters and comments: you *must*
# modify some of the parameters for your system.

# You probably have to modify:
# PARTITION
# and you very likely will have to modify these other parameters:
# HOME_PROJECT
# BUILD_PROJECT
# COMMIT
# See the notes below and edit the values. (Do *not* put spaces around
# the '=' sign; this is a 'shell' script, not a 'make' script.)


#Reduce the risk of the French or Polish decimal comma in shell commands
LANG=C


# For more help on --ntasks:
#
# info sbatch 
NTASKS=1  


# For more info on 'slurm partitions' you have access to, find documentation
# for users of your supercomputer. You can check the existing partitions
# (you might not have access to all of them) with:
# 
# sinfo --all -o "%.10P %.10a %.15l %.20F"
#
# You QUITE LIKELY have to modify this parameter.
#
#PARTITION=fast
PARTITION=standard
# standard: This might have parameters, such as, e.g. max running time
# 1 day, default RAM 2 Gb


# This is the number of cpus that you request per slurm 'task'.
# For 'lensing', something like 28 is a good choice; better not use less than 20.
# This parameter will be fed through to ${LENSING_SCRIPT}
export CPUS_PER_TASK=28


# This is a backed up area of disk where you have placed your source
# files, including:
# * the snapshot tar file
# * your main script file (called ${LENSING_SCRIPT} below)
#
# You *MUST* modify this parameter.
#
export HOME_PROJECT=/my/home/myproject
# double-check
if ! [ -d ${HOME_PROJECT} ]; then
    printf "Please fix the meta script, since the directory HOME_PROJECT=${HOME_PROJECT} does not seem to exist.\n"
    exit 1
fi
printf "Your final pdf file, if all goes well, should\n"
printf "be placed here in HOME_PROJECT=${HOME_PROJECT} .\n"

# This is the name of the script that calls the 'maneage' project,
# such as 'lensing'. The file is assumed to be in ${HOME_PROJECT}.
#
# You will need to modify this if you have changed the
# filename of the main script, e.g. to a newer version.
#
LENSING_SCRIPT=lensing_slurm_20221021.sh
# double-check
if ! [ -x ${HOME_PROJECT}/${LENSING_SCRIPT} ]; then
    printf "Please fix the script file LENSING_SCRIPT=${HOME_PROJECT}/${LENSING_SCRIPT}."
    printf "It must exist and be executable by the user (info chmod).\n"
    exit 1
fi

## Create a directory where output directly from slurm will be stored.
SLURM_LOG_DIR=${HOME_PROJECT}/slurm_log
mkdir -pv ${SLURM_LOG_DIR}
cd ${SLURM_LOG_DIR}
printf "You will find some output directly from slurm\n"
printf "in ${HOME_PROJECT}/slurm_log/ .\n"


## This is an existing directory in which to create a temporary build
## directory - the directory ${BUILD_PROJECT} may *not* be a
## subdirectory of your package directory.
#
# You *MUST* modify this parameter.
# 
export BUILD_PROJECT=/my/project/tmp_build_space
# double-check
if ! [ -d ${BUILD_PROJECT} ]; then
    printf "Please fix the meta script, since the directory BUILD_PROJECT=${BUILD_PROJECT} does not seem to exist.\n"
    exit 1
fi
printf "A temporary copy of your source files, and a log file,\n"
printf "and a build/ directory, should be placed in\n"
printf "BUILD_PROJECT=${BUILD_PROJECT} .\n"
printf "The *.log file, and the build/software/ and the build/analysis/\n"
printf "subdirectories there should provide extensive diagnostics and results.\n"


## This is the git commit ID of your source tar file. The name is assumed
## to be in the form such as   lensing-${COMMIT}.tar.gz  . This
## will be checked in the script ${LENSING_SCRIPT} once that script
## starts running,
#
# You *ALMOST CERTAINLY MUST* modify this parameter.
# 
COMMIT=abcd123

## As long as your host operating system (host OS) has a default, working
## version of gcc (and associated g++), you can ignore the parameters
## HOST_OS_GCC_VERSIONS

## leave this at 'no' - using modules requires some hacks
NEED_MODULE=no
## leave this at 'dummy' - using modules requires some hacks
HOST_OS_GCC_VERSIONS=dummy

## You will *probably* have to modify the values given to --mem and --time
## depending on the computer node available for your project.
##
## For 'lensing', you *might* be able to reduce from --mem=120000M
## (120 Gb) to a bit over 100 Gb. This is the amount that slurm will allow
## your project to use.
##
## For 'lensing', you *should* be able to reduce from --time=5760 minutes = 4
## days, but you might not have to; once your script has finished,
## your slurm task will be closed.

function lensing_submit {
    for HOST_OS_GCC_VERSION in ${HOST_OS_GCC_VERSIONS}; do
        for c in ${COMMIT}; do
            sbatch \
                --ntasks=${NTASKS} \
                --partition=${PARTITION} \
                --cpus-per-task=${CPUS_PER_TASK} \
                --export=COMMIT=${c},CPUS_PER_TASK=${CPUS_PER_TASK},NEED_MODULE=${NEED_MODULE},HOST_OS_GCC_VERSION=${HOST_OS_GCC_VERSION},HOME_PROJECT=${HOME_PROJECT},BUILD_PROJECT=${BUILD_PROJECT},SLURM_LOG_DIR=${SLURM_LOG_DIR} \
                --mem=120000M \
                --time=5760 \
                ${HOME_PROJECT}/${LENSING_SCRIPT}
            printf "${HOME_PROJECT}/${LENSING_SCRIPT} submitted...\n"
            sleep 15 # wait a bit (15 seconds) before submitting another commit
        done
    done
}

function submit_all {
        lensing_submit
}


submit_all

