# Verify the project outputs before building the paper.
#
# Copyright (C) 2020-2022 Mohammad Akhlaghi <mohammad@akhlaghi.org>
#
# This Makefile is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This Makefile is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this Makefile.  If not, see <http://www.gnu.org/licenses/>.


# Abbreviation for statistical verification script.
verify_parameter_statistically = $(bashdir)/verify-parameter-statistically.sh



# Verification functions
# ----------------------
#
# These functions are used by the final rule in this Makefile
verify-print-error-start = \
  echo; \
  echo "VERIFICATION ERROR"; \
  echo "------------------"; \
  echo

verify-print-tips = \
  echo "If you are still developing your project, you can disable"; \
  echo "verification by removing the value of the variable in the"; \
  echo "following file (from the top project source directory):"; \
  echo "    reproduce/analysis/config/verify-outputs.conf"; \
  echo; \
  echo "If this is the final version of the file, you can just copy"; \
  echo "and paste the calculated checksum (above) for the file in"; \
  echo "the following project source file:"; \
  echo "    reproduce/analysis/make/verify.mk"

# Removes following components of a plain-text file, calculates checksum
# and compares with given checksum:
#   - All commented lines (starting with '#') are removed.
#   - All empty lines are removed.
#   - All space-characters in remaining lines are removed (so the width of
#     the printed columns won't invalidate the verification).
#
# It takes three arguments:
#   - First argument: Full address of file to check.
#   - Second argument: Expected checksum of the file to check.
#   - File name to write result.
verify-txt-no-comments-no-space = \
  infile=$(strip $(1)); \
  inchecksum=$(strip $(2)); \
  innobdir=$$(echo $$infile | sed -e's|$(BDIR)/||g'); \
  if ! [ -f "$$infile" ]; then \
    $(call verify-print-error-start); \
    echo "The following file (that should be verified) doesn't exist:"; \
    echo "    $$infile"; \
    echo; exit 1; \
  fi; \
  checksum=$$(sed -e 's/[[:space:]][[:space:]]*//g' \
                  -e 's/\#.*$$//' \
                  -e '/^$$/d' $$infile \
                  | md5sum \
                  | awk '{print $$1}'); \
  if [ x"$$inchecksum" = x"$$checksum" ]; then \
    echo "%% (VERIFIED) $$checksum $$innobdir" >> $(3); \
  else \
    $(call verify-print-error-start); \
    $(call verify-print-tips); \
    echo; \
    echo "Checked file (without empty or commented lines):"; \
    echo "    $$infile"; \
    echo "Expected MD5 checksum:   $$inchecksum"; \
    echo "Calculated MD5 checksum: $$checksum"; \
    echo; exit 1; \
  fi;


count-verify-files = $$(ls $(verifydir)/commit_$(strip $(1))_*verify.tex | wc -l)


# Final verification TeX macro (can be empty)
# -------------------------------------------
#
# This is the FINAL analysis step (before going onto the paper. Please use
# this step to verify the contents of the figures/tables used in the paper
# and the LaTeX macros generated from all your processing. It should depend
# on all the LaTeX macro files that are generated (their contents will be
# checked), and any files that go into the tables/figures of the paper
# (generated in various stages of the analysis.
#
# Since each analysis step's data files are already prerequisites of their
# respective TeX macro file, its enough for 'verify.tex' to depend on the
# final TeX macro.
#
# USEFUL TIP: during the early phases of your research (when you are
# developing your analysis, and the values aren't final), you can comment
# the respective lines.
#
# Here is a description of the variables defined here.
#
#   verify-dep: The major step dependencies of 'verify.tex', this includes
#               all the steps that must be finished before it.
#
#   verify-changes: The files whose contents are important. This is
#               essentially the same as 'verify-dep', but it has removed
#               the 'initialize' step (which is information about the
#               pipeline, not the results).
#verify-dep = $(filter-out verify paper, $(makesrc))
verify-dep = $(subst verify,,$(subst paper,,$(makesrc)))

# Remove some .mk files from the list of those to be verified:
verify-check0 = $(subst initialize,,$(verify-dep))
verify-check1 = $(subst download,,$(verify-check0))
## create-trees does not have its own .conf file
verify-check2= $(subst create-trees,,$(verify-check1))
## 2023-02-05 calc_densities does not have its own .conf file
verify-check3= $(subst calc_densities,,$(verify-check2))
## 2023-02-05 analyse-plot-lensing is checked statistically below
verify-check = $(subst analyse-plot-lensing,,$(verify-check3))


#$(mtexdir)/verify.tex: $(foreach s, $(verify-dep), $(mtexdir)/$(s).tex)
#	printf "%% Automatically produced file.\n" > $@
#	printf "\\\\newcommand{\\\\HowWeVerifyResults}{This is how we verified the results of the pipeline.}\n\n" >> $@


$(mtexdir)/verify.tex: $(foreach s, $(verify-dep), $(mtexdir)/$(s).tex)
#	Make sure that verification is actually requested.
	printf "verify-check=$(verify-check)\n"
	LC_NUMERIC=C
	LANG=C
	if [ x"$(verify-outputs)" = xyes ]; then

#	Make sure the temporary output doesn't exist (because we want to
#	append to it). We are making a temporary output target so if there
#	is a crash in the middle, Make will not continue. If we write in
#	the final target progressively, the file will exist, and its date
#	will be more recent than all prerequisites, so next time the
#	project is run, Make will continue and ignore the rest of the
#	checks.
	  rm -f $@.tmp

#	Verify TeX macros (the values that go into the PDF text).
	  for m in $(verify-check); do
	    file=$(mtexdir)/$$m.tex
	    if [ $$m == init-N-body ]; then
	       s=a49834c0fc0d5586e9d0fa7f17613caf
	       egrep -v "NCPUSvalue" $${file} > $${file}.verify
	    elif [ $$m == run-simulations ]; then
	       s=929bc98ec2085a33ede482d62e610fd5
	       cat $${file} > $${file}.verify
	    elif [ $$m == detect-haloes ]; then
	       s=fb31641e94cc810843382819c3ac3373
	       cat $${file} > $${file}.verify
	    elif [ $$m == gal-recipes ]; then
	       s=2102f43214c3a29aae057012ec586cb9
	       cat $${file} > $${file}.verify
	    elif [ $$m == detect-voids ]; then
	       s=9e055d33e9c2f41806edafd14173c1c6
	       cat $${file} > $${file}.verify
	    elif [ $$m == lensing ]; then
	       s=83e1096244246a5b16dd4372e277b7ef
	       egrep -v "Match|NVoid" $${file} > $${file}.verify
	    else echo; echo "'$$m' not recognized."; exit 1
	    fi
	    if [ x"$(devmode)" != x1 ]; then
#	      default: check this file if not in devmode
	      $(call verify-txt-no-comments-no-space, $${file}.verify, $$s, $@.tmp)
	    elif [ $$m != lensing ]; then
#	      devmode: check this file except if it is lensing.tex
	      $(call verify-txt-no-comments-no-space, $${file}.verify, $$s, $@.tmp)
	    fi
	  done

	else
	  echo "% Verification was DISABLED!" > $@.tmp
	fi

	if [ x"$(verify-outputs)" = xyes ]; then
#	Verify the best values of the tables (the values that go into the
#	PDF text).  We leave verification of the uncertainties to an open
#	question not handled here.

#	Lensing project: these significance parameters are numerically
#	sensitive for independent calculations on the same machine; and
#	also sensitive to the choice of machine; generous allowances are
#	made for checking result consistency.

	  file=$(mtexdir)/analyse-plot-lensing.tex
	  n_verify_sigma=6

#####	  P_{x,z}(3D|2D)
	  param=SigmaSignificanceXZthreeDgiventwoD
	  x_mean=0.0350
	    x_sd=0.0007
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  #verify values: 1 true, 0 false
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" > $${file}.verify

	  param=gammaSignificanceXZthreeDgiventwoD
	  x_mean=0.0012
	    x_sd=0.0005
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify

	  param=thetaSignificanceXZthreeDgiventwoD
	  x_mean=0.0030
	    x_sd=0.0015
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify

	  param=sigmaSignificanceXZthreeDgiventwoD
	  x_mean=0.018
	    x_sd=0.008
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify


#####	  P_{x,z}(2D|3D)
	  param=SigmaSignificanceXZtwoDgiventhreeD
	  x_mean=0.00034
	    x_sd=0.00030
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  #verify values: 1 true, 0 false
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify

	  param=gammaSignificanceXZtwoDgiventhreeD
	  x_mean=0.00002
	    x_sd=0.00002
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify

	  param=thetaSignificanceXZtwoDgiventhreeD
	  x_mean=0.00002
	    x_sd=0.00002
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify

	  param=sigmaSignificanceXZtwoDgiventhreeD
	  x_mean=0.00025
	    x_sd=0.00009
	  val=$$(grep $${param} $${file} | sed -e 's/\\times 10^{\([^}]*\)}/e\1/' | \
	     awk -F '{' '{print $$2}' | \
	     sed -e 's/}//')
	  verify=$$(echo $${val} $${x_mean} $${x_sd}|awk -v n=$${n_verify_sigma} '{verify=($$1 <= $$2+n*$$3)?1:0; print verify; if(!verify)print " "$$1}')
	  printf "$${param} $${x_mean} $${x_sd} $${verify}\n" >> $${file}.verify

	  s=6ba47c8fd5d6f608a83b08d6655267b0
	  if [ x"$(devmode)" != x1 ]; then
	    $(call verify-txt-no-comments-no-space, $${file}.verify, $$s, $@.tmp)
	  fi
	fi

#	  # Galaxy size table - full mass range
#	  $(verify_parameter_statistically) \
#	          PlotsSizeiRAvgVoidsTotalvalue \
#	          PlotsSizeiErrorRAvgVoidsTotalvalue \
#	          $(n_verify_sigmas) \
#	          $(verifydir)/analyse-plot.tex.verify \
#	          $(mtexdir)/analyse-plot.tex $(cosmic_var_commit_ID) 2 $@.tmp $(verify-outputs)
#
#	  # Void parameter relation: haloinvoidfrac vs r/R_{eff}
#	  # - there is modest cosmic variance for this value:
#	  $(verify_parameter_statistically) \
#	          FracRreffSlopevalue \
#	          FracRreffSigSlopevalue \
#	          $(n_verify_sigmas) \
#	          $(verifydir)/analyse-plot.tex.verify \
#	          $(mtexdir)/analyse-plot.tex $(cosmic_var_commit_ID) 3 $@.tmp $(verify-outputs)
#
#	## Add LaTeX macros to quantify the verification criteria:
#	  printf "\\\\newcommand{\\\\CosmicVarCommitID}{{\\\\tt $(cosmic_var_commit_ID)}}\n" >> $@.tmp
#	  printf "\\\\newcommand{\\\\NVerifySigmas}{$(n_verify_sigmas)}\n" >> $@.tmp
#	  # TODO: Handle the startup case when $(cosmic_var_commit_ID) is empty, for generating new standard cases.
#	  count_verify_files=$(call count-verify-files, $(cosmic_var_commit_ID))
#	  printf "\\\\newcommand{\\\\NVerifyFiles}{$${count_verify_files}}\n" >> $@.tmp
#
#	else
#	  echo "% Verification was DISABLED; the LaTeX paper will fail due to missing values!" > $@.tmp
#	fi
#
#
#        # Move temporary file to final target.
	mv $@.tmp $@
