# analyse_void_matches - statistical analysis of matches between 2D and 3D voids
# (C) 2023 Boud Roukema, Marius Peper GPL-3+
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

import numpy as np
# Warning: 'import random' is for the python-level (not numpy) random generator
from numpy.random import SeedSequence, default_rng, PCG64, Generator
import scipy
import scipy.stats
from scipy.stats import permutation_test
import os

from lensing_calculations_unit_testable import lensing_calc_unit_testable as lens_calc_ut

## TODO (DRY: Don't Repeat Yourself): these should be read by
## lensing.py and this routine from a single place:
##
std_xz_void_prob = 5.0 # Mpc/h linear scale
std_R_void_prob = 0.1  # log10 of ratio of void radii (unit = "dex")



# detection method: density, weak lensing or Sachs optical scalar
# mass_def: projected mass deficit (underdensity) \Sigma
# gamma: weak lensing shear \gamma
# exp: Sachs optical expansion scalar \theta
# sig: Sachs optical shear scalar \sigma
detection_methods = ["mass_def","gamma","exp","sig"]
detection_method_symbols = ["\\Sigma","\\bargamma","\\theta","\\vert\\sigma\\vert"]

#template filename: void_matches_exp_given_2D.dat

# remove_duplicates - If two rows have identical values in columns k1,k2,k3
# to within a given tolerance 'tol', then select the row with the higher
# value of 'prob' and mark the other row for removal. Continue through all
# non-redundant and still valid pairs. After the double loop has
# terminated, remove the rows marked for removal, and return the resulting matrix.
#
# 2023-03-09 This is currently a *brute force* algorithm; only use this for small
# arrays. Update to a more efficient algorithm if the numbers are greater.
#
# xx = input array (given_2D or given_3D)
# k1, k2, k3 = columns of the parameters of the voids that might be duplicates
# k_prob = column of pseudo-probability of the quality of the match
def remove_duplicates(xx, k1,k2,k3, k_prob):
    # Hardwired tolerance; assumes that radii and positions are of order unity
    # give or take an order of magnitude or so.
    tol = 1e-10

    N_rows = xx.shape[0]
    rows_OK = np.full(N_rows,True) # mark all rows as OK
    for i in range(0,N_rows-1):
        for j in range(i+1,N_rows):
            # don't compare the row with itself
            if(i != j and rows_OK[i] and rows_OK[j]):
                # check if the three columns of data match
                if(np.abs(xx[i,k1] - xx[j,k1]) < tol and
                   np.abs(xx[i,k2] - xx[j,k2]) < tol and
                   np.abs(xx[i,k3] - xx[j,k3]) < tol):
                    # keep the better row
                    if(xx[i,k_prob] >= xx[j,k_prob]):
                        rows_OK[j] = False # reject row j
                    else:
                        rows_OK[i] = False # reject row i

    yy = xx[ np.where(rows_OK)[0] ]
    return yy

# INPUTS:
# detection_method: string
# N_sim: number of simulations to run, int

# OUTPUTS:
#
# p_xz_median_low_given_2D: probability that the median xz distance is
# as low as measured, given the 2D voids
#
# p_xz_median_low_given_3D: probability that the median xz distance is
# as low as measured, given the 3D voids
#
# p_R_spearmanr_high_given_2D: two-sided probability that the Spearman rank
# r is as high as measured, given the 2D voids
#
# p_R_spearmanr_high_given_3D: two-sided probability that the Spearman rank
# r is as high as measured, given the 3D voids
#
# p_R_spearmanr_high_given_2D_besthalf: as for p_R_spearmanr_high_given_2D, but
# only for the 2D and 3D voids matching better than the median match probability
#
# p_R_spearmanr_high_given_3D_besthalf: as for p_R_spearmanr_high_given_3D, but
# only for the 2D and 3D voids matching better than the median match probability
def analyse_matches(detection_method, N_pixels, N_sim_xz, N_sim_spearman, rm_duplicates = False):

    # read given_2D data file
    in_filename = lensing_dir + "/" + "void_matches_" + detection_method + "_given_2D.dat"
    given_2D = np.loadtxt(in_filename) # the defaults should work

    # If multiple 3D voids correspond to a 2D void, then keep only the most likely match:
    if(rm_duplicates):
        given_2D = remove_duplicates(given_2D, 0,1,2, 6)

    N_2D = given_2D.shape[0]

    # read given_3D data file
    in_filename = lensing_dir + "/" + "void_matches_" + detection_method + "_given_3D.dat"
    given_3D = np.loadtxt(in_filename) # the defaults should work

    # If multiple 3D voids correspond to a 2D void, then keep only the most likely match:
    if(rm_duplicates):
        given_3D = remove_duplicates(given_3D, 3,4,5, 6)

    N_3D = given_3D.shape[0]

    print("N_2D, N_3D = %d %d" % (N_2D,N_3D))

    #### given_2D case ####
    # Calculate x and z T^2 distances and the median xz_median
    # The suffix "_found" means the best found match.
    R_3D_found = given_2D[:,0]
    x_3D_found = given_2D[:,1]
    z_3D_found = given_2D[:,2]
    R_2D = given_2D[:,3]
    x_2D = given_2D[:,4]
    z_2D = given_2D[:,5]
    match_prob_2D = given_2D[:,6] # heuristic probability that the match is valid
    d_x = np.amin([np.abs(x_3D_found-x_2D),
                   np.abs(N_pixels + x_3D_found-x_2D),
                   np.abs(N_pixels + x_2D-x_3D_found)], axis=0)
    d_z = np.amin([np.abs(z_3D_found-z_2D),
                   np.abs(N_pixels + z_3D_found-z_2D),
                   np.abs(N_pixels + z_2D-z_3D_found)], axis=0)

    d_xz = np.sqrt(d_x*d_x + d_z*d_z)
    #print("d_xz = ",d_xz)
    x_median = np.median(d_x)
    z_median = np.median(d_z)
    xz_median = np.median(d_xz)
    print("x_median = ",x_median)
    print("z_median = ",z_median)
    if(rm_duplicates):
        print("xz_median_given_twoD_uniq_%s = %g" % (detection_method, xz_median))
    else:
        print("xz_median_given_twoD_%s = %g" % (detection_method, xz_median))

    R_3D = given_3D[:,0]
    x_3D = given_3D[:,1]
    z_3D = given_3D[:,2]
    R_2D_found = given_3D[:,3]
    x_2D_found = given_3D[:,4]
    z_2D_found = given_3D[:,5]
    match_prob_3D = given_3D[:,6] # heuristic probability that the match is valid

    x_median_sim = -1.0*np.ones(N_sim_xz)
    z_median_sim = -1.0*np.ones(N_sim_xz)
    xz_median_sim = -1.0*np.ones(N_sim_xz)

    ############## given_2D ################
    # Simulate 2D x and z, forgetting the info on the best matched 3D voids.
    for i_sim in range(0,N_sim_xz):
        # Generate N_2D intrinsic voids uniformly
        x_2D_sim = scipy.stats.uniform.rvs(-N_pixels/2.0,N_pixels,N_2D, random_state=rng_void_matches)
        z_2D_sim = scipy.stats.uniform.rvs(-N_pixels/2.0,N_pixels,N_2D, random_state=rng_void_matches)

        x_3D_best = -1.0*np.ones(N_2D)
        z_3D_best = -1.0*np.ones(N_2D)

        for i_2D in range(0,N_2D):
            p, i_best = lens_calc_ut.void_probability(R_3D, x_3D, z_3D,
                                         std_R_void_prob, std_xz_void_prob,
                                         R_2D[i_2D], x_2D_sim[i_2D], z_2D_sim[i_2D], N_pixels)
            x_3D_best[i_2D] = x_3D[i_best]
            z_3D_best[i_2D] = z_3D[i_best]

        # Calculate x and z T^2 distances and their medians
        d_x = np.amin([np.abs(x_3D_best-x_2D_sim),
                   np.abs(N_pixels + x_3D_best-x_2D_sim),
                   np.abs(N_pixels + x_2D_sim-x_3D_best)], axis=0)
        d_z = np.amin([np.abs(z_3D_best-z_2D_sim),
                   np.abs(N_pixels + z_3D_best-z_2D_sim),
                   np.abs(N_pixels + z_2D_sim-z_3D_best)], axis=0)

        d_xz = np.sqrt(d_x*d_x + d_z*d_z)
        #print("d_xz = ",d_xz)
        x_median_sim[i_sim] = np.median(d_x)
        z_median_sim[i_sim] = np.median(d_z)
        xz_median_sim[i_sim] = np.median(d_xz)


    # Calculate the probabilities
    #print("xz_median_sim = ",xz_median_sim)
    #print(np.where(xz_median_sim < xz_median)[0])

    p_x_median_low = len(np.where(x_median_sim < x_median)[0])/N_sim_xz
    p_z_median_low = len(np.where(z_median_sim < z_median)[0])/N_sim_xz

    print("p_x, p_z: ",p_x_median_low, p_z_median_low)
    p_xz_median_low_given_2D = (np.maximum(1,len(np.where(xz_median_sim < xz_median)[0])))/N_sim_xz


    ############## given_3D ################
    d_x = np.amin([np.abs(x_3D-x_2D_found),
                   np.abs(N_pixels + x_3D-x_2D_found),
                   np.abs(N_pixels + x_2D_found-x_3D)], axis=0)
    d_z = np.amin([np.abs(z_3D-z_2D_found),
                   np.abs(N_pixels + z_3D-z_2D_found),
                   np.abs(N_pixels + z_2D_found-z_3D)], axis=0)

    d_xz = np.sqrt(d_x*d_x + d_z*d_z)
    #print("d_xz = ",d_xz)
    x_median = np.median(d_x)
    z_median = np.median(d_z)
    xz_median = np.median(d_xz)
    print("x_median = ",x_median)
    print("z_median = ",z_median)

    if(rm_duplicates):
        print("xz_median_given_threeD_uniq_%s = %g" % (detection_method, xz_median))
    else:
        print("xz_median_given_threeD_%s = %g" % (detection_method, xz_median))

    def stat_spearman_2D_3D_found(R_2D):
        # R_3D_found has to be read from the wider scope:
        return scipy.stats.spearmanr(R_2D, R_3D_found).statistic
    perm_object = permutation_test(
        (R_2D,), stat_spearman_2D_3D_found,
        n_resamples=N_sim_spearman,
        permutation_type='pairings',
        random_state=rng_void_matches)

    p_R_spearmanr_high_given_2D = perm_object.pvalue

    sp_asymp, p_R_spearmanr_high_given_2D_asymp = scipy.stats.spearmanr(R_2D, R_3D_found)
    print("spearman: perm: %g asymp: %g\n" % (p_R_spearmanr_high_given_2D, p_R_spearmanr_high_given_2D_asymp))


    ## best half of matches
    p_median = np.median(match_prob_2D)
    i_besthalf = np.where(match_prob_2D > p_median)[0]
    len_i_besthalf = len(i_besthalf)

    def stat_spearman_2D_3D_found_besthalf(R_2D):
        # R_3D_found has to be read from the wider scope:
        return scipy.stats.spearmanr(R_2D, R_3D_found[i_besthalf]).statistic
    if(len_i_besthalf > 0):
        perm_object_besthalf = permutation_test(
            (R_2D[i_besthalf],), stat_spearman_2D_3D_found_besthalf,
            n_resamples=N_sim_spearman,
            permutation_type='pairings',
            random_state=rng_void_matches)

        p_R_spearmanr_high_given_2D_besthalf = perm_object_besthalf.pvalue
    else:
        p_R_spearmanr_high_given_2D_besthalf = 1.0

    print("spearman_besthalf: perm: %g \n" % (p_R_spearmanr_high_given_2D_besthalf))


    ############## given_3D ################
    # Simulate x and z, forgetting the info on the best matched 2D voids.
    for i_sim in range(0,N_sim_xz):
        # Generate N_2D intrinsic voids uniformly
        x_2D_sim = scipy.stats.uniform.rvs(-N_pixels/2.0,N_pixels,N_2D, random_state=rng_void_matches)
        z_2D_sim = scipy.stats.uniform.rvs(-N_pixels/2.0,N_pixels,N_2D, random_state=rng_void_matches)

        x_2D_best = -1.0*np.ones(N_3D)
        z_2D_best = -1.0*np.ones(N_3D)

        for i_3D in range(0,N_3D):
            p, i_best = lens_calc_ut.void_probability(R_3D[i_3D], x_3D[i_3D], z_3D[i_3D],
                                         std_R_void_prob, std_xz_void_prob,
                                         R_2D, x_2D_sim, z_2D_sim, N_pixels)
            x_2D_best[i_3D] = x_2D_sim[i_best]
            z_2D_best[i_3D] = z_2D_sim[i_best]

        # Calculate x and z T^2 distances and their medians
        d_x = np.amin([np.abs(x_3D-x_2D_best),
                   np.abs(N_pixels + x_3D-x_2D_best),
                   np.abs(N_pixels + x_2D_best-x_3D)], axis=0)
        d_z = np.amin([np.abs(z_3D-z_2D_best),
                   np.abs(N_pixels + z_3D-z_2D_best),
                   np.abs(N_pixels + z_2D_best-z_3D)], axis=0)

        d_xz = np.sqrt(d_x*d_x + d_z*d_z)
        #print("d_xz = ",d_xz)
        x_median_sim[i_sim] = np.median(d_x)
        z_median_sim[i_sim] = np.median(d_z)
        xz_median_sim[i_sim] = np.median(d_xz)


    # Calculate the probabilities
    #print("xz_median_sim = ",xz_median_sim)
    #print(np.where(xz_median_sim < xz_median)[0])

    p_x_median_low = len(np.where(x_median_sim < x_median)[0])/N_sim_xz
    p_z_median_low = len(np.where(z_median_sim < z_median)[0])/N_sim_xz

    print("p_x, p_z: ",p_x_median_low, p_z_median_low)
    p_xz_median_low_given_3D = (np.maximum(1,len(np.where(xz_median_sim < xz_median)[0])))/N_sim_xz

    def stat_spearman_3D_2D_found(R_3D):
        # R_3D_found has to be read from the wider scope:
        return scipy.stats.spearmanr(R_3D, R_2D_found).statistic
    perm_object = permutation_test(
        (R_3D,), stat_spearman_3D_2D_found,
        n_resamples=N_sim_spearman,
        permutation_type='pairings',
        random_state=rng_void_matches)

    p_R_spearmanr_high_given_3D = perm_object.pvalue

    sp_asymp, p_R_spearmanr_high_given_3D_asymp = scipy.stats.spearmanr(R_3D, R_2D_found)
    print("spearman: perm: %g asymp: %g\n" % (p_R_spearmanr_high_given_3D, p_R_spearmanr_high_given_3D_asymp))


    ## best half of matches
    p_median = np.median(match_prob_3D)
    i_besthalf = np.where(match_prob_3D > p_median)[0]
    len_i_besthalf = len(i_besthalf)

    def stat_spearman_3D_2D_found_besthalf(R_3D):
        # R_2D_found has to be read from the wider scope:
        return scipy.stats.spearmanr(R_3D, R_2D_found[i_besthalf]).statistic
    if(len_i_besthalf > 0):
        perm_object_besthalf = permutation_test(
            (R_3D[i_besthalf],), stat_spearman_3D_2D_found_besthalf,
            n_resamples=N_sim_spearman,
            permutation_type='pairings',
            random_state=rng_void_matches)

        p_R_spearmanr_high_given_3D_besthalf = perm_object_besthalf.pvalue
    else:
        p_R_spearmanr_high_given_3D_besthalf = 1.0

    print("spearman_besthalf: perm: %g \n" % (p_R_spearmanr_high_given_3D_besthalf))


    return p_xz_median_low_given_2D, p_xz_median_low_given_3D, p_R_spearmanr_high_given_2D, p_R_spearmanr_high_given_3D, p_R_spearmanr_high_given_2D_besthalf, p_R_spearmanr_high_given_3D_besthalf


if __name__ == '__main__':

    if os.environ.get('N_PIXELS'):
        N_pixels = int(os.environ.get('N_PIXELS'))
    else:
        N_pixels = 120

    if os.environ.get('LENSING_DIR'):
        lensing_dir = os.environ.get('LENSING_DIR')
    else:
        lensing_dir = ".build/analysis/lensing-calculations/"

    if os.environ.get('VOID_MATCHING_DIR'):
        void_matching_dir = os.environ.get('VOID_MATCHING_DIR')
    else:
        void_matching_dir = ".build/analysis/void_matching/"

    if os.environ.get('N_VOID_MATCHES_SIM'):
        N_sim = int(os.environ.get('N_VOID_MATCHES_SIM'))
    else:
        N_sim = 50000

    if os.environ.get('LENSING_SEED_VOID_MATCHES'):
        lensing_seed_void_matches = int(os.environ.get('LENSING_SEED_VOID_MATCHES'))
    else:
        lensing_seed_void_matches = 199501
    if(lensing_seed_void_matches >= 0):
        rng_void_matches = Generator(PCG64(SeedSequence(lensing_seed_void_matches)))
    else:
        rng_void_matches = np.random.default_rng()


    N_sim_xz = N_sim
    N_sim_spearman = N_sim
    print("prob_random: #Col 1: physical quantity used for detection of 2D voids\n")
    print("prob_random: #Col 2: p_xz_median_low_given_2D - probability that the median xz distance is closer than random, given a 2D void\n")
    print("prob_random: #Col 3: p_xz_median_low_given_3D - probability that the median xz distance is closer than random, given a 3D void\n")
    print("prob_random: #Col 4: p_R_spearmanr_high_given_2D - Spearman rank probability that the radii correlate more strongly than random, given a 2D void\n")
    print("prob_random: #Col 5: p_R_spearmanr_high_given_3D - Spearman rank probability that the radii correlate more strongly than random, given a 3D void\n")

    if(False):
        # Disabled: allow multiple matches:
        for i_method in range(0,4):
            print("detection method = %s" % (detection_methods[i_method]))
            p_xz_median_low_given_2D, p_xz_median_low_given_3D, p_R_spearmanr_high_given_2D, p_R_spearmanr_high_given_3D, p_R_spearmanr_high_given_2D_besthalf, p_R_spearmanr_high_given_3D_besthalf = analyse_matches(detection_methods[i_method], N_pixels, N_sim_xz, N_sim_spearman)

            print("p_xz_median_low_given_2D, p_xz_median_low_given_3D, p_R_spearmanr_high_given_2D, p_R_spearmanr_high_given_3D:")

            print("prob_random: %10s %#9.2g %#9.2g %#9.2g %#9.2g\n" %
                  (detection_method_symbols[i_method],
                   p_xz_median_low_given_2D, p_xz_median_low_given_3D, p_R_spearmanr_high_given_2D, p_R_spearmanr_high_given_3D))
            print("\n")

            print("p_R_spearmanr_high_given_2D_besthalf, p_R_spearmanr_high_given_3D_besthalf:")
            print("prob_random_besthalf: %10s %#9.2g %#9.2g\n" %
                  (detection_method_symbols[i_method],
                   p_R_spearmanr_high_given_2D_besthalf, p_R_spearmanr_high_given_3D_besthalf))
            print("\n")

    # Unique matches
    for i_method in range(0,4):
        print("detection method = %s" % (detection_methods[i_method]))
        p_xz_median_low_given_2D_uniq, p_xz_median_low_given_3D_uniq, p_R_spearmanr_high_given_2D_uniq, p_R_spearmanr_high_given_3D_uniq, p_R_spearmanr_high_given_2D_uniq_besthalf, p_R_spearmanr_high_given_3D_uniq_besthalf = analyse_matches(detection_methods[i_method], N_pixels, N_sim_xz, N_sim_spearman, rm_duplicates = True)

        print("p_xz_median_low_given_2D_uniq, p_xz_median_low_given_3D_uniq, p_R_spearmanr_high_given_2D_uniq, p_R_spearmanr_high_given_3D_uniq:")

        print("prob_random_uniq: %10s %#9.2g %#9.2g %#9.2g %#9.2g\n" %
              (detection_method_symbols[i_method],
               p_xz_median_low_given_2D_uniq, p_xz_median_low_given_3D_uniq, p_R_spearmanr_high_given_2D_uniq, p_R_spearmanr_high_given_3D_uniq))
        print("\n")
