#!/usr/bin/env python

import numpy as np
from os.path import getsize as getFileSize

from scipy.interpolate import LinearNDInterpolator
#from scipy.interpolate import Rbf
import scipy.integrate as integrate

import matplotlib.pyplot as plt

def read_gals(model_name, first_file, last_file):
    """
    Read in the data about all galaxies generated by sage.
    """
    
    # The input galaxy structure:
    Galdesc_full = [
        ('SnapNum'                      , np.int32),
        ('Type'                         , np.int32),
        ('GalaxyIndex'                  , np.int64),
        ('CentralGalaxyIndex'           , np.int64),
        ('SAGEHaloIndex'                , np.int32),
        ('SAGETreeIndex'                , np.int32),
        ('SimulationHaloIndex'          , np.int64),
        ('mergeType'                    , np.int32),
        ('mergeIntoID'                  , np.int32),
        ('mergeIntoSnapNum'             , np.int32),
        ('dT'                           , np.float32),
        ('Pos'                          , (np.float32, 3)),
        ('Vel'                          , (np.float32, 3)),
        ('Spin'                         , (np.float32, 3)),
        ('Len'                          , np.int32),
        ('Mvir'                         , np.float32),
        ('CentralMvir'                  , np.float32),
        ('Rvir'                         , np.float32),
        ('Vvir'                         , np.float32),
        ('Vmax'                         , np.float32),
        ('VelDisp'                      , np.float32),
        ('ColdGas'                      , np.float32),
        ('StellarMass'                  , np.float32),
        ('BulgeMass'                    , np.float32),
        ('HotGas'                       , np.float32),
        ('EjectedMass'                  , np.float32),
        ('BlackHoleMass'                , np.float32),
        ('IntraClusterStars'            , np.float32),
        ('MetalsColdGas'                , np.float32),
        ('MetalsStellarMass'            , np.float32),
        ('MetalsBulgeMass'              , np.float32),
        ('MetalsHotGas'                 , np.float32),
        ('MetalsEjectedMass'            , np.float32),
        ('MetalsIntraClusterStars'      , np.float32),
        ('SfrDisk'                      , np.float32),
        ('SfrBulge'                     , np.float32),
        ('SfrDiskZ'                     , np.float32),
        ('SfrBulgeZ'                    , np.float32),
        ('DiskRadius'                   , np.float32),
        ('Cooling'                      , np.float32),
        ('Heating'                      , np.float32),
        ('QuasarModeBHaccretionMass'    , np.float32),
        ('TimeOfLastMajorMerger'         , np.float32),
        ('TimeOfLastMinorMerger'         , np.float32),
        ('OutflowRate'                  , np.float32),
        ('infallMvir'                   , np.float32),
        ('infallVvir'                   , np.float32),
        ('infallVmax'                   , np.float32)
    ]
    names = [Galdesc_full[i][0] for i in range(len(Galdesc_full))]
    formats = [Galdesc_full[i][1] for i in range(len(Galdesc_full))]
    Galdesc = np.dtype({'names':names, 'formats':formats}, align=True)


    # Initialize variables.
    TotNTrees = 0
    TotNGals = 0
    FileIndexRanges = []

    print("Determining array storage requirements.")

    # Read each file and determine the total number of galaxies to be read in
    goodfiles = 0
    for fnr in range(first_file,last_file+1):
        fname = model_name+'_'+str(fnr)  # Complete filename

        if not os.path.isfile(fname):
            print("File\t%s  \tdoes not exist!  Skipping..." % (fname))
            continue

        if getFileSize(fname) == 0:
            print("File\t%s  \tis empty!  Skipping..." % (fname))
            continue

        fin = open(fname, 'rb')  # Open the file
        Ntrees = np.fromfile(fin,np.dtype(np.int32),1)  # Read number of trees in file
        NtotGals = np.fromfile(fin,np.dtype(np.int32),1)[0]  # Read number of gals in file.
        TotNTrees = TotNTrees + Ntrees  # Update total sim trees number
        TotNGals = TotNGals + NtotGals  # Update total sim gals number
        goodfiles = goodfiles + 1  # Update number of files read for volume calculation
        fin.close()

    print()
    print("Input files contain:\t%d trees ;\t%d galaxies ." % (TotNTrees, TotNGals))
    print()

    # Initialize the storage array
    G = np.empty(TotNGals, dtype=Galdesc)

    offset = 0  # Offset index for storage array

    # Open each file in turn and read in the preamble variables and structure.
    print("Reading in files.")
    for fnr in range(first_file,last_file+1):
        fname = model_name+'_'+str(fnr)  # Complete filename

        if not os.path.isfile(fname):
            continue

        if getFileSize(fname) == 0:
            continue

        print("read_gals: Will read: ",fname)
        fin = open(fname, 'rb')  # Open the file
        Ntrees = np.fromfile(fin, np.dtype(np.int32), 1)  # Read number of trees in file
        NtotGals = np.fromfile(fin, np.dtype(np.int32), 1)[0]  # Read number of gals in file.
        np.fromfile(fin, np.dtype((np.int32, Ntrees)),1) # skip over GalsPerTree = number of gals in each tree
        print(":   Reading N=", NtotGals, "   \tgalaxies from file: ", fname)
        GG = np.fromfile(fin, Galdesc, NtotGals)  # Read in the galaxy structures

        FileIndexRanges.append((offset,offset+NtotGals))

        # Slice the file array into the global array
        # The copy() operation is required! Otherwise, we point to
        # the GG data, which changes from file to file, rather
        # than making a new copy of the data.
        # This comes from python's strategies for fast vector operations.
        #
        G[offset:offset+NtotGals]=GG[0:NtotGals].copy()

        del(GG)
        offset = offset + NtotGals  # Update the offset position for the global array

        fin.close()  # Close the file


    print()
    print("Total galaxies considered:", TotNGals)

    # Convert the Galaxy array into a recarray
    G = G.view(np.recarray)

    w = np.where(G.StellarMass > 1.0)[0]
    print("Galaxies more massive than 10^10Msun/h:", len(w))

    print()

    # Calculate the volume given the first_file and last_file
    #self.volume = self.BoxSize**3.0 * goodfiles / self.MaxTreeFiles

    return G

if __name__ == '__main__':

    from optparse import OptionParser
    import os

    parser = OptionParser()
    parser.add_option(
        '-d',
        '--dir_name',
        dest='DirName',
        default='./',
        help='input directory name (default: ./millennium/)',
        metavar='DIR',
        )
    parser.add_option(
        '-f',
        '--file_base',
        dest='FileName',
        default='model_z-0.000',
        help='filename base (default: model_z0.000)',
        metavar='FILE',
        )
    parser.add_option(
        '-n',
        '--file_range',
        type='int',
        nargs=2,
        dest='FileRange',
        default=(0, 7),
        help='first and last filenumbers (default: 0 7)',
        metavar='FIRST LAST',
        )

    (opt, args) = parser.parse_args()

    fin_base = opt.DirName + opt.FileName
    FirstFile = opt.FileRange[0]
    LastFile = opt.FileRange[1]
    print("fin_base=",fin_base)
    print("FirstFile=",FirstFile)
    print("LastFile=",LastFile)
    
    G = read_gals(fin_base, FirstFile, LastFile)

    f= open("galpos.dat","w+")
    for i in range(0,len(G.Pos[:,0])):
        f.write("%.3f %.3f %.3f\n" % (G.Pos[i,0], G.Pos[i,1], G.Pos[i,2]))
