import numpy as np
import scipy.integrate as integrate

speed_of_light = 299792.458 # in km/s
Myr_in_s = 31557600000000 # Julian Myr in s
M_sun_in_kg = 1.989e30

class coordinate_calc:

    def calc_distance(self, redshift, OmegaM, OmegaLambda, Hubble):
        """
        The function will calculate the distance to an hypothetical
        observer at a given redhsift .

        Parameter
        ------------
        redshift: float
        The redshift of the observer.
        OmegaM: float
        Current value of the matter density parameter, Omega_{M0}
        OmegaLambda: float
        Current value of the dark energy density parameter, Omega_{\Lambda0}
        Hubble: float
        Current value of the Hubble-Lemaitre parameter, H_0, in km/s/Mpc

        returns
        ------------
        The distance to an hypothetical observer at a given redshift.
        """

        integrand = lambda z: 1.0/(np.sqrt(OmegaM*(1.0+z)**3 + OmegaLambda))
        dist_int = integrate.romberg(integrand, 0, redshift, tol=1e-30, rtol=1.e-6, divmax=10)

        return speed_of_light/Hubble * dist_int # Mpc

    def calc_angle(self, void_radius, distance):
        """
        This function will calculate the angle over which a void will span
        for an hypothetical observer at a given distance. The calculation is
        done under the assumption that the effective void radius is a reasonable
        approximation for the size of the void.

        Paramter
        ------------
        void_radius: float
        The effective radius of a void.
        distance: float
        The distance to an observer.

        returns
        ------------
        The angle that the void will span over.
        """
        
        return np.arctan(void_radius/distance)

    # Not needed at the moment

    def convert_cartesian_to_spherical(self,x, y, z):

        if(x > 0 and y > 0 and z > 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z)
            phi = np.arctan(y/x)
        if(x > 0 and y < 0 and z > 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z)
            phi = np.arctan(y/x)
        if(x < 0 and y > 0 and z > 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z)
            phi = np.arctan(y/x) +np.pi
        if(x < 0 and y < 0 and z > 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z)
            phi = np.arctan(y/x) - np.pi

        if(x > 0 and y > 0 and z < 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z) + np.pi
            phi = np.arctan(y/x)
        if(x > 0 and y < 0 and z < 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z) + np.pi
            phi = np.arctan(y/x)
        if(x < 0 and y > 0 and z < 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z) + np.pi
            phi = np.arctan(y/x) + np.pi
        if(x < 0 and y < 0 and z < 0):
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arctan(np.sqrt(x**2 + y**2)/z)
            phi = np.arctan(y/x) - np.pi
        return r, theta, phi

    def convert_cartesian_to_spherical_old(self,x, y, z):

        r = np.sqrt(x**2 + y**2 + z**2)
        polar_angle = np.arccos(z/r)
        azimuth_angle = np.arctan(y/x)

        return r, polar_angle, azimuth_angle

    def convert_spherical_to_cartesian(self,r, polar_angle, azimuth_angle, offset):
        """
        This function converts from spherical coordinates to cartesian 
        coordinates. We use spherical coordinates to model the path a
        photon that arrives at the observer.

        Parameters
        ------------
        r, polar_angle, azimuth_angle: floats
        The spherical coordinates.
        offset: array(3,float)
        The position of the hypothetical observer.

        returns
        ------------
        x, y, z_ floats
        Converted cartesian coordinates.
        """

        x = r*np.sin(polar_angle)*np.cos(azimuth_angle) - offset[0]
        y = r*np.sin(polar_angle)*np.sin(azimuth_angle) - offset[1]
        z = r*np.cos(polar_angle) - offset[2]

        return x, y, z


    def cylinder_coord(self, r, phi, pos, box_length):

        x=pos[0]+r*np.cos(phi)
        z=pos[1]+r*np.sin(phi)

        if(x >= (box_length/2)):
            x=-box_length+x
        if(x <= -(box_length/2)):
            x=box_length+x
        if(z >= (box_length/2)):
            z=-box_length+z
        if(z <= -(box_length/2)):
            z=box_length+z

        return x,z


    def construct_cart_unit_vect(self, x, y, z):

        length = np.sqrt((x*x)+(y*y)+(z*z))

        return [x/length, y/length, z/length]
