#!/usr/bin/env python

# lensing - geometrical optics, including lensing
# (C) 2021, 2022, 2023 Marius Peper, Boud Roukema GPL-3+
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

## WARNING: As of 2022-06-21 all functions that use spherical
## coordniates are not used anymore and thus not updated. We
## keep the functions in the script if, in the future, somebody
## might find them useful and wants to use them.

# system level
import sys
import os
from os.path import getsize as getFileSize
import glob
from operator import eq
import warnings

# standard science packages
#import matplotlib
#matplotlib.use('Agg')
import numpy as np
###import random # this is for a python-level (not numpy) random generator
from numpy.random import SeedSequence, default_rng, PCG64, Generator
from scipy.interpolate import interpn
from scipy.interpolate import LinearNDInterpolator
import scipy.integrate as integrate
from scipy.signal import savgol_filter

from itertools import product, combinations
import time
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from multiprocessing import Pool

from extra_calculations import calculations
from lensing_calculations import lensing_calc_grid
from lensing_calculations import LensingCalc
from lensing_calculations_unit_testable import lensing_calc_unit_testable as lens_calc_ut
from coordinates import coordinate_calc
from read import Read
from figures import figures

import ctypes
diff_solver_path = os.environ.get('DIFF_SOLVER_PATH')
mass_deficit_solver = ctypes.CDLL("%s/optical_scalar_solver.so" % diff_solver_path)
call_gsl_integration = ctypes.CDLL("%s/call_gsl_integration.so" % diff_solver_path)
from numpy.ctypeslib import ndpointer

N_grad = 11
Nsnaps_analyse = 5

speed_of_light = 299792.458 # in km/s
Myr_in_s = 31557600000000
#M_sun_in_kg = 1.989e30

## DEFINE GLOBAL PARAMETER TO MAKE THEM USABLE IN THE WHOLE SCRIPT ##
# The parameter hold general information about the siulation,
# i.e. the size of the void, the scale factors of the used snapshots
# or the mean density.
# To save memory the 3d interpolations of the velocities of DM particles,
# the matter density of DM particles and the gravitational potential are
# also declared as global variables.
global box_length
global r_comoving_min
global r_comoving_max
global distance
global void_position_from_observer
global redshift
global Hubble
global mean_density
global Nfiles
global snap_times
global scale_factors
global void_radius
global rho
global pot_deriv
global pot_time_slope
global pot_time_curve
global dtdy
global dtdx
global dtdz
global scale_factor_dot
global pot_y_slope
global pot_y_curve
global dydx
global dydz
global pot_x_slope
global pot_x_curve
global dxdy
global dxdz
global pot_z_slope
global pot_z_curve
global dzdy
global dzdx
global pot
global radii_comoving
global x_analyse_grid
global z_analyse_grid
global vel_x
global vel_y
global vel_z
## Here we only want to have functions that are needed to start
## the parallelised calculations. The main routines should be
## found in lensing_calculations.py

def test_weyl_tensor_cart(axis1, axis2):

    orth_basis1, orth_basis2 = np.array([1,0,0]), np.array([0,0,1])

    orth_basis1, orth_basis2 = lens_calc.construct_orth_plane(axis1,distance,axis2)

    dyad_basis1 = np.array([0, orth_basis1[0], orth_basis1[1], orth_basis1[2]])
    dyad_basis2 = np.array([0, orth_basis2[0], orth_basis2[1], orth_basis2[2]])

    # TODO before using test_weyl_tensor_cart: fix the parameters to match those of the function definition:
    print("Software error: see the TODO.")
    pot = lens_calc.interpolate_detector_variable_cart(distance, axis1, axis2, void_position_from_observer, 0.0, potsInter)

    pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot = lens_calc.time_deriv_cart(Nfiles, distance, axis1, axis2, snap_times, inter_pots_times, void_position_from_observer, scale_factors)

    pot_y_slope, pot_y_curve, dydx, dydz = lens_calc.y_deriv(potsInter, distance, axis1, axis2, void_position_from_observer)
    pot_x_slope, pot_x_curve, dxdy, dxdz = lens_calc.x_deriv(potsInter, distance, axis1, axis2, void_position_from_observer)
    pot_z_slope, pot_z_curve, dzdy, dzdx = lens_calc.z_deriv(potsInter, distance, axis1, axis2, void_position_from_observer)

    # TODO, redshift is a function of radius
    WeylTensor1010,  diag1010 = lens_calc.calc_weyl_cart([1,0,1,0], pot, pot_time_slope, pot_x_slope, pot_y_slope, pot_z_slope, pot_time_curve, pot_x_curve, pot_y_curve, pot_z_curve, dxdy, dxdz, dydx, dydz, dzdx, dzdy, dtdx, dtdy, dtdz, scale_factor_dot, dyad_basis1, dyad_basis2, redshift, scale_factors, distance)
    WeylTensor1020,  diag1020 = lens_calc.calc_weyl_cart([1,0,2,0], pot, pot_time_slope, pot_x_slope, pot_y_slope, pot_z_slope, pot_time_curve, pot_x_curve, pot_y_curve, pot_z_curve, dxdy, dxdz, dydx, dydz, dzdx, dzdy, dtdx, dtdy, dtdz, scale_factor_dot, dyad_basis1, dyad_basis2, redshift, scale_factors, distance)

    poisson = pot_x_curve + pot_y_curve + pot_z_curve

    return WeylTensor1010, WeylTensor1020, diag1010, diag1020, poisson


def velocity_centre_cart(axis1, axis2, x_analyse_grid, z_analyse_grid):

    # maybe faster when not parallelised?

    # move the calculation of the unit vector to the prepare step?
    unit_vector = coord.construct_cart_unit_vect(x_analyse_grid[axis1], distance, z_analyse_grid[axis2])
    unit_vector_rad = coord.construct_cart_unit_vect(x_analyse_grid[axis1], 0, z_analyse_grid[axis2])

    pixel_vel_y = vel_y[:,axis1, axis2]
    pixel_radial_vel = vel_x[:,axis1, axis2]*unit_vector_rad[0] + vel_y[:,axis1, axis2]*unit_vector_rad[1] + vel_z[:,axis1, axis2]*unit_vector_rad[2]
    pixel_vel_magnitude = np.sqrt((vel_x[:,axis1, axis2]*vel_x[:,axis1, axis2]) + (vel_y[:,axis1, axis2]*vel_y[:,axis1, axis2]) + (vel_z[:,axis1, axis2]*vel_z[:,axis1, axis2]))
    pixel_vel_los = vel_x[:,axis1, axis2]*unit_vector[0] + vel_y[:,axis1, axis2]*unit_vector[1] + vel_z[:,axis1, axis2]*unit_vector[2]

    return pixel_radial_vel, pixel_vel_magnitude, pixel_vel_los, pixel_vel_y


def mass_deficit_cart(pixelx, pixelz):

    #print("polar: ", polar, "azimuth: ", azimuth)
    #density_line_of_sight = lambda r: lens_calc.interpolate_detector_variable_spherical(r, polar, azimuth, void_position_from_observer, mean_density, linInter)
    #integral = integrate.romberg(density_line_of_sight, r_comoving_min, r_comoving_max, rtol=1.e-3, divmax=20)
    #print("integral: ",integral)
    #mass_def = integral
    c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")

    #mass_deficit_solver.mass_deficit.argtypes = [c_input_type,
    #                                             c_input_type, ctypes.c_double,
    #                                             ctypes.c_double, ctypes.c_double]
    #mass_deficit_solver.mass_deficit.restype = ctypes.c_double
    #X = mass_deficit_solver.mass_deficit(radii_comoving, rho, mean_density, radii_comoving[0], radii_comoving[int(len(radii_comoving)) -1])
    #print("X: ",X)
    #mass_def = np.float(X)

    call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]
    call_gsl_integration.call_gsl_integration.restype = ctypes.c_double
    # rho = densities - mean_density! Units: m_dm/(Mpc/h)^3
    X = call_gsl_integration.call_gsl_integration(radii_comoving, np.array(rho[:, pixelx, pixelz]), radii_comoving[0], radii_comoving[int(len(radii_comoving)) -1], 1.e-30, rel_err_gsl, int(len(radii_comoving)))
    mass_def=float(X)

    # m_dm/(Mpc/h)^2
    return mass_def


def ISW_cart(pixelx, pixelz):

    #print("polar: ", polar, "azimuth: ", azimuth)
    #pot_deriv_line_of_sight = lambda r: lens_calc.time_deriv(Nfiles, r, polar, azimuth, snap_times, inter_pots_times, void_position_from_observer, scale_factors)[0]
    #integral = integrate.romberg(pot_deriv_line_of_sight, r_min, r_max, rtol=1.e-3, divmax=20)
    #print("integral: ",integral)
    #ISW = integral

    c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")

    #mass_deficit_solver.ISW.argtypes = [c_input_type,
    #                                    c_input_type,
    #                                    ctypes.c_double, ctypes.c_double]
    #X = mass_deficit_solver.ISW(radii_comoving, pot_deriv, radii_comoving[0], radii_comoving[int(len(radii_comoving)) -1])

    #print("radii_comoving: ",radii_comoving)

    call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]
    call_gsl_integration.call_gsl_integration.restype = ctypes.c_double
    X = call_gsl_integration.call_gsl_integration(radii_comoving, np.array(pot_time_slope[:,pixelx, pixelz]), radii_comoving[0], radii_comoving[int(len(radii_comoving)) -1], 1.e-30, rel_err_gsl, int(len(radii_comoving)))
    # eq. 2.1 https://arxiv.org/pdf/1212.0776.pdf
    # 2.7K is the mean temperature of the CMB
    # 1.e-6 to convert the speed of light from km/s to Mpc/Myr
    ISW = np.float(X)*2*2.7/(speed_of_light*1.e-6) # in Kelvin

    #ISW = np.trapz(pot_deriv,radii_comoving)

    return ISW

def doppler_convergence_cart(axis1, axis2):

    slices_y=[-void_radius, -void_radius/2, 0, void_radius/2, void_radius]
    convergence=np.zeros((len(slices_y)))

    i=0
    for y in slices_y:

        z_source=redshift
        unit_vector = coord.construct_cart_unit_vect(axis1, distance, axis2)

        # TODO before using doppler_convergence_cart: fix the parameters to match those of the function definition:
        print("Software error: see the TODO.")
        vel_x = lens_calc.interpolate_detector_variable_cart(y+distance, axis1, axis2, void_position_from_observer, 0.0, vels_xInter)
        vel_y = lens_calc.interpolate_detector_variable_cart(y+distance, axis1, axis2, void_position_from_observer, 0.0, vels_yInter)
        vel_z = lens_calc.interpolate_detector_variable_cart(y+distance, axis1, axis2, void_position_from_observer, 0.0, vels_zInter)

        convergence[i] = (1- (speed_of_light*(1+z_source)/(Hubble*(y+distance))) )*(np.inner(unit_vector,[vel_x, vel_y, vel_z])/speed_of_light)
        i=i+1

    return convergence


def solve_optical_scalars_cart(axis1, axis2):

    distance=void_position_from_observer[1]
    orth_basis1, orth_basis2 = lens_calc.construct_orth_plane(x_analyse_grid[axis1],distance,z_analyse_grid[axis2])

    dyad_basis1 = np.array([0, orth_basis1[0], orth_basis1[1], orth_basis1[2]])
    dyad_basis2 = np.array([0, orth_basis2[0], orth_basis2[1], orth_basis2[2]])

    diff_eq_sol = lens_calc.solve_diff_cart(radii_comoving, axis1, axis2, void_position_from_observer, redshift, mass_dm, Nfiles, snap_times, scale_factors, dyad_basis1, dyad_basis2, mean_density, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, (rho+mean_density), pot, box_length, (Hubble/100.0))

    return diff_eq_sol


def prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, vel_y_range, pots, circ_center, do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW):


    calc = calculations()

    # Passing the interpolated data to the Pool processes takes a lot of memory.
    # However, we use data points that we pass to C scripts. In these C scripts
    # we interpolate the data again. There is no need to pass the full interpolator
    # to the Pool processes. We exactly know the path of the lightray for each
    # pixel and thus can create an array with the data that is needed. This should
    # be a rather fast operation. These arrays can be passed to the Pool processes,
    # needed considerable less RAM.

    ## COMOVING REFERENCE FRAMES: Euclid_obs, Euclid_sim, spherical_obs
    #
    # void_position_observer is a positive value, e.g. 1322 Mpc/h for z=0.5 in LCDM
    #
    # The observer is at
    # Euclid_sim:          (0,-void_position_from_observer,0)
    # Euclid_obs:          (0,0,0)
    # spherical_obs:       (0,\forall,\forall)  # \forall means irrelevant
    #
    # The simulation centre is at:
    # Euclid_sim:          (0,0,0)
    # Euclid_obs:          (0,+void_position_from_observer,0)
    # spherical_obs:       (0,pi/2,pi/2)

    # axis2, axis1 in Euclid_sim
    axis2, axis1 = np.meshgrid(x_analyse_grid, z_analyse_grid)

    # rad, polar, azimuth in spherical_obs
    rad = np.zeros((len(x_analyse_grid), len(z_analyse_grid)))
    polar = np.zeros((len(x_analyse_grid), len(z_analyse_grid)))
    azimuth = np.zeros((len(x_analyse_grid), len(z_analyse_grid)))
    for i in range(0,len(rad)):
        for j in range(0,len(rad)):
            # 'distance' is the shift between Euclid_sim and Euclid_obs:
            #         y_{Euclid_obs} = y_{Euclid_sim} + distance
            rad[i,j], polar[i,j], azimuth[i,j] = coord.convert_cartesian_to_spherical(x_analyse_grid[i], distance, z_analyse_grid[j])


    if(do_test_weyl==True or do_mass_def==True or do_solve_scalars==True):

        # interpolations always are using the Euclid_sim coordinate system
        linInter = LinearNDInterpolator(moved_tracers, densities, fill_value=np.mean(densities))

        # interpolate densities using cartesian coordinates.
        # 'dens' should be similar to 'rho' when plotted for the same distance.
        #
        # void_position_from_observer = [0,distance,0] - the shift vector
        # to add to coordinates in Euclid_sim to obtain coordinates in
        # Euclid_obs
        dens=lens_calc.interpolate_detector_variable_cart(distance, axis1, axis2, void_position_from_observer, mean_density, box_length, linInter)
        rho = np.zeros((len(radii_comoving), N_pixels, N_pixels))

        for i in range(0,len(radii_comoving)):
            # rho = density - mean_density!
            # rho is using rad, polar, azimuth in spherical_obs as input
            rho[i, :, :] = lens_calc.interpolate_detector_variable_spherical(radii_comoving[i], polar, azimuth, void_position_from_observer, mean_density, box_length, linInter)

        del linInter
        del densities

        # return rho, dens, pixel_pot MASS_DEF

    if(do_velocity_centre==True or do_doppler_lensing==True):

        # As of 13.03.2023 the velocity calculations are not used.
        # We keep the basic structure here to enable the velocity
        # calculations in the future.

        vel_x=np.zeros((len(vel_y_range), len(x_analyse_grid), len(z_analyse_grid)))
        vel_y=np.zeros((len(vel_y_range), len(x_analyse_grid), len(z_analyse_grid)))
        vel_z=np.zeros((len(vel_y_range), len(x_analyse_grid), len(z_analyse_grid)))

        vels_xInter = LinearNDInterpolator(moved_tracers, vels[:,0], fill_value=np.mean(vels[:,0]))
        for i in range(0,len(vel_y_range)):
            vel_x[i,:,:]=lens_calc.interpolate_detector_variable_cart(vel_y_range[i]+distance, axis1, axis2, void_position_from_observer, 0.0, box_length, vels_xInter)
        del vels_xInter

        vels_yInter = LinearNDInterpolator(moved_tracers, vels[:,1], fill_value=np.mean(vels[:,1]))
        for i in range(0,len(vel_y_range)):
            vel_y[i,:,:]=lens_calc.interpolate_detector_variable_cart(vel_y_range[i]+distance, axis1, axis2, void_position_from_observer, 0.0, box_length, vels_yInter)
        del vels_yInter

        vels_zInter = LinearNDInterpolator(moved_tracers, vels[:,2], fill_value=np.mean(vels[:,2]))
        for i in range(0,len(vel_y_range)):
            vel_z[i,:,:]=lens_calc.interpolate_detector_variable_cart(vel_y_range[i]+distance, axis1, axis2, void_position_from_observer, 0.0, box_length, vels_zInter)
        del vels_zInter
        del vels

        #return vel_x, vel_y, vel_z VELOCITIES

    if(do_ISW==True or do_test_weyl==True or do_solve_scalars==True):

        # As of 13.03.2023 the ISW calculations are not used.
        # We keep the basic structure here to enable the ISW
        # calculations in the future.
        # This part will still be used to estimate
        # the optical scalars.

        times = np.zeros(Nsnaps_analyse) # Myr
        scale_factor = np.zeros(Nsnaps_analyse) # dimensionless
        pot_time_slope = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/Myr
        pot = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # dimensionless

        pots_time_line = np.zeros((Nsnaps_analyse,len(radii_comoving),N_pixels,N_pixels)) # dimensionless

        #if(do_test_weyl==True or do_solve_scalars==True):
        pots_time_line_y = np.zeros((Nsnaps_analyse,N_grad,len(radii_comoving),N_pixels,N_pixels)) # dimensionless
        pots_time_line_x = np.zeros((Nsnaps_analyse,N_grad,len(radii_comoving),N_pixels,N_pixels)) # dimensionless
        pots_time_line_z = np.zeros((Nsnaps_analyse,N_grad,len(radii_comoving),N_pixels,N_pixels)) # dimensionless
        if(do_test_weyl==True or do_solve_scalars==True):
            pot_time_curve = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/Myr^2
            dtdy = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # h/(Myr Mpc)
            dtdx = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # h/(Myr Mpc)
            dtdz = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # h/(Myr Mpc)
            scale_factor_dot = np.zeros(Nsnaps_analyse) # 1/Mpc
            pot_x_slope = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            pot_x_curve = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            pot_y_slope = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            pot_y_curve = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            pot_z_slope = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            pot_z_curve = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            dxdy = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            dxdz = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            dydx = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            dydz = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            dzdx = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            dzdy = np.zeros((len(radii_comoving),N_pixels,N_pixels)) # 1/(Mpc/h)^2
            mag_on_line_xz = np.zeros((N_grad,N_grad,len(radii_comoving),N_pixels,N_pixels))
            mag_on_line_yx = np.zeros((N_grad,N_grad,len(radii_comoving),N_pixels,N_pixels))
            mag_on_line_zy = np.zeros((N_grad,N_grad,len(radii_comoving),N_pixels,N_pixels))

        x = x_analyse_grid
        z = z_analyse_grid

        i=0
        for l in range(Nfiles-Nsnaps_analyse + 1, Nfiles+1):
            # TODO: Think of a better method to select the snapshots
            if(i==0):
                snap_number = Nfiles-15
            if(i==1):
                snap_number = Nfiles-10
            if(i==2):
                snap_number = Nfiles-6
            if(i==3):
                snap_number = analyse_snap
            if(i==4):
                snap_number = Nfiles
            if snap_number < 10:
                this_tracer_file = tracer_file_base + '00%d' % snap_number
            elif(snap_number < 100 and snap_number >= 10):
                this_tracer_file = tracer_file_base + '0%d' % snap_number
            else:
                this_tracer_file = tracer_file_base + '%d' % snap_number

            print("this_tracer_file: ", this_tracer_file)

            this_tracers = read.read_tracers(this_tracer_file) # Mpc/h
            #Optional for speed: select a random 1/8-th of the tracers instead of using all of them:
            if(N_subsample_potential < Ncroot**3):
                i_ran = rng_potentials.integers(0, high=Ncroot**3, size=N_subsample_potential)
                this_tracers=this_tracers[i_ran]

            scale_factor[i] = read.read_scale_factor(this_tracer_file)
            time_snap = calc.calc_time(OmegaM, OmegaLambda, Hubble, scale_factor[i]) # Myr

            this_moved_tracers = calc.move_coordinates(this_tracers[:,0], this_tracers[:,1], this_tracers[:,2], ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)
            this_pots = read.read_pots(this_tracer_file) # dimensionless
            if(N_subsample_potential < Ncroot**3):
                this_pots=this_pots[i_ran] # dimensionless
                del i_ran # no need for these Mb or so of memory any more

            this_pots_inter=LinearNDInterpolator(this_moved_tracers, this_pots, fill_value=np.mean(this_pots)) # dimensionless
            times[i] = time_snap # Myr

            print("prepare_interpolations: radii_comoving = ",radii_comoving)
            for k in range(0,len(radii_comoving)):
                # y, radii_comoving, polar, azimuth are in the spherical_obs frame:
                y = radii_comoving[k] # - void_position_from_observer[1]
                # Convert_spherical_to_cartesian() transforms spherical_obs to Euclid_obs:
                xx,yy,zz = coord.convert_spherical_to_cartesian(radii_comoving[k], polar, azimuth, np.array([0.0, 0.0, 0.0]))
                # xx, yy, zz are now in the Euclid_obs frame.

                if(i<2 and (k==0 or k==75)):
                    print("===============================")
                    print("prepare_interpolations: i,k = ",i,k)
                    print("prepare_interpolations: polar = ",polar)
                    print("prepare_interpolations: azimuth = ",azimuth)
                    print("prepare_interpolations: xx = ",xx)
                    print("prepare_interpolations: yy = ",yy)
                    print("prepare_interpolations: zz = ",zz)
                    print("===============================")

                #print("xx,yy,zz : ", xx,yy,zz)
                x_range = np.linspace((xx-5.0),(xx+5.0),N_grad) # x_range.shape = (N_grad,N_pixels,N_pixels)
                y_range = np.linspace((radii_comoving[k]-5.0),(radii_comoving[k]+5.0),N_grad)
                #print("y_range: ", y_range)
                z_range = np.linspace((zz-5.0),(zz+5.0),N_grad)
                # radii_comoving is a coordinate in spherical_obs
                # xx, yy, zz, x_range, y_range, z_range are all in Euclid_obs.
                pots_time_line[i,k,:,:] = lens_calc.interpolate_detector_variable_cart(radii_comoving[k], xx, zz, void_position_from_observer, 0.0, box_length, this_pots_inter)
                if(do_test_weyl==True or do_solve_scalars==True):
                    for j in range(0,N_grad):
                        pots_time_line_y[i,j,k,:,:] = lens_calc.interpolate_detector_variable_cart(y_range[j], xx, zz, void_position_from_observer, 0.0, box_length, this_pots_inter)
                        pots_time_line_x[i,j,k,:,:] = lens_calc.interpolate_detector_variable_cart(yy, x_range[j], zz, void_position_from_observer, 0.0, box_length, this_pots_inter)
                        pots_time_line_z[i,j,k,:,:] = lens_calc.interpolate_detector_variable_cart(yy, xx, z_range[j], void_position_from_observer, 0.0, box_length, this_pots_inter)
                        pot[k, :, :] = lens_calc.interpolate_detector_variable_cart(radii_comoving[k], xx, zz, void_position_from_observer, 0.0, box_length, this_pots_inter)
                        if(l==Nfiles):
                            for m in range(0,N_grad):
                                mag_on_line_xz[j,m,k,:,:] = lens_calc.interpolate_detector_variable_cart(radii_comoving[k], x_range[j], z_range[m], void_position_from_observer, 0.0, box_length, this_pots_inter)
                                mag_on_line_yx[j,m,k,:,:] = lens_calc.interpolate_detector_variable_cart(y_range[j], x_range[m], zz, void_position_from_observer, 0.0, box_length, this_pots_inter)
                                mag_on_line_zy[j,m,k,:,:] = lens_calc.interpolate_detector_variable_cart(y_range[m], xx, z_range[j], void_position_from_observer, 0.0, box_length, this_pots_inter)

            i=i+1


        if(do_mass_def==True):
            pixel_pot=lens_calc.interpolate_detector_variable_cart(distance, axis1, axis2, void_position_from_observer, 0.0, box_length, this_pots_inter)

        del this_pots
        del this_pots_inter
        if(do_test_weyl==True or do_solve_scalars==True):
            # radii_comoving is a coordinate in spherical_obs
            # xx, yy, zz, x_range, y_range, z_range are all in Euclid_obs
            pot_time_slope[:,:], pot_time_curve[:,:], dtdy[:,:], dtdx[:,:], dtdz[:,:], scale_factor_dot = lens_calc.time_deriv_cart(Nfiles, radii_comoving[k], xx, zz, times, void_position_from_observer, scale_factor, box_length, N_pixels, pots_time_line, pots_time_line_y, pots_time_line_x, pots_time_line_z)
            pot_x_slope, pot_x_curve, pot_y_slope, pot_y_curve, pot_z_slope, pot_z_curve, dxdy, dxdz, dydx, dydz, dzdx, dzdy = lens_calc.spatial_pot_derivs(mag_on_line_xz, mag_on_line_yx, mag_on_line_zy, xx, yy, zz)
        else:
            pot_time_slope[:,:] = lens_calc.time_deriv_cart(Nfiles, radii_comoving[k], xx, zz, times, void_position_from_observer, scale_factor, box_length, N_pixels, pots_time_line, pots_time_line_y, pots_time_line_x, pots_time_line_z)[0]

    # return pot_deriv ISW

    if(do_mass_def==True and do_ISW==False and do_test_weyl==False and do_solve_scalars==False):
        potsInter = LinearNDInterpolator(moved_tracers, pots, fill_value=np.mean(pots))
        pixel_pot=lens_calc.interpolate_detector_variable_cart(distance, axis1, axis2, void_position_from_observer, 0.0, box_length, potsInter)
        del potsInter
        del pots

    if( (do_test_weyl==True or do_solve_scalars==True) and do_mass_def==True and (do_velocity_centre==True or do_doppler_lensing==True) and do_ISW==True):
        return rho, dens, pixel_pot, vel_x, vel_y, vel_z, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, pot, times, scale_factor
    elif(do_mass_def==True and do_test_weyl==False and do_solve_scalars==False and do_velocity_centre==False and do_doppler_lensing==False and do_ISW==False):
        return rho, dens, pixel_pot
    elif(do_ISW==True and do_test_weyl==False and do_solve_scalars==False and do_velocity_centre==False and do_doppler_lensing==False and do_mass_def==False):
        return pot_time_slope
    elif(do_velocity_centre==True and do_test_weyl==False and do_solve_scalars==False and do_mass_def==False and do_doppler_lensing==False and do_ISW==False):
        return vel_x, vel_y, vel_z
    elif((do_solve_scalars==True or do_test_weyl==True) and do_velocity_centre==False and do_mass_def==False and do_doppler_lensing==False and do_ISW==False):
        return pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, rho, pot, times, scale_factor
    elif( (do_test_weyl==True or do_solve_scalars==True) and do_mass_def==True and (do_velocity_centre==False or do_doppler_lensing==False) and do_ISW==False):
        return rho, dens, pixel_pot, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, pot, times, scale_factor
    elif( (do_test_weyl==True or do_solve_scalars==True) and do_mass_def==True and do_ISW==True and (do_velocity_centre==False or do_doppler_lensing==False) ):
        return rho, dens, pixel_pot, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, pot, times, scale_factor
    elif(do_mass_def==True and do_ISW==True and do_test_weyl==False and do_solve_scalars==False and (do_velocity_centre==False or do_doppler_lensing==False) ):
        return rho, dens, pixel_pot, pot_time_slope

        #vel_x, vel_y, vel_z = prepare_velocity_input(x_analyse_grid, z_analyse_grid, offsets)
        #rho, dens, pixel_pot = prepare_mass_def_input(x_analyse_grid, z_analyse_grid)
        #pot_deriv=prepare_ISW_input(x_analyse_grid, z_analyse_grid)

    #inter_pots_times, snap_times, scale_factors = lens_calc.prepare_time_deriv(tracer_file_base, analyse_snap, Nfiles, OmegaM, OmegaLambda, Hubble, circ_center_gals[void_id_gals], box_length)



# TODO: shift suitable versions of these functions to extra_calculations.py

class lensing:

    def find_equi_lines_min(self, polar, azimuth, values, equi_fraction, identifier):

        equi_polar_index = []
        equi_azimuth_index = []
        minimal_value = np.min(values)
        maximal_value = np.max(values)
        equi_line_value = minimal_value + (equi_fraction*abs(maximal_value-minimal_value))
        random_pixel_value = np.zeros((N_pixels,N_pixels), dtype=float)
        random_pixel_adjacencies = np.zeros((N_pixels,N_pixels), dtype=float)

        i=0
        for beta in azimuth:
            l=0
            for alpha in polar:
                if(i != (N_pixels-1) and l != (N_pixels-1)):
                    if((equi_line_value < values[i+1,l] and equi_line_value > values[i-1,l])
                       or (equi_line_value < values[i,l+1] and equi_line_value > values[i,l-1])
                       or (equi_line_value < values[i+1,l+1] and equi_line_value > values[i-1,l-1])
                       or (equi_line_value < values[i-1,l] and equi_line_value > values[i+1,l])
                       or (equi_line_value < values[i,l-1] and equi_line_value > values[i,l+1])
                       or (equi_line_value < values[i-1,l-1] and equi_line_value > values[i+1,l+1])
                    ):
                        equi_polar_index.append(i)
                        equi_azimuth_index.append(l)
                        random_pixel_value[i,l]=0.0
                        random_pixel_adjacencies[i-1,l-1]=random_pixel_adjacencies[i-1,l-1]+1
                        random_pixel_adjacencies[i-1,l]=random_pixel_adjacencies[i-1,l]+1
                        random_pixel_adjacencies[i-1,l+1]=random_pixel_adjacencies[i-1,l+1]+1
                        random_pixel_adjacencies[i,l-1]=random_pixel_adjacencies[i,l-1]+1
                        random_pixel_adjacencies[i,l+1]=random_pixel_adjacencies[i,l+1]+1
                        random_pixel_adjacencies[i+1,l-1]=random_pixel_adjacencies[i+1,l-1]+1
                        random_pixel_adjacencies[i+1,l]=random_pixel_adjacencies[i+1,l]+1
                        random_pixel_adjacencies[i+1,l+1]=random_pixel_adjacencies[i+1,l+1]+1
                    else:
                        random_pixel_value[i,l]=float('NaN')
                l += 1
            i += 1



        #########################################################################################################

        # The following lines can be used to narrow down the "equipotential" lines. The problem is that the current version
        # will not uniquely determine the line, neighbouring pixels will be marked and we need a filter that decides which pixel
        # is the correct one to determine the equipotential line.
        # The following solution could be used multiple times to remove falsely marked pixels, the problem is that it either removes
        # too few or too many pixels.


        azimuth, polar = np.meshgrid(azimuth, polar, indexing = 'ij', sparse = False)

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'$\theta$')
        plt.xlabel(r'$\phi$')
        plt.ticklabel_format(axis='both', style='sci', scilimits=(0,0))
        c = ax.pcolormesh(azimuth, polar, random_pixel_adjacencies, vmin=0, vmax=8)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho / m_{DM}$', rotation = 270, labelpad=15)
        figname ="random_pixel_adjacencies_massdef" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'$\theta$')
        plt.xlabel(r'$\phi$')
        plt.ticklabel_format(axis='both', style='sci', scilimits=(0,0))
        c = ax.pcolormesh(azimuth, polar, values, vmin=np.min(values), vmax=np.max(values))
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho / m_{DM}$', rotation = 270, labelpad=15)
        figname ="values_massdef" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        #too_few_adjacencies_index=np.transpose(np.where((random_pixel_adjacencies <=3) ))
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0],too_few_adjacencies_index[:,1]]=0

        # the for-loop is not the best solution. Better is
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0]-1,too_few_adjacencies_index[:,1]-1]=random_pixel_adjacencies[too_few_adjacencies_index[:,0]-1,too_few_adjacencies_index[:,1]-1]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0]-1,too_few_adjacencies_index[:,1]]=random_pixel_adjacencies[too_few_adjacencies_index[:,0]-1,too_few_adjacencies_index[:,1]]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0]-1,too_few_adjacencies_index[:,1]+1]=random_pixel_adjacencies[too_few_adjacencies_index[:,0]-1,too_few_adjacencies_index[:,1]+1]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0],too_few_adjacencies_index[:,1]-1]=random_pixel_adjacencies[too_few_adjacencies_index[:,0],too_few_adjacencies_index[:,1]-1]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0],too_few_adjacencies_index[:,1]+1]=random_pixel_adjacencies[too_few_adjacencies_index[:,0],too_few_adjacencies_index[:,1]+1]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0]+1,too_few_adjacencies_index[:,1]-1]=random_pixel_adjacencies[too_few_adjacencies_index[:,0]+1,too_few_adjacencies_index[:,1]-1]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0]+1,too_few_adjacencies_index[:,1]]=random_pixel_adjacencies[too_few_adjacencies_index[:,0]+1,too_few_adjacencies_index[:,1]]-1
        #random_pixel_adjacencies[too_few_adjacencies_index[:,0]+1,too_few_adjacencies_index[:,1]+1]=random_pixel_adjacencies[too_few_adjacencies_index[:,0]+1,too_few_adjacencies_index[:,1]+1]-1
        # The problem is that this solution fails at the edges.

        #for this_index in range(0,len(too_few_adjacencies_index[:,0])):
        #    if(too_few_adjacencies_index[this_index,0]  != (N_pixels-1) and too_few_adjacencies_index[this_index,1] != (N_pixels-1)):
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]-1,too_few_adjacencies_index[this_index,1]-1]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]-1,too_few_adjacencies_index[this_index,1]-1]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]-1,too_few_adjacencies_index[this_index,1]]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]-1,too_few_adjacencies_index[this_index,1]]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]-1,too_few_adjacencies_index[this_index,1]+1]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]-1,too_few_adjacencies_index[this_index,1]+1]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0],too_few_adjacencies_index[this_index,1]-1]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0],too_few_adjacencies_index[this_index,1]-1]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0],too_few_adjacencies_index[this_index,1]+1]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0],too_few_adjacencies_index[this_index,1]+1]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]+1,too_few_adjacencies_index[this_index,1]-1]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]+1,too_few_adjacencies_index[this_index,1]-1]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]+1,too_few_adjacencies_index[this_index,1]]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]+1,too_few_adjacencies_index[this_index,1]]-1
        #        random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]+1,too_few_adjacencies_index[this_index,1]+1]=random_pixel_adjacencies[too_few_adjacencies_index[this_index,0]+1,too_few_adjacencies_index[this_index,1]+1]-1

        #########################################################################################################


        return random_pixel_value, equi_polar_index, equi_azimuth_index


    def find_equi_lines_max(self, polar, azimuth, values, equi_fraction, identifier):

        equi_polar_index = []
        equi_azimuth_index = []
        minimal_value = np.min(values)
        maximal_value = np.max(values)
        equi_line_value = maximal_value - (equi_fraction*abs(maximal_value-minimal_value))
        random_pixel_value = np.zeros((N_pixels,N_pixels), dtype=float)
        random_pixel_adjacencies = np.zeros((N_pixels,N_pixels), dtype=float)
        pixel_test = np.zeros((N_pixels,N_pixels), dtype=float)
        pixel_test_in = np.zeros((N_pixels,N_pixels), dtype=float)

        i=0
        for beta in azimuth:
            l=0
            for alpha in polar:
                pixel_test[i,l] = 1
                if(i != (N_pixels-1) and l != (N_pixels-1)):
                    if((equi_line_value < values[i+1,l] and equi_line_value > values[i-1,l])
                       or (equi_line_value < values[i,l+1] and equi_line_value > values[i,l-1])
                       or (equi_line_value < values[i+1,l+1] and equi_line_value > values[i-1,l-1])
                       or (equi_line_value < values[i-1,l] and equi_line_value > values[i+1,l])
                       or (equi_line_value < values[i,l-1] and equi_line_value > values[i,l+1])
                       or (equi_line_value < values[i-1,l-1] and equi_line_value > values[i+1,l+1])
                    ):
                        equi_polar_index.append(i)
                        equi_azimuth_index.append(l)
                        random_pixel_value[i,l]=0.0
                        pixel_test_in[i,l] = 1
                        random_pixel_adjacencies[i-1,l-1]=random_pixel_adjacencies[i-1,l-1]+1
                        random_pixel_adjacencies[i-1,l]=random_pixel_adjacencies[i-1,l]+1
                        random_pixel_adjacencies[i-1,l+1]=random_pixel_adjacencies[i-1,l+1]+1
                        random_pixel_adjacencies[i,l-1]=random_pixel_adjacencies[i,l-1]+1
                        random_pixel_adjacencies[i,l+1]=random_pixel_adjacencies[i,l+1]+1
                        random_pixel_adjacencies[i+1,l-1]=random_pixel_adjacencies[i+1,l-1]+1
                        random_pixel_adjacencies[i+1,l]=random_pixel_adjacencies[i+1,l]+1
                        random_pixel_adjacencies[i+1,l+1]=random_pixel_adjacencies[i+1,l+1]+1
                    else:
                        random_pixel_value[i,l]=float('NaN')
                l += 1
            i += 1

        azimuth, polar = np.meshgrid(azimuth, polar, indexing = 'ij', sparse = False)

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'$\theta$')
        plt.xlabel(r'$\phi$')
        plt.ticklabel_format(axis='both', style='sci', scilimits=(0,0))
        c = ax.pcolormesh(azimuth, polar, pixel_test, vmin=0, vmax=1)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho / m_{DM}$', rotation = 270, labelpad=15)
        figname ="pixeltest_" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'$\theta$')
        plt.xlabel(r'$\phi$')
        plt.ticklabel_format(axis='both', style='sci', scilimits=(0,0))
        c = ax.pcolormesh(azimuth, polar, pixel_test_in, vmin=0, vmax=1)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho / m_{DM}$', rotation = 270, labelpad=15)
        figname ="pixeltest_condition_" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'$\theta$')
        plt.xlabel(r'$\phi$')
        plt.ticklabel_format(axis='both', style='sci', scilimits=(0,0))
        c = ax.pcolormesh(azimuth, values, polar, vmin=np.percentile(values,10), vmax=np.percentile(values,90))
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho / m_{DM}$', rotation = 270, labelpad=15)
        figname ="values_pot" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'$\theta$')
        plt.xlabel(r'$\phi$')
        plt.ticklabel_format(axis='both', style='sci', scilimits=(0,0))
        c = ax.pcolormesh(azimuth, polar, random_pixel_adjacencies, vmin=0, vmax=8)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho / m_{DM}$', rotation = 270, labelpad=15)
        figname ="random_pixel_adjacencies_pot" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        return random_pixel_value, equi_polar_index, equi_azimuth_index


    def find_pixel_centre(self, polar, azimuth, equi_polar_index, equi_azimuth_index, distance):

        if(len(equi_polar_index) == len(equi_azimuth_index)):
            pixels_x=np.empty(len(equi_polar_index))
            pixels_y=np.empty(len(equi_polar_index))
            pixels_z=np.empty(len(equi_polar_index))
            i=0
            for this_pixel in equi_polar_index:
                this_pixel_polar=polar[this_pixel]
                this_pixel_azimuth=azimuth[this_pixel]
                pixels_x[i],pixels_y[i],pixels_z[i]=coord.convert_spherical_to_cartesian(distance, this_pixel_polar, this_pixel_azimuth, void_position_from_observer)
                i += 1
            avg_pixel_x=np.average(pixels_x)
            avg_pixel_y=np.average(pixels_y)
            avg_polar=np.average(polar[equi_polar_index])
            avg_azimuth=np.average(azimuth[equi_azimuth_index])
        else:
            print("Error: Something went wrong when finding the equipotential lines")

        return avg_pixel_x, avg_pixel_y, avg_polar, avg_azimuth


    def dens_test(self, G, void_gal_ids, void_radius, tracers, gal_pos, densities, circ_center, void_id, box_length, identifier):
        """
        The function is supposed to help us checking that the functions work
        properly. This function only creates two plots, it plots the tracers
        in the simulation for a chosen slice and the corresponding densities.

        Parameter
        ------------
        void_radius: float
        The length scale of the box. In our case we use the void_radius
        since we want to create gridded data of the void.
        tracers: np.array(3,dtype=float)
        The spatial coordinates of the DM particles.
        densities: np.array(dtype=float)
        The densities estimated in Revolver using the VTFE.
        circ_center: np.array(3)
        The position of the circumcentre of the void. It will be used to
        move our particles according to it.
        void_id: int
        Specifies which void we are going to investigate.

        returns
        ------------
        Two plots; The tracers in the simulation for a chosen slice and the
        corresponding densities for this slice.
        """
        calc = calculations()

        #densities = np.log(densities)

        #x = np.linspace(-2*void_radius, 2*void_radius, 40)
        #y = np.linspace(-2*void_radius, 2*void_radius, 40)

        x = np.linspace(-(0.95*box_length/2), (0.95*box_length/2), 128)
        #y = np.linspace(-(box_length/2), (box_length/2), 128)
        z = np.linspace(-(0.95*box_length/2), (0.95*box_length/2), 128)

        moved_tracers=calc.move_coordinates(tracers[:,0], tracers[:,1], tracers[:,2],
                                            ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)

        moved_gals=calc.move_coordinates(G.Pos[:,0], G.Pos[:,1], G.Pos[:,2],
                                         ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)


        gal_in_slice_nonvoid_ids = np.delete(np.arange(0,len(moved_gals)),void_gal_ids)
        gal_in_slice_nonvoid_index=np.where(abs(0 - moved_gals[gal_in_slice_nonvoid_ids,1]) <= 5)[0]
        gal_in_slice_nonvoid_ids=gal_in_slice_nonvoid_ids[gal_in_slice_nonvoid_index]
        void_gal_in_slice_index=np.where(abs(0 - moved_gals[void_gal_ids,1]) <= 5)[0]
        void_gal_in_slice_ids=void_gal_ids[void_gal_in_slice_index]


        x, z = np.meshgrid( x, z, indexing = 'ij', sparse = False)

        #moved_gal_pos = calc.move_coordinates(gal_pos[:,0], gal_pos[:,1], gal_pos[:,2],
        #                                      circ_center[void_id], box_length)

        linInter= LinearNDInterpolator(moved_tracers, densities, fill_value=np.mean(densities))
        #linInter= LinearNDInterpolator(moved_gal_pos, densities)

        ax = plt.subplot(1, 1, 1)

        #FD_void = patches.Rectangle((-void_radius,-void_radius),
        #                            (2*void_radius),(2*void_radius),
        #                            linewidth=1, edgecolor='r',
        #                            facecolor='none')
        #ax.add_patch(FD_void)
        plt.xlabel("x in Mpc/h")
        plt.ylabel("z in Mpc/h")

        #z=0
        y=0
        g = linInter(x,y,z)/mean_density
        c=ax.pcolormesh(x, z, g, vmin=np.percentile(g,10), vmax=np.percentile(g,90))
        cbar=plt.colorbar(c,ax=ax)
        void_gal_plot=plt.scatter(moved_gals[void_gal_in_slice_ids,0], moved_gals[void_gal_in_slice_ids,2],
                                  marker="o", color="blue", s=1, label="void galaxies")
        nonvoid_gal_plot=plt.scatter(moved_gals[gal_in_slice_nonvoid_ids,0], moved_gals[gal_in_slice_nonvoid_ids,2],
                                     marker="o", color="red", s=1, label="non-void galaxies")
        void_gal_plot.set_alpha(1)
        nonvoid_gal_plot.set_alpha(1)
        cbar.set_label(r'$\rho / \bar{\rho}$', rotation = 270, labelpad=15)
        figname="densities_full" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        ax = plt.subplot(1, 1, 1)
        plt.xlabel("x in Mpc/h")
        plt.ylabel("z in Mpc/h")
        c=ax.pcolormesh(x, z, g)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\rho (m_{DM})$', rotation = 270, labelpad=15)
        figname="densities_full_no_cut" + identifier + ".eps"
        plt.savefig(figname)
        plt.close()

        ## plot particles to check the density.

        particle_ids=np.where(abs(moved_tracers[:,1]) < 1)[0]

        print("particle_ids.size=",particle_ids.size)

        plt.scatter(moved_tracers[particle_ids,0], moved_tracers[particle_ids,2], marker='.', alpha=0.1)

        plt.xlabel("x in Mpc/h")
        plt.ylabel("z in Mpc/h")
        figname="tracers_full" + identifier + ".png"
        plt.savefig(figname, format='png')
        plt.close()


    def plot_first_snap_pots(self, tracer_file_base, circ_center_gals, void_id_gals):

        read = Read()
        calc = calculations()
        x = np.linspace(-(0.95*box_length/2), (0.95*box_length/2), 128)
        #y = np.linspace(-(box_length/2), (box_length/2), 128)
        z = np.linspace(-(0.95*box_length/2), (0.95*box_length/2), 128)

        tracer_file = tracer_file_base + '000'
        tracers = read.read_tracers(tracer_file)
        pots = read.read_pots(tracer_file)
        moved_tracers=calc.move_coordinates(tracers[:,0], tracers[:,1], tracers[:,2],
                                            ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)

        x, z = np.meshgrid( x, z, indexing = 'ij', sparse = False)
        y=0
        linInter= LinearNDInterpolator(moved_tracers, pots, fill_value=np.mean(pots))
        g = linInter(x,y,z)

        ax = plt.subplot(1, 1, 1)
        plt.xlabel("x in Mpc/h")
        plt.ylabel("z in Mpc/h")
        c=ax.pcolormesh(x, z, g)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\Phi$', rotation = 270, labelpad=15)
        figname="initial_potential_0.eps"
        plt.savefig(figname)
        plt.close()

        tracer_file = tracer_file_base + '001'
        tracers = read.read_tracers(tracer_file)
        pots = read.read_pots(tracer_file)
        moved_tracers=calc.move_coordinates(tracers[:,0], tracers[:,1], tracers[:,2],
                                            ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)

        #x, z = np.meshgrid( x, z, indexing = 'ij', sparse = False)
        #y=0
        linInter= LinearNDInterpolator(moved_tracers, pots, fill_value=np.mean(pots))
        g = linInter(x,y,z)

        ax = plt.subplot(1, 1, 1)
        plt.xlabel("x in Mpc/h")
        plt.ylabel("z in Mpc/h")
        c=ax.pcolormesh(x, z, g)
        cbar=plt.colorbar(c,ax=ax)
        cbar.set_label(r'$\Phi$', rotation = 270, labelpad=15)
        figname="initial_potential_1.eps"
        plt.savefig(figname)
        plt.close()


if __name__ == '__main__':

    from optparse import OptionParser
    import os

    box_length=float(sys.argv[1])
    print(box_length)

    time_interval=float(sys.argv[2])
    print(time_interval)

    Nfiles=int(sys.argv[3])
    print(Nfiles)

    analyse_snap=int(sys.argv[4])
    print(analyse_snap)

    N_stack=int(sys.argv[5])
    print(N_stack)

    parser = OptionParser()
    parser.add_option(
        '-d',
        '--dir_name',
        dest='DirName',
        default='./',
        help='input directory name (default: ./millennium/)',
        metavar='DIR',
        )
    parser.add_option(
        '-f',
        '--file_base',
        dest='FileName',
        default='model_z-0.000',
        help='filename base (default: model_z0.000)',
        metavar='FILE',
        )
    parser.add_option(
        '-n',
        '--file_range',
        type='int',
        nargs=2,
        dest='FileRange',
        default=(0, 7),
        help='first and last filenumbers (default: 0 7)',
        metavar='FIRST LAST',
        )
    parser.add_option(
        '-t',
        '--tracer_file_base',
        dest='TracerFile',
        default='./out_gadget',
        help='path and filename of tracer file',
        metavar='FILE',
        )
    parser.add_option(
        '-v',
        '--voids_dm_directory',
        dest='DMVoidsDirectory',
        default='./Revolver',
        help='path to the Revolver output files',
        metavar='Revolver_directory',
        )
    parser.add_option(
        '-g',
        '--voids_gals_directory',
        dest='GalVoidsDirectory',
        default='./Revolver',
        help='path to the Revolver output files',
        metavar='Revolver_directory',
        )
    parser.add_option(
        '-l',
        '--lensing_directory',
        dest='LensingDirectory',
        default='./lensing-calculations',
        help='path to the lensing calculations output files',
        metavar='Lensing_directory',
        )

    (opt, args) = parser.parse_args()

    tracer_file_base = opt.TracerFile
    DM_voids_dir = opt.DMVoidsDirectory
    Gal_voids_dir = opt.GalVoidsDirectory
    lensing_dir = opt.LensingDirectory
    densities_file = DM_voids_dir + '/' + 'outputrawZOBOV/test.vol'
    densities_file_gal = Gal_voids_dir + '/' + 'outputrawZOBOV/test.vol'
    adjacency_file = DM_voids_dir + '/' + 'outputrawZOBOV/test.adj'
    void_file_dm = DM_voids_dir + '/' + 'outputzobov-Voids_cat.txt'
    #void_file_gals = Gal_voids_dir + '/' + 'outputzobov-Voids_cat.txt'
    # Use geometrical centroids (misleadingly named "bary"centres):
    void_file_gals = Gal_voids_dir + '/' + 'outputbarycentres/zobov-Voids_baryC_cat.txt'
    void_matches_file_base = lensing_dir + '/' + 'void_matches'
    if analyse_snap >= 100:
        tracer_file = tracer_file_base + '%d' % analyse_snap
    elif 10 <= analyse_snap < 100:
        tracer_file = tracer_file_base + '0%d' % analyse_snap
    elif analyse_snap < 10:
        tracer_file = tracer_file_base + '00%d' % analyse_snap
    id_file = DM_voids_dir + '/Revolver_Ids.txt'
    void_gal_id_file = Gal_voids_dir + '/Revolver_Ids_circ.txt'
    FirstFile = opt.FileRange[0]
    LastFile = opt.FileRange[1]
    fin_base = opt.DirName + opt.FileName

    # set variables from the simulation
    if os.environ.get('NCROOT'):
        Ncroot = int(os.environ.get('NCROOT'))
    else:
        Ncroot = 32
    if os.environ.get('OMEGA_M'):
        OmegaM = float(os.environ.get('OMEGA_M'))
    else:
        OmegaM = 0.3
    if os.environ.get('OMEGA_LAMBDA'):
        OmegaLambda = float(os.environ.get('OMEGA_LAMBDA'))
    else:
        OmegaLambda = 0.7

    if os.environ.get('HUBBLE0'):
        Hubble = float(os.environ.get('HUBBLE0'))
    else:
        Hubble = 70.0

    # set distance to the observer
    if os.environ.get('MAIN_VOID_REDSHIFT'):
        redshift = float(os.environ.get('MAIN_VOID_REDSHIFT'))
    else:
        redshift = 1.0

    if os.environ.get('GAL_DIR'):
        gal_dir = str(os.environ.get('GAL_DIR'))
    else:
        gal_dir = './'

    if os.environ.get('EQUI_LINE_PERCENT1'):
        equi_line_percent1 = float(os.environ.get('EQUI_LINE_PERCENT1'))
    else:
        equi_line_percent1 = 0.02

    if os.environ.get('PROB_THRESHOLD'):
        prob_threshold = float(os.environ.get('PROB_THRESHOLD'))
    else:
        prob_threshold = 0.5

    if os.environ.get('NCPU_LENSING'):
        NCPU = int(os.environ.get('NCPU_LENSING'))
    else:
        NCPU = 4

    if os.environ.get('N_PIXELS'):
        N_pixels = int(os.environ.get('N_PIXELS'))
    else:
        N_pixels = 64

    if os.environ.get('N_SUBSAMPLE_POTENTIAL'):
        N_subsample_potential = int(os.environ.get('N_SUBSAMPLE_POTENTIAL'))
    else:
        N_subsample_potential = 2097152


    if os.environ.get('VOID_STDDEV_FRAC_CRIT'):
        void_stddev_frac_crit = float(os.environ.get('VOID_STDDEV_FRAC_CRIT'))
    else:
        void_stddev_frac_crit = 0.5
    if os.environ.get('VOID_DIFF_FRAC_CRIT'):
        void_diff_frac_crit = float(os.environ.get('VOID_DIFF_FRAC_CRIT'))
    else:
        void_diff_frac_crit = 1.5
    if os.environ.get('VOID_STDDEV_FRAC_CRIT_THETA'):
        void_stddev_frac_crit_theta = float(os.environ.get('VOID_STDDEV_FRAC_CRIT_THETA'))
    else:
        void_stddev_frac_crit_theta = 0.5
    if os.environ.get('VOID_DIFF_FRAC_CRIT_THETA'):
        void_diff_frac_crit_theta = float(os.environ.get('VOID_DIFF_FRAC_CRIT_THETA'))
    else:
        void_diff_frac_crit_theta = 1.5
    if os.environ.get('VOID_STDDEV_FRAC_CRIT_SIGMA'):
        void_stddev_frac_crit_sigma = float(os.environ.get('VOID_STDDEV_FRAC_CRIT_SIGMA'))
    else:
        void_stddev_frac_crit_sigma = 0.5
    if os.environ.get('VOID_DIFF_FRAC_CRIT_SIGMA'):
        void_diff_frac_crit_sigma = float(os.environ.get('VOID_DIFF_FRAC_CRIT_SIGMA'))
    else:
        void_diff_frac_crit_sigma = 1.5
    if os.environ.get('VOID_GAMMA_FRAC_SLOPE_MAX_PERCENTILE'):
        void_gamma_frac_slope_max_percentile = float(os.environ.get('VOID_GAMMA_FRAC_SLOPE_MAX_PERCENTILE'))
    else:
        void_gamma_frac_slope_max_percentile = 0.45


    if os.environ.get('N_AVERAGE'):
        N_average = int(os.environ.get('N_AVERAGE'))
    else:
        N_average = 7
    if os.environ.get('N_NEIGHBOURS'):
        N_neighbours = int(os.environ.get('N_NEIGHBOURS'))
    else:
        N_neighbours = 5
    if os.environ.get('N_SIGNIFICANCE'):
        N_significance = int(os.environ.get('N_SIGNIFICANCE'))
    else:
        N_significance = 7
    if os.environ.get('TOLERANCE_FRAC_STD'):
        tolerance_frac_std = float(os.environ.get('TOLERANCE_FRAC_STD'))
    else:
        tolerance_frac_std = 0.65
    if os.environ.get('TOLERANCE_FRAC_MEAN'):
        tolerance_frac_mean = float(os.environ.get('TOLERANCE_FRAC_MEAN'))
    else:
        tolerance_frac_mean = 0.85
    if os.environ.get('MERGE_DISTANCE'):
        merge_distance = float(os.environ.get('MERGE_DISTANCE'))
    else:
        merge_distance = 10.0

    # Widths of distributions for comparing photometrically detected voids
    # with intrinsic 3D voids. The initial model uses Gaussians, so 'std'
    # means 'standard deviation' of the Gaussian.
    if os.environ.get('STD_XZ_VOID_PROB'):
        std_xz_void_prob = float(os.environ.get('STD_XZ_VOID_PROB'))
    else:
        std_xz_void_prob = 5.0 # Mpc/h linear scale
    if os.environ.get('STD_R_VOID_PROB'):
        std_R_void_prob = float(os.environ.get('STD_R_VOID_PROB'))
    else:
        std_R_void_prob = 0.1 # log10 of ratio of void radii (unit = "dex")

    if os.environ.get('REL_ERR_GSL'):
        rel_err_gsl = float(os.environ.get('REL_ERR_GSL'))
    else:
        rel_err_gsl = 1e-7

    if os.environ.get('COSMIC_RAY_PERCENTILE'):
        cosmic_ray_percentile = float(os.environ.get('COSMIC_RAY_PERCENTILE'))
    else:
        cosmic_ray_percentile = 90.0


    if os.environ.get('LENSING_SEED_SAMPLE_POTENTIALS'):
        lensing_seed_sample_potentials = int(os.environ.get('LENSING_SEED_SAMPLE_POTENTIALS'))
    else:
        lensing_seed_sample_potentials = 98717
    if(lensing_seed_sample_potentials >= 0):
        rng_potentials = Generator(PCG64(SeedSequence(lensing_seed_sample_potentials)))
    else:
        rng_potentials = np.random.default_rng()

    if os.environ.get('LENSING_SEED_SAMPLE_VELOCITIES'):
        lensing_seed_sample_velocities = int(os.environ.get('LENSING_SEED_SAMPLE_VELOCITIES'))
    else:
        lensing_seed_sample_velocities = 94777
    if(lensing_seed_sample_velocities >= 0):
        rng_velocities = Generator(PCG64(SeedSequence(lensing_seed_sample_velocities)))
    else:
        rng_velocities = np.random.default_rng()

    if os.environ.get('TEST_WEYL'):
        if(str(os.environ.get('TEST_WEYL'))=="True"):
            do_test_weyl = True
        else:
            do_test_weyl = False
    else:
        do_test_weyl = False
    if os.environ.get('VELOCITY_CENTRE'):
        if(str(os.environ.get('VELOCITY_CENTRE'))=="True"):
            do_velocity_centre = True
        else:
            do_velocity_centre = False
    else:
        do_velocity_centre = False
    if os.environ.get('MASS_DEF'):
        if(str(os.environ.get('MASS_DEF'))=="True"):
            do_mass_def = True
        else:
            do_mass_def = False
    else:
        do_mass_def = False
    if os.environ.get('ISW'):
        if(str(os.environ.get('ISW'))=="True"):
            do_ISW = True
        else:
            do_ISW = False
    else:
        do_ISW = False
    if os.environ.get('DOPPLER_LENSING'):
        if(str(os.environ.get('DOPPLER_LENSING'))=="True"):
            do_doppler_lensing = True
        else:
            do_doppler_lensing = False
    else:
        do_doppler_lensing = False
    if os.environ.get('OPTICAL_SCALARS'):
        if(str(os.environ.get('OPTICAL_SCALARS'))=="True"):
            do_solve_scalars = True
        else:
            do_solve_scalars = False
    else:
        do_solve_scalars = False
    if os.environ.get('STACKING'):
        if(str(os.environ.get('STACKING'))=="YES"):
            stacking_step = True
        else:
            stacking_step = False
    else:
        stacking_step = False


    print("do_test_weyl:", do_test_weyl)
    print("do_velocity_centre: ",do_velocity_centre)
    print("do_mass_def: ",do_mass_def)
    print("do_ISW: ", do_ISW)
    print("do_doppler_lensing: ",do_doppler_lensing)
    print("do_solve_scalars: ", do_solve_scalars)
    read = Read()
    calc = calculations()
    lens_calc_grid=lensing_calc_grid()
    lens_calc=LensingCalc() # create instance of class LensingCalc
    lensing=lensing() # create instance of class 'lensing'
    coord=coordinate_calc()
    figs=figures()

    # Create and assign data attributes (input parameters) that should not
    # be changed.
    #
    # TODO: shift other input parameters here instead of having incredibly
    # long lists of parameters to pass.
    lens_calc.void_gamma_frac_slope_max_percentile = void_gamma_frac_slope_max_percentile

    print("read")

    time_interval_in_sec = time_interval*Myr_in_s
    #eff_void_radius_dm = read.read_void_radius(void_file_dm)
    #num_voids_dm = eff_void_radius_dm.size

    print("fin_base: ",fin_base)
    print("FirstFile: ",FirstFile)
    print("LastFile: ",LastFile)
    G = read.read_gals(fin_base, FirstFile, LastFile)
    #circ_center_dm = read.read_circ_center(void_file_dm,num_voids_dm)

    print("tracer_file: ", tracer_file)

    vels = read.read_vels(tracer_file) # unit: km/s
    #print("vels.shape: ", vels.shape)
    if(N_subsample_potential < Ncroot**3):
        i_ran = rng_velocities.integers(0, high=Ncroot*3, size=N_subsample_potential)
        vels = vels[i_ran]
    #print("vels.shape: ", vels.shape)
    tracers = read.read_tracers(tracer_file) # unit: Mpc/h
    #print("sys.getsizeof(tracers): ", sys.getsizeof(tracers))
    #tracers = tracers[i_ran]
    #print("sys.getsizeof(tracers): ", sys.getsizeof(tracers))
    gal_pos = read.read_galpos(gal_dir + '/' + 'galpos.dat') # unit: Mpc/h
    #densities = read.read_densities(densities_file,tracer_file,box_length)
    #print("densities.shape: ", densities.shape)
    #densities = densities[i_ran]
    #print("densities.shape: ", densities.shape)
    pots = read.read_pots(tracer_file) # unit: dimensionless
    #pots = pots[i_ran]

    densities=np.load("densities_voronoi.npy") # unit: m_DM/(Mpc/h)^3 = 1.e10*M_sun_in_kg/(Mpc/h)^3


    if(N_subsample_potential < Ncroot**3):
        del i_ran # no need for these Mb or so of memory any more

    #dm_adjacencies = read.read_adjacencies(adjacency_file, len(tracers))
    #dm_adjacencies = dm_adjacencies[i_ran]

    #pots = calc.adjacencies_smoothing(pots, dm_adjacencies)
    #densities = calc.adjacencies_smoothing(densities, dm_adjacencies)

    mass_dm = read.read_mass_dm(tracer_file) # 1.e10*M_sun_in_kg

    print("mass_dm: ",mass_dm)

    mean_density = lens_calc.calc_mean_density(box_length, Ncroot, tracer_file) # = Npart/Volume => unit: m_DM/(Mpc/h)^3

    print("densities: ",densities)
    vols=1/densities
    print("vols: ", vols)
    print("np.sum(vols): ", np.sum(vols))
    print("np.sum(densities): ", np.sum(densities))
    print("np.mean(densities): ", np.mean(densities))
    print("np.mean(pots): ", np.mean(pots))

    print("mean_density: ",mean_density)
    distance = coord.calc_distance(redshift, OmegaM, OmegaLambda, Hubble)*(Hubble/100.0) # Mpc/h
    distance_file=open("observer_void_distance.dat", "w+")
    distance_file.write("observer_void_distance=%f\n" % distance )
    distance_file.close()

    eff_void_radius_gals = read.read_void_radius(void_file_gals)
    num_voids_gals = eff_void_radius_gals.size
    void_id_gals = read.largest_void(eff_void_radius_gals)
    circ_center_gals = read.read_circ_center(void_file_gals,num_voids_gals)

    void_ids = np.where(eff_void_radius_gals>20)[0]
    np.save(("void_ids_%i") % N_stack, void_ids)
    np.save(("eff_void_radius_gals_%i") % N_stack, eff_void_radius_gals)

    best_void_index=np.zeros((3,len(eff_void_radius_gals)))
    best_void_index[:]=-1
    lensing_void_radii=np.zeros((3,len(eff_void_radius_gals)))
    # Attention: This also sets the upper limit for the void size
    rads=np.linspace(0.1, box_length/3, 101)

    #for void_count in range(0,len(void_ids)):
    for void_count in range(0,1):
        print("void_count: ", void_count)
        #void_id_gals=void_ids[void_count]
        void_id_gals = read.largest_void(eff_void_radius_gals)
        print("void_id_gals: ", void_id_gals)

        void_radius = eff_void_radius_gals[void_id_gals]

        # TODO: put it in the ini?
        void_position_from_observer = [0,distance,0]

        # estimate the patch of the simulation that is of interest
        #id_file_void_id, dm_in_void_id = read.dm_in_void(Ncroot, id_file)
        id_file_gal_in_void_id, gal_in_void_id = read.dm_in_void((int(len(G.Pos[:,0])**(1/3)+1)), void_gal_id_file)

        where_in_revolver_file=void_id_gals
        void_gal_ids = gal_in_void_id[where_in_revolver_file]

        moved_gal_pos = calc.move_coordinates(gal_pos[:,0], gal_pos[:,1], gal_pos[:,2],
                                              ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)

        moved_revolver_centres = calc.move_coordinates(circ_center_gals[:,0], circ_center_gals[:,1], circ_center_gals[:,2],
                                                       ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)

        moved_tracers=calc.move_coordinates(tracers[:,0], tracers[:,1], tracers[:,2],
                                            ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)

        void_radius = eff_void_radius_gals[void_id_gals]

        # TODO: Find a good place for this function - velocities part?
        redshift_mean,redshift_std,gal_redshift=lens_calc.galaxy_redshift_centre(G, void_gal_ids, moved_gal_pos, redshift, distance, circ_center_gals, void_id_gals, box_length)
        np.save(("gal_redshift_%i_%i" % (N_stack, void_count)), gal_redshift)

        if(stacking_step==False):
            gal_z_out_file=open(("gal_redshift_%i.dat") % void_count,"w+")
            gal_z_out_file.write("centre_redshift=%f\nredshift_std=%f\n" % (redshift_mean, redshift_std) )
            gal_z_out_file.write("N_voidgal=%i\n" % (len(gal_redshift)) )
            gal_z_out_file.close()

        # TODO: verify direction of integration
        x_analyse_grid = np.linspace(-(box_length/2), (box_length/2), N_pixels) # Mpc/h
        z_analyse_grid = np.linspace(-(box_length/2), (box_length/2), N_pixels) # Mpc/h
        r_comoving_max = distance + 0.95*(box_length/2) # Mpc/h
        r_comoving_min = distance - 0.95*(box_length/2) # Mpc/h
        offsets = [-void_radius, -void_radius/2, 0, void_radius/2, void_radius] # Mpc/h
        # WARNING: hardwired value: changing the number of radii_comoving would require recompiling the diff solver codes!
        # TODO: make this more robust against changes (DRY vs WET: https://en.wikipedia.org/wiki/Don%27t_repeat_yourself)
        radii_comoving = np.linspace(r_comoving_min, r_comoving_max, 150)

        #lensing.plot_first_snap_pots(tracer_file_base, circ_center_gals, void_id_gals)

        if(do_mass_def==True and (do_test_weyl==True or do_solve_scalars==True) and (do_velocity_centre==True or do_doppler_lensing==True) and do_ISW==True):
            rho, dens, pixel_pot, vel_x, vel_y, vel_z, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, pot, snap_times, scale_factors = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif(do_mass_def==True and do_test_weyl==False and do_solve_scalars==False and do_velocity_centre==False and do_doppler_lensing==False and do_ISW==False):
            rho, dens, pixel_pot = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif(do_ISW==True and do_test_weyl==False and do_solve_scalars==False and do_velocity_centre==False and do_doppler_lensing==False and do_mass_def==False):
            pot_time_slope = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif(do_velocity_centre==True and do_test_weyl==False and do_solve_scalars==False and do_mass_def==False and do_doppler_lensing==False and do_ISW==False):
            vel_x, vel_y, vel_z = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif((do_solve_scalars==True or do_test_weyl==True) and do_velocity_centre==False and do_mass_def==False and do_doppler_lensing==False and do_ISW==False):
            pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, rho, pot, snap_times, scale_factors = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif( (do_test_weyl==True or do_solve_scalars==True) and do_mass_def==True and (do_velocity_centre==False or do_doppler_lensing==False) and do_ISW==False):
            rho, dens, pixel_pot, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, pot, snap_times, scale_factors = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif( (do_test_weyl==True or do_solve_scalars==True) and do_mass_def==True and do_ISW==True and (do_velocity_centre==False or do_doppler_lensing==False)):
            rho, dens, pixel_pot, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, pot, snap_times, scale_factors = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)
        elif( do_mass_def==True and do_ISW==True and do_test_weyl==False and do_solve_scalars==False and (do_velocity_centre==False or do_doppler_lensing==False)):
            rho, dens, pixel_pot, pot_time_slope = prepare_interpolations(radii_comoving, x_analyse_grid, z_analyse_grid, moved_tracers, densities, vels, offsets, pots, circ_center_gals[void_id_gals], do_test_weyl, do_mass_def, do_solve_scalars, do_velocity_centre, do_doppler_lensing, do_ISW)

        print("rho ", rho)
        print("dens" , dens)
        print("pixel_pot ", pixel_pot)
        print("pot_time_slope ", pot_time_slope)
        print("pot_time_curve ", pot_time_curve)
        print("dtdy ", dtdy)
        print("dtdx ", dtdx)
        print("dtdz ", dtdz)
        print("scale_factor_dot ", scale_factor_dot)
        print("pot_y_slope ", pot_y_slope)
        print("pot_y_curve ", pot_y_curve)
        print("dydx ", dydx)
        print("dydz ", dydz)
        print("pot_x_slope ", pot_x_slope)
        print("pot_x_curve ", pot_x_curve)
        print("dxdy ", dxdy)
        print("dxdz ", dxdz)
        print("pot_z_slope ", pot_z_slope)
        print("pot_z_curve ", pot_z_curve)
        print("dzdy ", dzdy)
        print("dzdx ",dzdx)
        print("pot ", pot)
        print("snap_times ", snap_times)
        print("scale_factors ", scale_factors)

        if(do_test_weyl==True):

            # TODO: This routine is outdated!
            x_analyse_grid = np.linspace(-(analyse_size_void_cart/2), (analyse_size_void_cart/2), N_pixels)
            z_analyse_grid = np.linspace(-(analyse_size_void_cart/2), (analyse_size_void_cart/2), N_pixels)
            r_comoving_max = distance + 0.95*(box_length/2)
            r_comoving_min = distance - 0.95*(box_length/2)

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time before Weyl loop =",current_time)

            ### Test Weyl ####
            WeylTensor1010 = np.zeros((N_pixels,N_pixels), dtype=float)
            WeylTensor1020 = np.zeros((N_pixels,N_pixels), dtype=float)
            diag1010 = np.zeros((N_pixels,N_pixels), dtype=float)
            diag1020 = np.zeros((N_pixels,N_pixels), dtype=float)
            signWeyl1010 = np.zeros((N_pixels,N_pixels), dtype=float)
            signWeyl1020 = np.zeros((N_pixels,N_pixels), dtype=float)
            diag1010_median = np.zeros((N_pixels,N_pixels), dtype=float)
            poisson_test = np.zeros((N_pixels,N_pixels), dtype=float)

            pool = Pool(NCPU)
            xz_grid = []
            for alpha in x_analyse_grid:
                for beta in z_analyse_grid:
                    xz_grid.append([alpha, beta])
            res = pool.starmap_async(test_weyl_tensor_cart, xz_grid)
            pool.close()
            pool.join()

            Weyl_res = np.array(res.get())

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time after Weyl loop =",current_time)


            which_pixel_x = 0
            which_pixel_z = 0

            for i in range(0,len(Weyl_res)):
                WeylTensor1010[which_pixel_x,which_pixel_z] = Weyl_res[i,0]
                WeylTensor1020[which_pixel_x,which_pixel_z] = Weyl_res[i,1]
                diag1010[which_pixel_x,which_pixel_z] = Weyl_res[i,2]
                diag1020[which_pixel_x,which_pixel_z] = Weyl_res[i,3]
                poisson_test[which_pixel_x,which_pixel_z] = Weyl_res[i,4]
                which_pixel_z += 1
                if(which_pixel_z == N_pixels):
                    which_pixel_z = 0
                    which_pixel_x += 1

            i=0
            for beta in x_analyse_grid:
                l = 0
                for alpha in z_analyse_grid:
                    if(WeylTensor1010[i,l] > 0):
                        signWeyl1010[i,l] = 1
                    else:
                        signWeyl1010[i,l] = -1
                    if(WeylTensor1020[i,l] > 0):
                        signWeyl1020[i,l] = 1
                    else:
                        signWeyl1020[i,l] = -1
                    l += 1
                i += 1


            i = 0
            for beta in x_analyse_grid:
                l = 0
                for alpha in z_analyse_grid:
                    diag1010_median[i,l]=diag1010[i,l]
                    if(i != (N_pixels-1) and l != (N_pixels-1)):
                        neighbours = [diag1010[i,l], diag1010[i+1,l], diag1010[i-1,l], diag1010[i,l+1], diag1010[i,l-1]]
                        pixel_median = np.median(neighbours)
                    if(abs(diag1010[i,l]) > 2*abs(pixel_median)):
                        diag1010_median[i,l] = pixel_median
                    l += 1
                i += 1
            i=0
            for beta in x_analyse_grid:
                l = 0
                for alpha in z_analyse_grid:
                    if(i != (N_pixels-1) and l != (N_pixels-1)):
                        neighbours = [diag1010_median[i,l], diag1010_median[i+1,l], diag1010_median[i-1,l], diag1010_median[i,l+1], diag1010_median[i,l-1]]
                        pixel_median = np.median(neighbours)
                    if(abs(diag1010_median[i,l]) > 2*abs(pixel_median)):
                        diag1010_median[i,l] = pixel_median
                    l += 1
                i += 1

            np.save(("WeylTensor1010_%i_%i" % (N_stack, void_count)), WeylTensor1010)
            np.save(("WeylTensor1020_%i_%i" % (N_stack, void_count)), WeylTensor1020)

            if(stacking_step==False):
                figs.plot_test_weyl_cart(x_analyse_grid, z_analyse_grid, WeylTensor1010, WeylTensor1020, signWeyl1010, signWeyl1020, diag1010, diag1010_median, poisson_test)

            #lens_calc.z_deriv_plot(potsInter, distance, 0, 0, void_position_from_observer)


        ### Test Weyl ####


        ### Velocities ####

        if (do_velocity_centre==True):

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time before velocity loop =",current_time)

            pixel_radial_vel = np.zeros((len(offsets),N_pixels,N_pixels), dtype=float)
            pixel_vel_magnitude = np.zeros((len(offsets),N_pixels,N_pixels), dtype=float)
            pixel_vel_los = np.zeros((len(offsets),N_pixels,N_pixels), dtype=float)
            pixel_vel_y = np.zeros((len(offsets),N_pixels,N_pixels), dtype=float)

            #for i in range(0,len(offsets)):

            xz_grid = []
            xz_grid_average = []
            for alpha in range(0,len(x_analyse_grid)):
                for beta in range(0,len(z_analyse_grid)):
                    xz_grid.append([alpha, beta, x_analyse_grid, z_analyse_grid])
                    xz_grid_average.append([x_analyse_grid[alpha], x_analyse_grid[beta]])
            pool = Pool(NCPU)
            res = pool.starmap_async(velocity_centre_cart, xz_grid)
            pool.close()
            pool.join()

            vel_res = np.array(res.get())

            which_pixel_x = 0
            which_pixel_z = 0

            write_out_file=True
            filename=("vels.dat")
            if(write_out_file):
                F = open(filename+".dat","w+")
            for i in range(0,len(vel_res)):
                pixel_radial_vel[:,which_pixel_x,which_pixel_z] = vel_res[i,0]
                pixel_vel_magnitude[:,which_pixel_x,which_pixel_z] = vel_res[i,1]
                pixel_vel_los[:,which_pixel_x,which_pixel_z] = vel_res[i,2]
                pixel_vel_y[:,which_pixel_x,which_pixel_z] = vel_res[i,3]
                if(write_out_file):
                    F.write("%f %f %.3e\n" % (x_analyse_grid[which_pixel_x], z_analyse_grid[which_pixel_z], pixel_vel_los[2,which_pixel_x,which_pixel_z]))
                which_pixel_z += 1
                if(which_pixel_z == N_pixels):
                    which_pixel_z = 0
                    which_pixel_x += 1
            if(write_out_file):
                F.close()

            np.save(("pixel_radial_vel_%i_%i" % (N_stack, void_count)), pixel_radial_vel)
            np.save(("pixel_vel_magnitude_%i_%i" % (N_stack, void_count)), pixel_vel_magnitude)
            np.save(("pixel_vel_los_%i_%i" % (N_stack, void_count)), pixel_vel_los)
            np.save(("pixel_vel_y_%i_%i" % (N_stack, void_count)), pixel_vel_y)

            if(stacking_step==False):
                los_vel_inter=LinearNDInterpolator(xz_grid_average, vel_res[:,2,2], fill_value=np.mean(vel_res[:,2,2]))
                radial_vel_inter=LinearNDInterpolator(xz_grid_average, vel_res[:,0,2], fill_value=np.mean(vel_res[:,0,2]))
                mean_los_vel=np.mean(np.abs(vel_res[:,2,2]))
                mean_radial_vel=np.mean(np.abs(vel_res[:,2,2]))
                centre_los=lens_calc.los_velocity_centre(los_vel_inter, x_analyse_grid, z_analyse_grid, mean_los_vel, box_length)
                centre_radial=lens_calc.radial_velocity_centre(radial_vel_inter, x_analyse_grid, z_analyse_grid, mean_radial_vel, box_length)
                centre=[centre_los, centre_radial]
                #centre=[[0,0],[0,0]]
                figs.plot_vels_cart(x_analyse_grid, z_analyse_grid, pixel_radial_vel[2,:,:], pixel_vel_magnitude[2,:,:], pixel_vel_los[2,:,:], pixel_vel_y[2,:,:], centre, void_count)

                deviation_to_dens_centre_los = calc.calc_centre_deviation([0,0], centre[0])
                deviation_to_dens_centre_radial = calc.calc_centre_deviation([0,0], centre[1])

                vel_out_file=open(("velocities_%i.dat") % void_count,"w+")
                vel_out_file.write("centre_x_los=%f\ncentre_z_los=%f\n" % (centre[0][0], centre[0][1]) )
                vel_out_file.write("centre_x_radial=%f\ncentre_z_radial=%f\n" % (centre[1][0], centre[1][1]) )
                vel_out_file.write("deviation_los=%f\n" % deviation_to_dens_centre_los )
                vel_out_file.write("deviation_radial=%f\n" % deviation_to_dens_centre_radial )
                vel_out_file.close()

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time after velocity loop =",current_time)

        ### Velocities ####


        ### Mass deficit ###

        if (do_mass_def==True):

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time before mass deficit loop =",current_time)

            mass_def = np.zeros((N_pixels,N_pixels), dtype=float)

            xz_grid = []
            xz_grid_average = []
            for alpha in range(0,len(x_analyse_grid)):
                for beta in range(0,len(z_analyse_grid)):
                    xz_grid.append([alpha, beta])
                    xz_grid_average.append([x_analyse_grid[alpha], z_analyse_grid[beta]])

            pool = Pool(NCPU)

            res = pool.starmap_async(mass_deficit_cart, xz_grid)
            pool.close()
            pool.join()

            mass_deficit_res = np.array(res.get())

            which_pixel_x = 0
            which_pixel_z = 0

            write_out_file=True
            if(write_out_file):
                F = open(("mass_def_results_%i.dat") % void_count,"w+")
            for i in range(0,len(mass_deficit_res)):
                mass_def[which_pixel_x,which_pixel_z] = mass_deficit_res[i]
                if(write_out_file):
                    F.write("%f %f %.3e\n" % (x_analyse_grid[which_pixel_x], z_analyse_grid[which_pixel_z], mass_def[which_pixel_x,which_pixel_z]))
                which_pixel_z += 1
                if(which_pixel_z == N_pixels):
                    which_pixel_z = 0
                    which_pixel_x += 1
            if(write_out_file):
                F.close()

            #print("Max mass_def: ", np.max(mass_def))
            k=0
            mass_def_inter_array=np.zeros((N_pixels*N_pixels))

            mass_def_top_percentile = np.percentile(mass_def,cosmic_ray_percentile)
            pixels_too_high = np.where(mass_def >= mass_def_top_percentile)
            if( np.array(pixels_too_high).size > 0):
                mass_def[pixels_too_high] = mass_def_top_percentile

            for i in range(0,len(x_analyse_grid)):
                for l in range(0,len(z_analyse_grid)):
                    if(i == (N_pixels-1) and l != (N_pixels-1)):
                        neighbours = [mass_def[i,l], mass_def[0,l], mass_def[i-1,l], mass_def[i,l+1], mass_def[i,l-1]]
                        pixel_median = np.median(neighbours)
                        pixel_mean = np.mean(neighbours)
                    elif(i != (N_pixels-1) and l == (N_pixels-1)):
                        neighbours = [mass_def[i,l], mass_def[i+1,l], mass_def[i-1,l], mass_def[i,0], mass_def[i,l-1]]
                        pixel_median = np.median(neighbours)
                        pixel_mean = np.mean(neighbours)
                    elif(i == (N_pixels-1) and l == (N_pixels-1)):
                        neighbours = [mass_def[i,l], mass_def[0,l], mass_def[i-1,l], mass_def[i,0], mass_def[i,l-1]]
                        pixel_median = np.median(neighbours)
                        pixel_mean = np.mean(neighbours)
                    elif(i != (N_pixels-1) and l != (N_pixels-1)):
                        neighbours = [mass_def[i,l], mass_def[i+1,l], mass_def[i-1,l], mass_def[i,l+1], mass_def[i,l-1]]
                        pixel_median = np.median(neighbours)
                        pixel_mean = np.mean(neighbours)
                    else:
                        print("Error: lensing.py, do_mass_def section: Programming error.\n")
                    #if(abs(mass_def[i,l]) > 2*abs(pixel_median)):

                    mass_def_inter_array[k]=mass_def[i,l]
                    k=k+1

            # Normalise our data to values between 0 and 1
            #most_negative_mass_def = np.amin(mass_def)
            #norm_mass_def = mass_def - most_negative_mass_def
            #maximal_mass_def = np.amax(norm_mass_def)
            #norm_mass_def = norm_mass_def/maximal_mass_def

            # TODO: Split the weak lensing calculations into an own routine!
            mass_def_inter=LinearNDInterpolator(xz_grid_average, mass_def_inter_array, fill_value=np.mean(mass_def_inter_array))
            median_mass_def=np.median(mass_deficit_res)

            #centre, centre_radii, average_mass_def_dens, average_mass_def, average_mass_def_gamma_centre, average_mass_def_combined, average_mass_def_lensing_dens, average_mass_def_lensing, average_mass_def_gamma_centre_lensing, average_mass_def_combined_lensing, kappa_grid, gamma_grid=lens_calc.mass_def_centre(mass_def_inter, x_analyse_grid, z_analyse_grid, rads, mean_mass_def, distance, mass_dm, OmegaM, OmegaLambda, Hubble, mass_def, N_pixels, box_length, void_radius, void_count)
            centre, centre_radii, centre_gamma, centre_radii_gamma = lens_calc.mass_def_centre(mass_def_inter, x_analyse_grid, z_analyse_grid, rads, median_mass_def, distance, mass_dm, OmegaM, OmegaLambda, Hubble, mass_def, N_pixels, box_length, void_stddev_frac_crit, void_diff_frac_crit, N_average, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean, merge_distance, void_count)


            np.save(("mass_def_%i_%i" % (N_stack, void_count)), mass_def)
            np.save(("dens_%i_%i" % (N_stack, void_count)), dens)
            np.save(("pixel_pot_%i_%i" % (N_stack, void_count)), pixel_pot)

            #np.save(("kappa_grid_%i_%i" % (N_stack, void_count)), kappa_grid)
            #np.save(("gamma_grid_%i_%i" % (N_stack, void_count)), gamma_grid)

            #np.save(("average_mass_def_lensing_dens_%i_%i" % (N_stack, void_count)),average_mass_def_lensing_dens)
            #np.save(("average_mass_def_gamma_centre_lensing_%i_%i" % (N_stack, void_count)),average_mass_def_gamma_centre_lensing)
            #np.save(("average_mass_def_lensing_%i_%i" % (N_stack, void_count)),average_mass_def_lensing)
            #np.save(("average_mass_def_combined_lensing_%i_%i" % (N_stack, void_count)),average_mass_def_combined_lensing)

            #np.save(("average_mass_def_dens_%i_%i" % (N_stack, void_count)),average_mass_def_dens)
            #np.save(("average_mass_def_gamma_centre_%i_%i" % (N_stack, void_count)),average_mass_def_gamma_centre)
            #np.save(("average_mass_def_%i_%i" % (N_stack, void_count)),average_mass_def)
            #np.save(("average_mass_def_combined_%i_%i" % (N_stack, void_count)),average_mass_def_combined)


            #if(stacking_step==False):

            #minima=[centre[0][0][0], centre[0][1][0]]
            #deviation_to_dens_centre = calc.calc_centre_deviation([0,0], [centre[0][0][0], centre[0][1][0]])
            #deviation_to_dens_centre_gamma = calc.calc_centre_deviation([0,0], [centre[1][0][0], centre[1][1][0]])
            #deviation_to_dens_centre_combined = calc.calc_centre_deviation([0,0], [centre[2][0][0], centre[2][1][0]])

            average_mass_def_revolver=np.zeros((len(eff_void_radius_gals), 101))
            average_mass_def_revolver_cut=np.zeros((len(void_ids), 101))
            average_mass_def=np.zeros((len(centre_radii), 101))
            average_gamma_revolver=np.zeros((len(eff_void_radius_gals), 101))
            average_gamma_revolver_cut=np.zeros((len(void_ids), 101))
            average_gamma=np.zeros((len(centre_radii_gamma), 101))
            matched_voids_revolver=0
            probs_mass_def=[]
            radii_mass_def=[] # revolver radius of the matched void
            probs_gamma=[]
            radii_gamma=[] # revolver radius of the matched void
            k=0

            f_matches_mass_def = open(void_matches_file_base + '_mass_def_given_3D.dat',"w") # given a 3D void, what is the best 2D void?
            f_matches_gamma = open(void_matches_file_base + '_gamma_given_3D.dat',"w") # given a 3D void, what is the best 2D void?
            f_matches_mass_def.write("# Radii and positions of 3D voids and each best-matched 2D void for the mass deficit Sigma\n")
            f_matches_mass_def.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_mass_def.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_mass_def.write("# Column  7: heuristic probability of the 2D void being a detection of the 3D void\n")
            f_matches_gamma.write("# Radii and positions of 3D voids and each best-matched 2D void for the lensing paramer gamma\n")
            f_matches_gamma.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_gamma.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_gamma.write("# Column  7: heuristic probability of the 2D void being a detection of the 3D void\n")

            i_2D_best_mass_def = -1*np.ones(len(eff_void_radius_gals),dtype=int)
            i_2D_best_gamma = -1*np.ones(len(eff_void_radius_gals),dtype=int)
            for this_void_id in range(0,len(eff_void_radius_gals)):
                prob, i_2D_best_mass_def[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2], std_R_void_prob, std_xz_void_prob, centre_radii, centre[:,0], centre[:,1], N_pixels)
                prob_gamma, i_2D_best_gamma[this_void_id]  = lens_calc_ut.void_probability(eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2], std_R_void_prob, std_xz_void_prob, centre_radii_gamma, centre_gamma[:,0], centre_gamma[:,1], N_pixels)
                #best_matched_void=np.where(prob==np.max(prob))[0][0]
                best_matched_void = i_2D_best_mass_def[this_void_id]
                best_matched_void_id=best_matched_void
                #best_matched_void_gamma=np.where(prob_gamma==np.max(prob_gamma))[0][0]
                best_matched_void_gamma = i_2D_best_gamma[this_void_id]
                probs_mass_def.append(prob[best_matched_void])
                radii_mass_def.append(eff_void_radius_gals[this_void_id])
                probs_gamma.append(prob_gamma[best_matched_void_gamma])
                radii_gamma.append(eff_void_radius_gals[this_void_id])
                if(prob[best_matched_void] >= prob_threshold):
                    matched_voids_revolver=matched_voids_revolver+1
                    best_void_index[0,this_void_id]=best_matched_void_id
                    lensing_void_radii[0,this_void_id]=centre_radii[int(best_matched_void_id)]
                norm_rads=np.linspace(0.1, 1.5*eff_void_radius_gals[this_void_id],101)
                average_mass_def_revolver[this_void_id] = lens_calc.average_mass_deficit([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, mass_def_inter)

                kappa, gamma, kappa_grid, gamma_grid, Sigma_bar, Sigma_crit = lens_calc.mass_def_lensing(mass_def, [moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], distance, mass_dm, OmegaM, OmegaLambda, Hubble, N_pixels, x_analyse_grid, z_analyse_grid, rads, box_length)

                f_matches_mass_def.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2],
                    centre_radii[best_matched_void], centre[best_matched_void,0], centre[best_matched_void,1],
                    prob[best_matched_void]))
                f_matches_gamma.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2],
                    centre_radii_gamma[best_matched_void_gamma], centre_gamma[best_matched_void_gamma,0], centre_gamma[best_matched_void_gamma,1],
                    prob_gamma[best_matched_void_gamma]))

                xz_grid=[]
                mass_deficit_lensing=np.zeros((N_pixels*N_pixels))
                j=0
                for l in range(0,len(x_analyse_grid)):
                    for m in range(0,len(z_analyse_grid)):
                        xz_grid.append([x_analyse_grid[l], z_analyse_grid[m]])
                        mass_deficit_lensing[j]=gamma_grid[l,m]
                        j=j+1

                mass_def_lensing_inter=LinearNDInterpolator(xz_grid, mass_deficit_lensing)
                average_gamma_revolver[this_void_id]=lens_calc.average_mass_def_lensing([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, mass_def_lensing_inter)
                # Ignore this - rough hack to with hardwired cutoff size of 20.0 Mpc ...
                if(eff_void_radius_gals[this_void_id] > 20):
                    average_mass_def_revolver_cut[k] = lens_calc.average_mass_deficit([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, mass_def_inter)
                    average_gamma_revolver_cut[k] = lens_calc.average_mass_def_lensing([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, mass_def_lensing_inter)
                    k=k+1

            f_matches_mass_def.close()
            f_matches_gamma.close()


            # Use the same file pointers, but for different files:
            f_matches_mass_def = open(void_matches_file_base + '_mass_def_given_2D.dat',"w") # given a 2D void, what is the best 3D void?
            f_matches_gamma = open(void_matches_file_base + '_gamma_given_2D.dat',"w") # given a 2D void, what is the best 3D void?
            f_matches_mass_def.write("# Radii and positions of 2D voids and each best-matched 3D void for the mass deficit Sigma\n")
            f_matches_mass_def.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_mass_def.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_mass_def.write("# Column  7: heuristic probability of the 3D void being a detection of the 2D void\n")
            f_matches_gamma.write("# Radii and positions of 2D voids and each best-matched 3D void for the lensing paramer gamma\n")
            f_matches_gamma.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_gamma.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_gamma.write("# Column  7: heuristic probability of the 3D void being a detection of the 2D void\n")

            probs_mass_def_projected_void=[]
            radii_mass_def_projected_void=[]
            matched_voids_mass_def=0
            i_3D_best_mass_def = -1*np.ones(len(centre_radii),dtype=int)
            for this_void_id in range(0,len(centre_radii)):
                prob, i_3D_best_mass_def[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals, moved_revolver_centres[:,0], moved_revolver_centres[:,2], std_R_void_prob, std_xz_void_prob, centre_radii[this_void_id], centre[this_void_id,0], centre[this_void_id,1], N_pixels) # returns a prob for all revolver voids
                #best_matched_void=np.where(prob==np.max(prob))[0][0]
                best_matched_void = i_3D_best_mass_def[this_void_id]
                best_matched_void_id=best_matched_void
                probs_mass_def_projected_void.append(prob[best_matched_void])
                radii_mass_def_projected_void.append(centre_radii[this_void_id])
                if(prob[best_matched_void] >= prob_threshold):
                    matched_voids_mass_def=matched_voids_mass_def+1
                    #best_void_index[0,this_void_id]=best_matched_void_id
                    #lensing_void_radii[0,this_void_id]=centre_radii[int(best_matched_void_id)]

                # Evaluate a range of radii normalised according to
                # centre_radii[this_void_id], the radius of this particular 2D void
                norm_rads=np.linspace(0.1, 1.5*centre_radii[this_void_id],101)
                average_mass_def[this_void_id] = lens_calc.average_mass_deficit([centre[this_void_id,0], centre[this_void_id,1]], norm_rads, box_length, mass_def_inter)

                f_matches_mass_def.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[best_matched_void], moved_revolver_centres[best_matched_void,0], moved_revolver_centres[best_matched_void,2],
                    centre_radii[this_void_id], centre[this_void_id,0], centre[this_void_id,1],
                    prob[this_void_id]))

            f_matches_mass_def.close()


            probs_gamma_projected_void=[]
            radii_gamma_projected_void=[]
            matched_voids_gamma=0
            i_3D_best_gamma = -1*np.ones(len(centre_radii_gamma),dtype=int)

            for this_void_id in range(0,len(centre_radii_gamma)):
                kappa, gamma, kappa_grid, gamma_grid, Sigma_bar, Sigma_crit = lens_calc.mass_def_lensing(mass_def, centre_gamma[this_void_id], distance, mass_dm, OmegaM, OmegaLambda, Hubble, N_pixels, x_analyse_grid, z_analyse_grid, rads, box_length)

                xz_grid=[]
                mass_deficit_lensing=np.zeros((N_pixels*N_pixels))
                k=0
                for l in range(0,len(x_analyse_grid)):
                    for m in range(0,len(z_analyse_grid)):
                        xz_grid.append([x_analyse_grid[l], z_analyse_grid[m]])
                        mass_deficit_lensing[k]=gamma_grid[l,m]
                        k=k+1

                mass_def_lensing_inter=LinearNDInterpolator(xz_grid, mass_deficit_lensing)

                prob, i_3D_best_gamma[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals, moved_revolver_centres[:,0], moved_revolver_centres[:,2], std_R_void_prob, std_xz_void_prob, centre_radii_gamma[this_void_id], centre_gamma[this_void_id,0], centre_gamma[this_void_id,1], N_pixels)
                #best_matched_void=np.where(prob==np.max(prob))[0][0]
                best_matched_void = i_3D_best_gamma[this_void_id]
                best_matched_void_id=best_matched_void
                probs_gamma_projected_void.append(prob[best_matched_void])
                radii_gamma_projected_void.append(centre_radii_gamma[this_void_id])
                if(prob[best_matched_void] >= prob_threshold):
                    matched_voids_gamma=matched_voids_gamma+1
                    #best_void_index[0,this_void_id]=best_matched_void_id
                    #lensing_void_radii[0,this_void_id]=centre_radii[int(best_matched_void_id)]
                norm_rads=np.linspace(0.1, 1.5*centre_radii_gamma[this_void_id],101)
                average_gamma[this_void_id]=lens_calc.average_mass_def_lensing(centre_gamma[this_void_id], norm_rads, box_length, mass_def_lensing_inter)

                f_matches_gamma.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[best_matched_void], moved_revolver_centres[best_matched_void,0], moved_revolver_centres[best_matched_void,2],
                    centre_radii_gamma[this_void_id], centre_gamma[this_void_id,0], centre_gamma[this_void_id,1],
                    prob[this_void_id]))

            f_matches_gamma.close()


            np.save("average_gamma", average_gamma)

            mass_def_out_file=open(("mass_def_%i.dat") % void_count,"w+")
            mass_def_out_file.write("N_average_pixel=%i\n" % (N_average*N_average))
            mass_def_out_file.write("N_void=%i\n" % (len(centre_radii)))
            mass_def_out_file.write("N_void_gamma=%i\n" % (len(centre_radii_gamma)))
            mass_def_out_file.write("N_void_revolver=%i\n" % (len(eff_void_radius_gals)))
            mass_def_out_file.write("N_void_revolver_matched=%i\n" % matched_voids_revolver)
            mass_def_out_file.write("matched_revolver_fraction=%f\n" % (matched_voids_revolver/len(eff_void_radius_gals)))
            mass_def_out_file.write("N_void_mass_def_matched=%i\n" % matched_voids_mass_def)
            mass_def_out_file.write("N_void_matched_gamma=%i\n" % matched_voids_gamma)
            mass_def_out_file.write("matched_mass_def_fraction=%f\n" % (matched_voids_mass_def/len(centre_radii)))
            mass_def_out_file.write("matched_fraction_gamma=%f\n" % (matched_voids_gamma/len(centre_radii_gamma)))
            mass_def_out_file.write("Revolver_id=%i\nour_id=%i\n" % (void_id_gals, best_matched_void_id))
            mass_def_out_file.write("radius_revolver=%f\n" % centre_radii[0] )
            mass_def_out_file.write("centre_x=%f\ncentre_z=%f\nradius=%f\n" % (centre[0,0], centre[0,1], centre_radii[1]) )
            #mass_def_out_file.write("centre_x_gamma=%f\ncentre_z_gamma=%f\nradius_gamma=%f\n" % (centre[1][0][0], centre[1][1][0], centre_radii[2]) )
            #mass_def_out_file.write("centre_x_combined=%f\ncentre_z_combined=%f\n" % (centre[2][0][0], centre[2][1][0]) )
            #mass_def_out_file.write("deviation=%f\n" % deviation_to_dens_centre )
            #mass_def_out_file.write("deviation_gamma=%f\n" % deviation_to_dens_centre_gamma )
            #mass_def_out_file.write("deviation_combined=%f\n" % deviation_to_dens_centre_combined )
            mass_def_out_file.close()

            figs.plot_density_pot_slice(x_analyse_grid, z_analyse_grid, dens, pixel_pot, moved_gal_pos, void_gal_ids, void_count)
            figs.plot_mass_def_cart(x_analyse_grid, z_analyse_grid, mass_def, centre, rads, average_mass_def_revolver,
                                    average_mass_def_revolver_cut, average_mass_def, average_mass_def, centre_radii,
                                    moved_revolver_centres, best_matched_void_id, box_length, void_count)
            figs.plot_rads_vs_prob(radii_mass_def, probs_mass_def, "mass_def")
            figs.plot_rads_vs_prob(radii_gamma, probs_gamma, "gamma")
            figs.plot_rads_vs_prob(radii_mass_def_projected_void, probs_mass_def_projected_void, "mass_def_projected_void")
            figs.plot_rads_vs_prob(radii_gamma_projected_void, probs_gamma_projected_void, "gamma_projected_void")
            figs.plot_weak_lensing_cart(x_analyse_grid, z_analyse_grid, kappa_grid, gamma_grid, centre_gamma, centre_radii_gamma, moved_revolver_centres, average_gamma, average_gamma_revolver, average_gamma_revolver_cut, box_length, void_count)
            figs.plot_void_population(eff_void_radius_gals, centre_radii, "mass_def")
            figs.plot_void_population(eff_void_radius_gals, centre_radii_gamma, "gamma")

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time after mass deficit loop =",current_time)

        ### Mass deficit ###


        ### ISW effect ###

        if(do_ISW==True):

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time before ISW loop =",current_time)


            ISW = np.zeros((N_pixels,N_pixels), dtype=float)

            xz_grid = []
            for alpha in range(0,len(x_analyse_grid)):
                for beta in range(0,len(z_analyse_grid)):
                    xz_grid.append([alpha, beta])

            pool = Pool(NCPU)

            res = pool.starmap_async(ISW_cart, xz_grid)
            pool.close()
            pool.join()

            ISW_res = np.array(res.get())

            which_pixel_x = 0
            which_pixel_z = 0

            for i in range(0,len(ISW_res)):
                ISW[which_pixel_x,which_pixel_z] = ISW_res[i]
                which_pixel_z += 1
                if(which_pixel_z == N_pixels):
                    which_pixel_z = 0
                    which_pixel_x += 1

            np.save(("ISW_%i_%i" % (N_stack, void_count)), ISW)

            if(stacking_step==False):
                centre, minima = calc.find_minimum(ISW, x_analyse_grid, z_analyse_grid)
                deviation_to_dens_centre = calc.calc_centre_deviation([0,0], centre)

                ISW_out_file=open(("ISW_%i.dat") % void_count,"w+")
                ISW_out_file.write("centre_x=%f\ncentre_z=%f\n" % (centre[0], centre[1]) )
                ISW_out_file.write("deviation=%f\n" % deviation_to_dens_centre )
                ISW_out_file.close()

                figs.plot_ISW_cart(x_analyse_grid, z_analyse_grid, ISW, pot_time_slope[20,:,:], centre, minima, void_count)

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time after ISW loop =",current_time)

        ### ISW effect ###


        ### Doppler lensing ###

        if(do_doppler_lensing ==True):

            # TODO: This routine is outdated
            xz_grid = []
            for alpha in x_analyse_grid:
                for beta in z_analyse_grid:
                    xz_grid.append([alpha, beta])

            pool = Pool(NCPU)

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time before doppler loop =",current_time)

            res = pool.starmap_async(doppler_convergence_cart, xz_grid)
            pool.close()
            pool.join()

            doppler_convergence_res = np.array(res.get())

            doppler_convergence = np.zeros((N_pixels,N_pixels,len(doppler_convergence_res[0])), dtype=float)

            which_pixel_x = 0
            which_pixel_z = 0

            for i in range(0,len(doppler_convergence_res[:,0])):
                doppler_convergence[which_pixel_x,which_pixel_z,0] = doppler_convergence_res[i,0]
                doppler_convergence[which_pixel_x,which_pixel_z,1] = doppler_convergence_res[i,1]
                doppler_convergence[which_pixel_x,which_pixel_z,2] = doppler_convergence_res[i,2]
                doppler_convergence[which_pixel_x,which_pixel_z,3] = doppler_convergence_res[i,3]
                doppler_convergence[which_pixel_x,which_pixel_z,4] = doppler_convergence_res[i,4]
                which_pixel_z += 1
                if(which_pixel_z == N_pixels):
                    which_pixel_z = 0
                    which_pixel_x += 1

            np.save(("doppler_convergence_%i" % N_stack), doppler_convergence)

            if(stacking_step==False):
                figs.plot_doppler_cart(x_analyse_grid, z_analyse_grid, doppler_convergence)

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time after doppler loop =",current_time)

        ### Doppler lensing ###

        ### Solve equation optical scalars ###

        if(do_solve_scalars==True):

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time before diff solver loop =",current_time)

            pool = Pool(NCPU)
            xz_grid = []
            xz_grid_average = []
            for alpha in range(0,len(x_analyse_grid)):
                for beta in range(0,len(z_analyse_grid)):
                    xz_grid.append([alpha, beta])
                    xz_grid_average.append([x_analyse_grid[alpha], z_analyse_grid[beta]])

            res = pool.starmap_async(solve_optical_scalars_cart, xz_grid)
            pool.close()
            pool.join()

            optical_scalar_res = np.array(res.get())

            real_theta=np.zeros((N_pixels,N_pixels), dtype=float)
            real_sigma=np.zeros((N_pixels,N_pixels), dtype=float)
            img_sigma=np.zeros((N_pixels,N_pixels), dtype=float)
            distance_check=np.zeros((N_pixels,N_pixels), dtype=float)

            which_pixel_x = 0
            which_pixel_z = 0

            mean_theta = np.mean(optical_scalar_res[:,0])
            rel_deviation = np.zeros((N_pixels,N_pixels), dtype=float)

            for i in range(0,len(optical_scalar_res)):
                real_theta[which_pixel_x,which_pixel_z] = optical_scalar_res[i,0]
                rel_deviation[which_pixel_x,which_pixel_z] = (mean_theta-optical_scalar_res[i,0])/mean_theta
                real_sigma[which_pixel_x,which_pixel_z] = optical_scalar_res[i,1]
                img_sigma[which_pixel_x,which_pixel_z] = optical_scalar_res[i,2]
                #distance_check[which_pixel_x,which_pixel_z] = optical_scalar_res[i,3]
                which_pixel_z += 1
                if(which_pixel_z == N_pixels):
                    which_pixel_z = 0
                    which_pixel_x += 1

            mag_sigma = np.sqrt((real_sigma**2)+(img_sigma**2))
            k=0
            expansion_inter_array=np.zeros((N_pixels*N_pixels))
            shear_inter_array=np.zeros((N_pixels*N_pixels))
            for i in range(0,len(x_analyse_grid)):
                for l in range(0,len(z_analyse_grid)):
                    if(i != (N_pixels-1) and l != (N_pixels-1)):
                        neighbours = [real_theta[i,l], real_theta[i+1,l], real_theta[i-1,l], real_theta[i,l+1], real_theta[i,l-1]]
                        neighbours_real_sigma = [real_sigma[i,l], real_sigma[i+1,l], real_sigma[i-1,l], real_sigma[i,l+1], real_sigma[i,l-1]]
                        neighbours_img_sigma = [img_sigma[i,l], img_sigma[i+1,l], img_sigma[i-1,l], img_sigma[i,l+1], img_sigma[i,l-1]]
                        pixel_median = np.median(neighbours)
                        pixel_median_real_sigma = np.median(neighbours_real_sigma)
                        pixel_median_img_sigma = np.median(neighbours_img_sigma)
                        pixel_mean = np.mean(neighbours)
                        pixel_mean_real_sigma = np.mean(neighbours_real_sigma)
                        pixel_mean_img_sigma = np.mean(neighbours_img_sigma)
                    if(abs(real_theta[i,l]) > 2*abs(pixel_mean)):
                        if(abs(pixel_mean) > 2*abs(pixel_median)):
                            # TODO: probably a faulty integration
                            real_theta[i,l] = pixel_median
                        else:
                            real_theta[i,l] = pixel_mean
                    if(abs(real_sigma[i,l]) > 2*abs(pixel_mean_real_sigma)):
                        if(abs(pixel_mean_real_sigma) > 2*abs(pixel_median_real_sigma)):
                            real_sigma[i,l] = pixel_median_real_sigma
                        else:
                            real_sigma[i,l] = pixel_mean_real_sigma
                    if(abs(img_sigma[i,l]) > 2*abs(pixel_mean_img_sigma)):
                        if(abs(pixel_mean_img_sigma) > 2*abs(pixel_median_img_sigma)):
                            img_sigma[i,l] = pixel_median_img_sigma
                        else:
                            img_sigma[i,l] = pixel_mean_img_sigma
                    expansion_inter_array[k]=real_theta[i,l]
                    shear_inter_array[k]=mag_sigma[i,l]
                    k=k+1

            np.save(("real_theta_%i_%i" % (N_stack, void_count)), real_theta)
            np.save(("real_sigma_%i_%i" % (N_stack, void_count)), real_sigma)
            np.save(("img_sigma_%i_%i" % (N_stack, void_count)), img_sigma)
            np.save(("mag_sigma_%i_%i" % (N_stack, void_count)), mag_sigma)

            #if(stacking_step==False):
            expansion_inter=LinearNDInterpolator(xz_grid_average, expansion_inter_array, fill_value=np.mean(expansion_inter_array))
            shear_inter=LinearNDInterpolator(xz_grid_average, shear_inter_array, fill_value=np.mean(shear_inter_array))
            median_expansion=np.median(real_theta)
            median_shear=np.median(mag_sigma)
            centre_exp, radius_exp=lens_calc.expansion_centre(expansion_inter, x_analyse_grid, z_analyse_grid, rads, median_expansion, box_length, N_pixels, void_stddev_frac_crit_theta, void_diff_frac_crit_theta, N_average, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean, merge_distance, void_count)
            centre_sigma, radius_sigma=lens_calc.sigma_centre(shear_inter, x_analyse_grid, z_analyse_grid, rads, median_shear, box_length, N_pixels, void_stddev_frac_crit_sigma, void_diff_frac_crit_sigma, N_average, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean, merge_distance, void_count)


            average_expansion_revolver=np.zeros((len(eff_void_radius_gals), 101))
            average_sigma_revolver=np.zeros((len(eff_void_radius_gals), 101))
            average_expansion_revolver_cut=np.zeros((len(void_ids), 101))
            average_sigma_revolver_cut=np.zeros((len(void_ids), 101))
            average_expansion=np.zeros((len(radius_exp), 101))
            average_sigma=np.zeros((len(radius_sigma), 101))

            matched_voids_revolver_exp=0
            matched_voids_revolver_sig=0
            matched_probs_exp=[]
            matched_radii_exp=[] # revolver radius of the matched void
            matched_probs_sigma=[]
            matched_radii_sigma=[] # revolver radius of the matched void

            f_matches_exp = open(void_matches_file_base + '_exp_given_3D.dat',"w") # given a 3D void, what is the best 2D void?
            f_matches_sig = open(void_matches_file_base + '_sig_given_3D.dat',"w") # given a 3D void, what is the best 2D void?
            f_matches_exp.write("# Radii and positions of 3D voids and each best-matched 2D void for the expansion theta\n")
            f_matches_exp.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_exp.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_exp.write("# Column  7: heuristic probability of the 2D void being a detection of the 3D void\n")
            f_matches_sig.write("# Radii and positions of 3D voids and each best-matched 2D void for the lensing parameter sigma\n")
            f_matches_sig.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_sig.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_sig.write("# Column  7: heuristic probability of the 2D void being a detection of the 3D void\n")

            k=0
            i_2D_best_exp = -1*np.ones(len(eff_void_radius_gals),dtype=int)
            i_2D_best_sigma = -1*np.ones(len(eff_void_radius_gals),dtype=int)
            for this_void_id in range(0,len(eff_void_radius_gals)):
                prob_exp, i_2D_best_exp[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2], std_R_void_prob, std_xz_void_prob, radius_exp, centre_exp[:,0], centre_exp[:,1], N_pixels)
                prob_sig, i_2D_best_sigma[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2], std_R_void_prob, std_xz_void_prob, radius_sigma, centre_sigma[:,0], centre_sigma[:,1], N_pixels)
                print("prob_exp: ", prob_exp)
                #best_matched_void_exp=np.where(prob_exp==np.max(prob_exp))[0][0]
                best_matched_void_exp = i_2D_best_exp[this_void_id]
                best_matched_void_id_exp=best_matched_void_exp
                #best_matched_void_sig=np.where(prob_sig==np.max(prob_sig))[0][0]
                best_matched_void_sig = i_2D_best_sigma[this_void_id]
                best_matched_void_id_sig=best_matched_void_sig
                matched_probs_exp.append(prob_exp[best_matched_void_exp])
                matched_radii_exp.append(eff_void_radius_gals[this_void_id])
                matched_probs_sigma.append(prob_sig[best_matched_void_sig])
                matched_radii_sigma.append(eff_void_radius_gals[this_void_id])
                if(prob_exp[best_matched_void_exp] >= prob_threshold):
                    matched_voids_revolver_exp=matched_voids_revolver_exp+1
                    best_void_index[1,this_void_id]=best_matched_void_id_exp
                    lensing_void_radii[1,this_void_id]=radius_exp[int(best_matched_void_id_exp)]
                if(prob_sig[best_matched_void_sig] >= prob_threshold):
                    matched_voids_revolver_sig=matched_voids_revolver_sig+1
                    best_void_index[2,this_void_id]=best_matched_void_id_sig
                    lensing_void_radii[2,this_void_id]=radius_sigma[int(best_matched_void_id_sig)]
                norm_rads=np.linspace(0.1, 1.5*eff_void_radius_gals[this_void_id],101)
                average_expansion_revolver[this_void_id] = lens_calc.average_expansion([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, expansion_inter)
                average_sigma_revolver[this_void_id] = lens_calc.average_sigma([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, shear_inter)
                if(eff_void_radius_gals[this_void_id] > 20):
                    average_expansion_revolver_cut[k] = lens_calc.average_expansion([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, expansion_inter)
                    average_sigma_revolver_cut[k] = lens_calc.average_sigma([moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2]], norm_rads, box_length, shear_inter)
                    k=k+1

                f_matches_exp.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2],
                     radius_exp[best_matched_void_exp], centre_exp[best_matched_void_exp,0], centre_exp[best_matched_void_exp,1],
                     prob_exp[best_matched_void_exp]))
                f_matches_sig.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[this_void_id], moved_revolver_centres[this_void_id,0], moved_revolver_centres[this_void_id,2],
                     radius_sigma[best_matched_void_sig], centre_sigma[best_matched_void_sig,0], centre_sigma[best_matched_void_sig,1],
                     prob_sig[best_matched_void_sig]))

            f_matches_exp.close()
            f_matches_sig.close()


            # Re-use the file pointers for finding the best 3D void given a 2D void.
            f_matches_exp = open(void_matches_file_base + '_exp_given_2D.dat',"w") # given a 2D void, what is the best 3D void?
            f_matches_sig = open(void_matches_file_base + '_sig_given_2D.dat',"w") # given a 2D void, what is the best 3D void?
            f_matches_exp.write("# Radii and positions of 2D voids and each best-matched 3D void for the expansion theta\n")
            f_matches_exp.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_exp.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_exp.write("# Column  7: heuristic probability of the 3D void being a detection of the 2D void\n")
            f_matches_sig.write("# Radii and positions of 2D voids and each best-matched 3D void for the lensing parameter sigma\n")
            f_matches_sig.write("# Columns 1-3: radius, x, z of 3D void\n")
            f_matches_sig.write("# Columns 4-6: radius, x, z of best-matched 2D void\n")
            f_matches_sig.write("# Column  7: heuristic probability of the 3D void being a detection of the 2D void\n")

            matched_probs_exp_projected_void=[]
            matched_radii_exp_projected_void=[]
            matched_voids_exp=0
            i_3D_best_exp = -1*np.ones(len(radius_exp),dtype=int)

            for this_void_id in range(0,len(radius_exp)):
                prob, i_3D_best_exp[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals, moved_revolver_centres[:,0], moved_revolver_centres[:,2], std_R_void_prob, std_xz_void_prob, radius_exp[this_void_id], centre_exp[this_void_id,0], centre_exp[this_void_id,1], N_pixels)
                #best_matched_void=np.where(prob==np.max(prob))[0][0]
                best_matched_void = i_3D_best_exp[this_void_id]
                best_matched_void_id=best_matched_void
                matched_probs_exp_projected_void.append(prob[best_matched_void])
                matched_radii_exp_projected_void.append(radius_exp[this_void_id])
                if(prob[best_matched_void] >= prob_threshold):
                    matched_voids_exp=matched_voids_exp+1
                    #best_void_index[0,this_void_id]=best_matched_void_id
                    #lensing_void_radii[0,this_void_id]=centre_radii[int(best_matched_void_id)]
                norm_rads=np.linspace(0.1, 1.5*radius_exp[this_void_id],101)
                average_expansion[this_void_id] = lens_calc.average_expansion([centre_exp[this_void_id,0], centre_exp[this_void_id,1]], norm_rads, box_length, expansion_inter)

                f_matches_exp.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[best_matched_void], moved_revolver_centres[best_matched_void,0], moved_revolver_centres[best_matched_void,2],
                     radius_exp[this_void_id], centre_exp[this_void_id,0], centre_exp[this_void_id,1],
                     prob_exp[this_void_id]))

            f_matches_exp.close()


            matched_probs_sigma_projected_void=[]
            matched_radii_sigma_projected_void=[]
            matched_voids_sig=0
            i_3D_best_sigma = -1*np.ones(len(radius_sigma),dtype=int)

            for this_void_id in range(0,len(radius_sigma)):
                prob, i_3D_best_sigma[this_void_id] = lens_calc_ut.void_probability(eff_void_radius_gals, moved_revolver_centres[:,0], moved_revolver_centres[:,2], std_R_void_prob, std_xz_void_prob, radius_sigma[this_void_id], centre_sigma[this_void_id,0], centre_sigma[this_void_id,1], N_pixels)
                #best_matched_void=np.where(prob==np.max(prob))[0][0]
                best_matched_void = i_3D_best_sigma[this_void_id]
                best_matched_void_id = best_matched_void
                matched_probs_sigma_projected_void.append(prob[best_matched_void])
                matched_radii_sigma_projected_void.append(radius_sigma[this_void_id])
                if(prob[best_matched_void] >= prob_threshold):
                    matched_voids_sig=matched_voids_sig+1
                    #best_void_index[0,this_void_id]=best_matched_void_id
                    #lensing_void_radii[0,this_void_id]=centre_radii[int(best_matched_void_id)]
                norm_rads=np.linspace(0.1, 1.5*radius_sigma[this_void_id],101)
                average_sigma[this_void_id] = lens_calc.average_sigma([centre_sigma[this_void_id,0], centre_sigma[this_void_id,1]], norm_rads, box_length, shear_inter)

                print("Matched voids sigma: ",
                      eff_void_radius_gals[best_matched_void],moved_revolver_centres[best_matched_void,0], moved_revolver_centres[best_matched_void,2],
                      radius_sigma[this_void_id], centre_sigma[this_void_id,0], centre_sigma[this_void_id,1],
                      prob[best_matched_void])

                f_matches_sig.write("%12.3f  %12.3f  %12.3f   %12.3f  %12.3f  %12.3f   %12.3g\n" %
                    (eff_void_radius_gals[best_matched_void], moved_revolver_centres[best_matched_void,0], moved_revolver_centres[best_matched_void,2],
                     radius_sigma[this_void_id], centre_sigma[this_void_id,0], centre_sigma[this_void_id,1],
                     prob_sig[this_void_id]))

            f_matches_sig.close()


            figs.plot_optical_scalars_cart(x_analyse_grid, z_analyse_grid, rads, real_theta, rel_deviation, real_sigma, img_sigma, distance_check, centre_exp, radius_exp, centre_sigma, radius_sigma, average_expansion_revolver, average_expansion_revolver_cut, average_expansion, average_sigma_revolver, average_sigma_revolver_cut, average_sigma, moved_revolver_centres, best_matched_void_id_exp, best_matched_void_id_sig, box_length, void_count)
            figs.plot_rads_vs_prob(matched_radii_exp, matched_probs_exp, "expansion")
            figs.plot_rads_vs_prob(matched_radii_sigma, matched_probs_sigma, "sigma")
            figs.plot_rads_vs_prob(matched_radii_exp_projected_void, matched_probs_exp_projected_void, "expansion_projected_void")
            figs.plot_rads_vs_prob(matched_radii_sigma_projected_void, matched_probs_sigma_projected_void, "sigma_projected_void")
            figs.plot_void_population(eff_void_radius_gals, radius_exp, "expansion")
            figs.plot_void_population(eff_void_radius_gals, radius_sigma, "sigma")

            #deviation_to_dens_centre = calc.calc_centre_deviation([0,0], centre_exp[0,0])

            optical_scalars_file=open(("optical_scalars_%i.dat") % void_count,"w+")
            optical_scalars_file.write("N_void_exp=%i\n" % (len(radius_exp)))
            optical_scalars_file.write("N_void_sigma=%i\n" % (len(radius_sigma)))
            optical_scalars_file.write("N_void_revolver_matched_exp=%i\n" % matched_voids_revolver_exp)
            optical_scalars_file.write("matched_revolver__fraction_exp=%f\n" % (matched_voids_revolver_exp/len(eff_void_radius_gals)))
            optical_scalars_file.write("N_void_revolver_matched_sigma=%i\n" % matched_voids_revolver_sig)
            optical_scalars_file.write("matched_revolver_fraction_sigma=%f\n" % (matched_voids_revolver_sig/len(eff_void_radius_gals)))

            optical_scalars_file.write("N_void_matched_exp=%i\n" % matched_voids_exp)
            optical_scalars_file.write("matched_fraction_exp=%f\n" % (matched_voids_exp/len(radius_exp)))
            optical_scalars_file.write("N_void_matched_sigma=%i\n" % matched_voids_sig)
            optical_scalars_file.write("matched_fraction_sigma=%f\n" % (matched_voids_sig/len(radius_sigma)))
            optical_scalars_file.write("Revolver_id=%i\nour_id_exp=%i\nour_id_sig=%i\n" % (void_id_gals, best_matched_void_id_exp, best_matched_void_id_sig))
            optical_scalars_file.write("centre_x=%f\ncentre_z=%f\nradius=%f\n" % (centre_exp[0,0], centre_exp[0,1], radius_exp[0]) )
            #optical_scalars_file.write("deviation=%f\n" % deviation_to_dens_centre )
            optical_scalars_file.close()

            t = time.localtime()
            current_time = time.strftime("%H:%M:%S", t)
            print("time after diff solver loop =",current_time)

            del real_theta
            del rel_deviation
            del real_sigma
            del img_sigma

            del expansion_inter

        #if(do_solve_scalars==True and do_mass_def==True):

        #    x_analyse_grid = np.linspace(-(box_length/2), (box_length/2), N_pixels)
        #    z_analyse_grid = np.linspace(-(box_length/2), (box_length/2), N_pixels)
        #    rads =np.linspace(1, 2*void_radius, 101)

        #    kappa_grid=np.load("kappa_grid_1_0.npy")
        #    gamma_grid=np.load("gamma_grid_1_0.npy")
        #    real_theta=np.load("real_theta_1_0.npy")
        #    real_sigma=np.load("real_sigma_1_0.npy")
        #    img_sigma=np.load("img_sigma_1_0.npy")

        #    mag_sigma = (real_sigma**2)+(img_sigma**2)

        #    expansion_dev_map = real_theta/kappa_grid
        #    shear_dev_map = (mag_sigma/np.mean(mag_sigma))/((gamma_grid**2)/np.mean(gamma_grid**2))

        #    expansions=np.zeros((N_pixels*N_pixels))
        #    shears=np.zeros((N_pixels*N_pixels))
        #    xz_grid=[]
        #    k=0
        #    for l in range(0,len(x_analyse_grid)):
        #        for m in range(0,len(z_analyse_grid)):
        #            xz_grid.append([x_analyse_grid[l], z_analyse_grid[m]])
        #            expansions[k]=expansion_dev_map[l,m]
        #            shears[k]=shear_dev_map[l,m]
        #            k=k+1

        #    expansions_inter=LinearNDInterpolator(xz_grid, expansions, fill_value=np.mean(expansions))
        #    shears_inter=LinearNDInterpolator(xz_grid, shears, fill_value=np.mean(shears))

        #    average_deviation_expansions=lens_calc.average_shear_deviation([0,0], expansions_inter, rads, box_length)
        #    average_deviation_shears=lens_calc.average_shear_deviation([0,0], shears_inter, rads, box_length)

        #    figs.plot_optical_scalars_vs_weak_lens_cart(x_analyse_grid, z_analyse_grid, expansion_dev_map, shear_dev_map, rads, average_deviation_expansions, average_deviation_shears, void_count)

        void_ids=np.arange(0,len(eff_void_radius_gals))
        failed_mass_def=np.zeros(len(void_ids),dtype=int)
        same_mass_def=np.zeros(len(void_ids),dtype=int)
        failed_exp=np.zeros(len(void_ids),dtype=int)
        same_exp=np.zeros(len(void_ids),dtype=int)
        failed_sigma=np.zeros(len(void_ids),dtype=int)
        same_sigma=np.zeros(len(void_ids),dtype=int)
        for i in range(0,len(best_void_index)):
            if(best_void_index[0,i] == -1):
                failed_mass_def[i]=1
            if(best_void_index[1,i] == -1):
                failed_exp[i]=1
            if(best_void_index[2,i] == -1):
                failed_sigma[i]=1
            if(best_void_index[0,i] == void_ids[i]):
                same_mass_def[i]=1
            if(best_void_index[1,i] == void_ids[i]):
                same_exp[i]=1
            if(best_void_index[2,i] == void_ids[i]):
                same_sigma[i]=1
        identification_file=open("identification.dat","w+")
        identification_file.write("matched_voids_mass_def=%i\n" % (len(void_ids)-len(np.where(failed_mass_def==1)[0])) )
        identification_file.write("failed_match_mass_def=%i\n" % len(np.where(failed_mass_def==1)[0]) )
        identification_file.write("matched_voids_exp=%i\n" % (len(void_ids)-len(np.where(failed_exp==1)[0])) )
        identification_file.write("failed_match_exp=%i\n" % len(np.where(failed_exp==1)[0]) )
        identification_file.write("matched_voids_sigma=%i\n" % (len(void_ids)-len(np.where(failed_sigma==1)[0])) )
        identification_file.write("failed_match_sigma=%i\n" % len(np.where(failed_sigma==1)[0]) )


        #radii = optical_scalar_res[0,:,0]
        #real_theta = optical_scalar_res[0,:,1]
        #img_theta = optical_scalar_res[0,:,2]
        #real_sigma = optical_scalar_res[0,:,3]
        #img_sigma = optical_scalar_res[0,:,4]
#
#        plt.plot(radii, real_theta, 'r-', label='theta')
#        plt.plot(radii, real_sigma, 'b-', label='sigma')
#        plt.grid()
#        plt.legend(loc='best')
#        plt.xlabel('radius')
#        plt.ylabel('optical scalars')
#        figname="optical_scalars_" + "gals" + ".eps"
#        plt.savefig(figname)
#        plt.close()
#
#        plt.plot(radii, real_theta, 'r-', label='theta')
#        plt.grid()
#        plt.legend(loc='best')
#        plt.xlabel(r'$r$')
#        plt.ylabel(r'$\theta$')
#        figname="theta_" + "gals" + ".eps"
#        plt.savefig(figname)
#        plt.close()
#
#        plt.plot(radii, real_sigma, 'r-', label='sigma')
#        plt.grid()
#        plt.legend(loc='best')
#        plt.xlabel(r'$r$')
#        plt.ylabel(r'$\sigma$')
#        figname="sigma_" + "gals" + ".eps"
#        plt.savefig(figname)
#        plt.close()
#

    ### Solve equation optical scalars ###
    #lensing.dens_test(G, void_gal_ids, eff_void_radius_gals[void_id_gals], tracers, gal_pos, densities, circ_center_gals, void_id_gals, box_length, "gals")

    #lensing.stacking(num_voids_dm, circ_center_dm, tracers, eff_void_radius_dm, box_length, "dm")
    #lensing.stacking(num_voids_gals, circ_center_gals, gal_pos, eff_void_radius_gals, box_length, "gals")


    #lensing.mass_deficit(angular_size_void_dm_x, angular_size_void_dm_y, void_size_dm_z, distance, mass_dm, mean_density, eff_void_radius_dm[void_id_dm], tracers, densities, pots, circ_center_dm, void_id_dm, box_length, "dm", vels)
    #lensing.dens_test(eff_void_radius_dm[void_id_dm], tracers, densities, circ_center_dm, void_id_dm, box_length, "dm")
