#!/usr/bin/env python

# lensing_calculations - heavy work for geometrical optics
# (C) 2021, 2022, 2023 Marius Peper, Boud Roukema GPL-3+
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import LinearNDInterpolator
#from scipy.interpolate import Rbf
import scipy.integrate as integrate
import scipy.stats
from scipy.signal import savgol_filter
from multiprocessing import Pool

from extra_calculations import calculations
from coordinates import coordinate_calc
from read import Read
from lensing_calculations_unit_testable import lensing_calc_unit_testable as lens_calc_ut
import os

import ctypes
diff_solver_path = os.environ.get('DIFF_SOLVER_PATH')
optical_scalar_solver = ctypes.CDLL("%s/optical_scalar_solver.so" % diff_solver_path)
call_gsl_integration = ctypes.CDLL("%s/call_gsl_integration.so" % diff_solver_path)
from numpy.ctypeslib import ndpointer

import matplotlib.pyplot as plt

N_grad = 11
Nsnaps_analyse = 5

speed_of_light_Mpc_over_Myr = 0.306357515 # in Mpc/Myr
convert_Mpc_km = 3.086e19
M_sun_in_kg = 1.989e30
G = 6.67e-20 # km^3/(kg * s^2)

eta0_grid_extreme_high = 1.0e20

if os.environ.get('REL_ERR_GSL'):
    rel_err_gsl = float(os.environ.get('REL_ERR_GSL'))
else:
    rel_err_gsl = 1.0e-7

class lensing_calc_grid:

    ## none of the functions in this class are currently used!

    def time_deriv_grid(self, tracer_file_base, Nfiles, x, y, z, Ngrid, circ_center, void_id, time_interval):
        """
        This function calculates the gradient in time of the potential.
        It will read in the data of several time steps and in order to estimate
        the potential at previously chosen positions in the void for
        the time steps and finally it will calculate the gradient using
        the estimated potentials.

        Parameter
        ------------
        tracer_file_base: string
        The name and path of the tracer files. This routine opens several
        tracers files for different time steps. It assumes that the base
        string will stay the same and only varies the the number at the end
        which is the usual convention to numerate time steps.
        Nfiles: int
        Specifies how many files will be opened to model the time-evolution
        of the potential.
        x,y,z: np.array
        The positions organised in a meshgrid.
        Ngrid: int
        How many grid points to we want to generate in each direction.
        circ_center: np.array(3)
        The position of the circumcentre of the void. It will be used to
        move our particles according to it.
        void_id: int
        Specifies which void we are going to investigate.
        time_interval: float
        The time interval we chose for the simulation. It is given in Myr.
        returns
        ------------
        pots_time_deriv: np.array
        Holds the time derivative for all chosen time steps at all points
        in the grid.
        pots_inter: np.array
        Holds the potential at all points in the grid.
        """

        read = Read()
        calc = calculations()
        pots_inter = np.zeros((Nfiles, Ngrid, Ngrid, Ngrid))

        for i in range(Nfiles-10,Nfiles):
            print("i=",i)
            if i < 10:
                tracer_file = tracer_file_base + '00%d' % i
            elif(i < 100 and i >= 10):
                tracer_file = tracer_file_base + '0%d' % i
            else:
                tracer_file = tracer_file_base + '%d' % i
            print("tracer_file=",tracer_file)
            tracers = read.read_tracers(tracer_file)
            moved_tracers = calc.move_coordinates(tracers[:,0], tracers[:,1], tracers[:,2], circ_center[void_id], box_length)
            pots = read.read_pots(tracer_file)

            linInter= LinearNDInterpolator(moved_tracers, pots)

            pots_inter[i] = linInter(x,y,z)

        pots_time_line = np.empty(Nfiles)
        pots_time_deriv = np.empty((Nfiles, Ngrid, Ngrid, Ngrid))

        for l in range(0,Ngrid):
            for m in range(0,Ngrid):
                for n in range(0,Ngrid):
                    for i in range(Nfiles-10,Nfiles):
                        print("i=",i)
                        pots_time_line[i] = pots_inter[i,l,m,n]
                    print("pots_time_line[i]=",pots_time_line)
                    pots_time_deriv[:,l,m,n] = np.gradient(pots_time_line[:], time_interval)

        return pots_time_deriv, pots_inter

    def interpolate_on_grid(self, detector_variable, tracers, x, y, z, Ngrid, origin):
        """
        This function calculates interpolates a given scalar on
        all points of the given grid.

        Parameter
        ------------
        detector_variable: np.array(dtype=float)
        The quantity that will be interpolated.
        tracers: np.array(3,dtype=float)
        The spatial coordinates of the DM particles.
        x,y,z: np.array
        The positions organised in a meshgrid.
        Ngrid: int
        How many grid points to we want to generate in each direction.
        origin: np.array(3)
        The position of the centre, e.g. the centre of a void.
        returns
        ------------
        inter: np.array
        The interpolated detector_variable at the grid points.
        """

        calc = calculations()

        inter = np.empty((Ngrid, Ngrid, Ngrid))
        moved_tracers = calc.move_coordinates(tracers[:,0], tracers[:,1], tracers[:,2],
                                              origin, box_length)
        linInter= LinearNDInterpolator(moved_tracers, detector_variable)

        inter = linInter(x,y,z)

        return inter

class LensingCalc:

    def calc_mean_density(self, box_length, Ncroot, tracer_file):
        """
        The function will calculate the mean density of the simulated box.

        Parameter
        ------------
        box_length:
        The length of the sides of the box in Mpc/h.
        Ncroot:
        The number of particles in the simulation

        returns
        ------------
        The mean matter density of the simulated region.
        """

        read = Read()
        mass_dm = read.read_mass_dm(tracer_file)
        #return (Ncroot**3)*mass_dm*1.e10 * M_sun_in_kg/(box_length**3);
        return (Ncroot**3)/((box_length)**3);

    def interpolate_detector_variable_spherical(self, r, polar_angle, azimuth_angle, void_position_from_observer, offset, box_length, linInter):
        """
        This function will interpolate a detector_variable in a spherical coordinate
        system.

        Parameters
        ------------
        r, polar_angle, azimuth_angle: floats
        The spherical coordinates.

        void_position_from_observer: array(3,float)
        The distance of the hypothetical observer.

        offset:
        A possible offset that is universal to all values.

        linInter:
        The interpolated detector_variable in cartesian coordinates.

        returns
        ------------
        dens_inter: float
        Interpolated density at the coordinates given in the input.
        """

        coord = coordinate_calc()
        ## See
        ##
        ## 'COMOVING REFERENCE FRAMES: Euclid_obs, Euclid_sim, spherical_obs'
        ##
        ## in lensing.py for descriptions of these three systems.

        # r, polar_angle, azimuth_angle are the coordinates in spherical_obs

        # In this call to convert_spherical_to_cartesian,
        # * we first transform r, polar_angle, azimuth_angle to Euclid_obs; and
        # * then make a shift, using void_position_from_observer, to Euclid_sim.
        x,y,z = coord.convert_spherical_to_cartesian(r, polar_angle, azimuth_angle, void_position_from_observer)
        # x,y,z are now in Euclid_sim

        if(hasattr(x, '__len__')):
            leaking=np.where(x >= box_length/2)
            x[leaking]=-box_length+x[leaking]
            leaking=np.where(x <= -box_length/2)
            x[leaking]=box_length+x[leaking]
        else:
            if(x >= box_length/2):
                x=-box_length+x
            if(x <= -box_length/2):
                x=box_length+x
        if(hasattr(y, '__len__')):
            leaking=np.where(y >= box_length/2)
            y[leaking]=-box_length+y[leaking]
            leaking=np.where(y <= -box_length/2)
            y[leaking]=box_length+y[leaking]
        else:
            if(y >= box_length/2):
                y=-box_length+y
            if(y <= -box_length/2):
                y=box_length+y
        if(hasattr(z, '__len__')):
            leaking=np.where(z >= box_length/2)
            z[leaking]=-box_length+z[leaking]
            leaking=np.where(z <= -box_length/2)
            z[leaking]=box_length+z[leaking]
            # search elements at the edge, they might are out of the interpolation range
            # TODO: Imptove this!
        else:
            if(z >= box_length/2):
                z=-box_length+z
            if(z <= -box_length/2):
                z=box_length+z

        # x,y,z are in Euclid_sim
        inter = linInter(x,y,z) - offset

        return inter

    def interpolate_detector_variable_cart(self, dist, axis1, axis2, void_position_from_observer, offset, box_length, linInter):
        """
        This function will interpolate a detector_variable in a spherical coordinate
        system.

        Parameters
        ------------
        dist: distance from observer to the simulation box (float)

        axis1, axis2: comoving spatial locations x, z in Euclid_obs coordinates (floats)

        void_position_from_observer: array(3,float)
        The distance of the void from the hypothetical observer.

        offset:
        A possible offset in 'detector_variable' that is universal to all values.

        linInter:
        The interpolated detector_variable in cartesian coordinates.

        returns
        ------------
        interpolate_detector_variable_inter: float
        Interpolated detector_variable at the coordinates given in the input.
        """

        coord = coordinate_calc()
        x=axis1
        z=axis2
        # The result 'y' will typically be positive or negative, of much
        # smaller amplitude than void_position_from_observer[1].
        y=dist - void_position_from_observer[1]
        # x,y,z are in Euclid_sim after these three lines

        if(hasattr(x, '__len__')):
            leaking=np.where(x >= box_length/2)
            x[leaking]=-box_length+x[leaking]
            leaking=np.where(x <= -box_length/2)
            x[leaking]=box_length+x[leaking]
        else:
            if(x >= box_length/2):
                x=-box_length+x
            if(x <= -box_length/2):
                x=box_length+x
        if(hasattr(y, '__len__')):
            leaking=np.where(y >= box_length/2)
            y[leaking]=-box_length+y[leaking]
            leaking=np.where(y <= -box_length/2)
            y[leaking]=box_length+y[leaking]
        else:
            if(y >= box_length/2):
                y=-box_length+y
            if(y <= -box_length/2):
                y=box_length+y
        if(hasattr(z, '__len__')):
            leaking=np.where(z >= box_length/2)
            z[leaking]=-box_length+z[leaking]
            leaking=np.where(z <= -box_length/2)
            z[leaking]=box_length+z[leaking]
            # search elements at the edge, they might are out of the interpolation range
            # TODO: Imptove this!
        else:
            if(z >= box_length/2):
                z=-box_length+z
            if(z <= -box_length/2):
                z=box_length+z

        # x,y,z are in Euclid_sim
        inter = linInter(x,y,z) - offset

        return inter


    def construct_orth_plane(self, x, y, z):
        """
        Construct the two basis vectors that will span an orthonogal plane with
        respect to a given vector (x,y,z).
        """


        #print("input vector", x,y,z)
        ortho_basis_1 = np.empty(3)
        ortho_basis_2 = np.empty(3)
        # take a fixed vector to start the gram-schmidt process

        # assume x is given in Mpc
        v = np.array([0, 0, 1])

        norm_input_vector = np.sqrt(np.inner([x,y,z],[x,y,z]))

        if( np.abs(np.inner([x,y,z],v)) <= 1.e-10):
            # v is perpendicular to (x,y,z)
            #print("vectors are perpendicular")
            ortho_basis_1 = v

        elif(v[0]/x == v[1]/y == v[2]/z):
            # v and (x,y,z) are parallel
            #print("vectors are orthogonal")
            ortho_basis_1 = [0,1,0]
            #ortho_basis_2 = [0,0,1]

        else:
            #print("contruct vector")
            #norm = np.inner([x,y,z],v)/norm_input_vector
            #ortho_basis_1[0] = v[0] - (norm * x/norm_input_vector)
            #ortho_basis_1[1] = v[1] - (norm * y/norm_input_vector)
            #ortho_basis_1[2] = v[2] - (norm * z/norm_input_vector)
            input_vector=np.array([x,y,z])/norm_input_vector
            #print("input_vector: ", input_vector)
            ortho_basis_1 = v-(np.inner(input_vector,v)/np.inner(input_vector,input_vector))*input_vector
            #print("ortho_basis_1:", ortho_basis_1)

        ortho_basis_2 = np.cross([x,y,z],ortho_basis_1)

        ortho_norm_1 = np.sqrt(np.inner(ortho_basis_1,ortho_basis_1))
        ortho_norm_2 = np.sqrt(np.inner(ortho_basis_2,ortho_basis_2))
        #print("ortho_norm_1: ",ortho_norm_1)

        ortho_basis_1 = ortho_basis_1/ortho_norm_1
        ortho_basis_2 = ortho_basis_2/ortho_norm_2

        #print("check for orthogonality")

        #print("[x,y,z]=",[x,y,z])
        #print("ortho_basis_1=",ortho_basis_1)
        #print("ortho_basis_2=",ortho_basis_2)

        #print("np.inner(ortho_basis_1,[x,y,z]) = ",np.inner(ortho_basis_1,[x,y,z]))
        #print("np.inner(ortho_basis_2,[x,y,z]) = ",np.inner(ortho_basis_2,[x,y,z]))
        #print("np.inner(ortho_basis_1,ortho_basis_2) = ",np.inner(ortho_basis_1,ortho_basis_2))

        return ortho_basis_1, ortho_basis_2


    def time_deriv_cart(self, Nfiles, dist, x_analyse_grid, z_analyse_grid, snap_times, void_position_from_observer, scale_factor, box_length, N_pixels, pots_time_line, pots_time_line_y, pots_time_line_x, pots_time_line_z):
        """
        This function calculates the gradient in time of the potential.
        It will read in the data of several time steps and in order to estimate
        the potential at previously chosen positions in the void for
        the time steps and finally it will calculate the gradient using
        the estimated potentials.

        Parameter
        ------------


        returns
        ------------
        """

        # TODO: this routine does not include any mixed second derivatives
        # with the spatial coordinates.

        coord=coordinate_calc()

        time_deriv_on_y = np.zeros((N_grad,Nsnaps_analyse,N_pixels,N_pixels))
        time_deriv_on_x = np.zeros((N_grad,Nsnaps_analyse,N_pixels,N_pixels))
        time_deriv_on_z = np.zeros((N_grad,Nsnaps_analyse,N_pixels,N_pixels))

        axis1 = x_analyse_grid
        axis2 = z_analyse_grid

        x = axis1
        y = dist # - void_position_from_observer[1]
        z = axis2

        y_range = np.linspace((y-5.0),(y+5.0),N_grad)
        x_range = np.linspace((x-5.0),(x+5.0),N_grad) # x_range.shape = (N_grad,N_pixels,N_pixels)
        z_range = np.linspace((z-5.0),(z+5.0),N_grad)

        # Idea: use Schwarz's theorem 1/(dt dr) = 1/(dr dt) to shorten calculation time.

        #for i in range(0,N_grad):
        time_deriv_on_y = np.gradient(pots_time_line_y, snap_times, axis=0, edge_order=2)
        time_deriv_on_x = np.gradient(pots_time_line_x, snap_times, axis=0, edge_order=2)
        time_deriv_on_z = np.gradient(pots_time_line_z, snap_times, axis=0, edge_order=2)

        pots_time_deriv = np.gradient(pots_time_line, snap_times, axis=0, edge_order=2)
        scale_time_deriv = np.gradient(scale_factor, snap_times, edge_order=2)
        #dtdy = np.gradient(time_deriv_on_y[:,int(Nsnaps_analyse-1)], y_range, edge_order=2)
        #dtdx = np.gradient(time_deriv_on_x[:,int(Nsnaps_analyse-1)], x_range, edge_order=2)
        #dtdz = np.gradient(time_deriv_on_z[:,int(Nsnaps_analyse-1)], z_range, edge_order=2)
        dtdy = np.gradient(time_deriv_on_y[3,:,:,:,:], y_range, axis=0, edge_order=2)
        dtdx = np.gradient(time_deriv_on_x[3,:,:,:,:], abs(x_range[0,0,0]-x_range[1,0,0]), axis=0, edge_order=2)
        dtdz = np.gradient(time_deriv_on_z[3,:,:,:,:], abs(z_range[0,0,0]-z_range[1,0,0]), axis=0, edge_order=2)
        pots_time_curve = np.gradient(pots_time_deriv, snap_times, axis=0, edge_order=2)

        return pots_time_deriv[3,:,:,:], pots_time_curve[3,:,:,:], dtdy[int((N_grad-1)/2),:,:,:], dtdx[int((N_grad-1)/2),:,:,:], dtdz[int((N_grad-1)/2),:,:,:], scale_time_deriv[3]
        #return pots_time_deriv[3,:,:], pots_time_curve[3,:,:], scale_time_deriv[3]


    def spatial_pot_derivs(self, mag_on_line_xz, mag_on_line_yx, mag_on_line_zy ,x, y, z):

        y_range = np.linspace((y-5.0),(y+5.0),N_grad) # Mpc/h
        x_range = np.linspace((x-5.0),(x+5.0),N_grad) # Mpc/h
        z_range = np.linspace((z-5.0),(z+5.0),N_grad) # Mpc/h
        step=10/N_grad

        slope_x = np.gradient(mag_on_line_xz, step, axis=0, edge_order=2) # 1/(Mpc/h)
        slope_y = np.gradient(mag_on_line_yx, step, axis=0, edge_order=2) # 1/(Mpc/h)
        slope_z = np.gradient(mag_on_line_zy, step, axis=0, edge_order=2) # 1/(Mpc/h)
        curve_x = np.gradient(slope_x[:,int((N_grad-1)/2),:,:,:], step, axis=0, edge_order=2) # 1/(Mpc/h)^2
        curve_y = np.gradient(slope_y[:,int((N_grad-1)/2),:,:,:], step, axis=0, edge_order=2) # 1/(Mpc/h)^2
        curve_z = np.gradient(slope_z[:,int((N_grad-1)/2),:,:,:], step, axis=0, edge_order=2) # 1/(Mpc/h)^2


        dxdz = np.gradient(slope_x[int((N_grad-1)/2),:,:,:,:], step, axis=0, edge_order=2)
        dydx = np.gradient(slope_y[int((N_grad-1)/2),:,:,:,:], step, axis=0, edge_order=2)
        dydz = np.gradient(slope_z[int((N_grad-1)/2),:,:,:,:], step, axis=0, edge_order=2)
        dxdy = dydx
        dzdx = dxdz
        dzdy = dydz

        return ( slope_x[int((N_grad-1)/2),int((N_grad-1)/2),:,:,:], curve_x[int((N_grad-1)/2),:,:,:],
                 slope_y[int((N_grad-1)/2),int((N_grad-1)/2),:,:,:], curve_y[int((N_grad-1)/2),:,:,:],
                 slope_z[int((N_grad-1)/2),int((N_grad-1)/2),:,:,:], curve_z[int((N_grad-1)/2),:,:,:],
                 dxdy[int((N_grad-1)/2),:,:,:], dxdz[int((N_grad-1)/2),:,:,:],
                 dydx[int((N_grad-1)/2),:,:,:], dydz[int((N_grad-1)/2),:,:,:],
                 dzdx[int((N_grad-1)/2),:,:,:], dzdy[int((N_grad-1)/2),:,:,:] )


    def galaxy_redshift_centre(self, G, void_gal_ids, moved_gal_pos, redshift, distance, circ_center, void_id, box_length):

        coord=coordinate_calc()
        calc = calculations()

        gal_vels=G.Vel[void_gal_ids]
        #gal_positions=G.Pos[void_gal_ids]

        unit_vector = np.transpose(coord.construct_cart_unit_vect(moved_gal_pos[np.array(void_gal_ids),0], moved_gal_pos[np.array(void_gal_ids),1]+distance, moved_gal_pos[np.array(void_gal_ids),2]))
        gal_los_vel = np.diagonal(np.inner(gal_vels,unit_vector))

        gal_redshift=((1+redshift)*(1+gal_los_vel/3.e5))-1

        redshift_mean=np.mean(gal_redshift,axis=0)
        redshift_std=np.std(gal_redshift,axis=0)

        moved_gals=calc.move_coordinates(G.Pos[:,0], G.Pos[:,1], G.Pos[:,2],
                                        ([0.5*box_length,0.5*box_length,0.5*box_length]), box_length)


        gal_in_slice_nonvoid_ids = np.delete(np.arange(0,len(moved_gals)),void_gal_ids)
        gal_in_slice_nonvoid_index=np.where(abs(0 - moved_gals[gal_in_slice_nonvoid_ids,1]) <= 5)[0]
        gal_in_slice_nonvoid_ids=gal_in_slice_nonvoid_ids[gal_in_slice_nonvoid_index]
        void_gal_in_slice_index=np.where(abs(0 - moved_gals[void_gal_ids,1]) <= 10)[0]
        void_gal_in_slice_ids=void_gal_ids[void_gal_in_slice_index]

        plt.figure()
        plt.xlabel("x in Mpc/h")
        plt.ylabel("z in Mpc/h")
        void_gal_plot=plt.scatter(moved_gals[void_gal_in_slice_ids,0], moved_gals[void_gal_in_slice_ids,2],
                                  marker="o", color="blue", s=10, label="void galaxies")
        nonvoid_gal_plot=plt.scatter(moved_gals[gal_in_slice_nonvoid_ids,0], moved_gals[gal_in_slice_nonvoid_ids,2],
                                     marker="o", color="red", s=10, label="non-void galaxies")
        #plt.scatter(redshift_mean[0], redshift_mean[2], color='red', marker='x')

        plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), ncol=3, fontsize=12)
        void_gal_plot.set_alpha(1)
        nonvoid_gal_plot.set_alpha(1)
        plt.tight_layout()
        plt.savefig("void_galaxies_slice.eps")
        plt.close()

        return redshift_mean, redshift_std, gal_redshift

    def prior(self, pos):

        if(pos[0] >= 40):
            return np.inf
        if(pos[0] <= -40):
            return np.inf
        if(pos[1] >= 40):
            return np.inf
        if(pos[1] <= -40):
            return np.inf

        return 0


    def average_mass_deficit(self, pos, rads, box_length, mass_def_inter):

        coord=coordinate_calc()
        average_mass_def=np.zeros((len(rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(rads)):

            #if(self.rads[i] < 5):
            #    angles=np.linspace(0,2*np.pi,11)
            if(rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            mass_def_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                mass_def_ring[j]=mass_def_inter(coord.cylinder_coord(rads[i], phi, pos, box_length)[0], coord.cylinder_coord(rads[i], phi, pos, box_length)[1] )
                j=j+1

            mass_def_ring=savgol_filter(mass_def_ring,15,3)
            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, mass_def_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, int(len(angles))  )
            integral=float(X)

            average_mass_def[i]=integral/(2*np.pi)

        return average_mass_def


    def mass_def_sky_plane_void_search(self, x):

        # step (ii) in the paper
        average_mass_def = self.average_mass_deficit([x[0], x[1]], self.rads, self.box_length, self.mass_def_inter)
        slopes=np.gradient(average_mass_def,self.rads,edge_order=2)
        continue_crit=0
        i=5
        this_slope=slopes[i]

        # step (iii) in the paper
        diff=average_mass_def[i]-average_mass_def[i-5]
        #while(diff <= 90 ):
        while(continue_crit < 4):
            i=i+1
            this_slope=slopes[i]
            diff=average_mass_def[i]-average_mass_def[i-5]
            if(this_slope > np.mean(np.abs(slopes)) and average_mass_def[i]-average_mass_def[i-1] > self.void_stddev_frac_crit*np.std(average_mass_def[:i])):
                continue_crit = continue_crit + 1
            if(average_mass_def[i] < self.void_diff_frac_crit * np.mean(average_mass_def[:i])):
                # increase is withing error margin, do not break here!
                diff=0
                continue_crit = 0
            if(i==100):
                break
        # step (iv) in the paper
        if(i<100):
            while(average_mass_def[i+1]>average_mass_def[i]):
                i=i+1
                if(i>=100):
                    break

        # step (v) in the paper
        if(i>=95):
            delta_deficit=0.001/eta0_grid_extreme_high # not currently used further
            delta_rads=0.001/eta0_grid_extreme_high
        else:
            delta_deficit = np.max(average_mass_def[i+5]) - average_mass_def[0]
            delta_rads=self.rads[np.where(average_mass_def==np.max(average_mass_def[i+5]))[0]]-self.rads[0]

        return 1/(delta_rads), i


    def mass_def_starter(self, x, z):

        mass_def_eta0_grid, edge_i=self.mass_def_sky_plane_void_search([x, z])
        # Warning: Calculating gamma for a pixel like this would require a symmetric projected surface density Sigma!
        # In general such a Sigma is not given in a real matter distribution.
        # The calculation of gamma_sym_model might be interesting to check but it is not a real estimate for the
        # weak lensing shear gamma at a given position!
        kappa, gamma_sym_model, kappa_grid, gamma_sym_model_grid, Sigma_bar, Sigma_crit = self.mass_def_lensing(self.mass_def, [x, z], self.distance, self.mass_dm, self.OmegaM, self.OmegaLambda, self.Hubble, self.N_pixels, self.x_analyse_grid, self.z_analyse_grid, self.rads, self.box_length)

        # 2023-03-25 Is this commented section (for loop, if clause) obsolete?
        # Delete it if it's unlikely to be recovered.

        #xz_grid=[]
        # TODO: replace the name 'mass_deficit_lensing' in the three occurrences in this
        # function by a less confusing name such as 'gamma_lensing'! This *should* not
        # modify the function's behaviour, but should clarify its role.
        #mass_deficit_lensing=np.zeros((self.N_pixels*self.N_pixels))
        #k=0
        #for l in range(0,len(self.x_analyse_grid)):
        #    for m in range(0,len(self.z_analyse_grid)):
        #        xz_grid.append([self.x_analyse_grid[l], self.z_analyse_grid[m]])
        #        mass_deficit_lensing[k]=gamma_sym_model_grid[l,m]
        #        k=k+1

        #self.mass_def_lensing_inter=LinearNDInterpolator(xz_grid, mass_deficit_lensing)
        #if(i==31 and j==31):
        #    average_mass_def_lensing_dens=self.average_mass_def_lensing([x_analyse_grid[i],z_analyse_grid[j]])
        #    np.save("average_mass_def_lensing_dens",average_mass_def_lensing_dens)

        self.Sigma_bar=Sigma_bar
        self.Sigma_crit=Sigma_crit

        gamma_eta0_grid, mass_lensing_edge_i=self.weak_lensing_shear_sky_plane_void_search([x, z])

        return mass_def_eta0_grid, edge_i, gamma_eta0_grid, mass_lensing_edge_i



    def mass_def_centre(self, mass_def_inter_in, x_analyse_grid, z_analyse_grid, rads_in, mass_def_median_in, distance, mass_dm, OmegaM, OmegaLambda, Hubble, mass_def, N_pixels, box_length_in, void_stddev_frac_crit, void_diff_frac_crit, N_average, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean, merge_distance, void_count):
        from scipy.optimize import minimize
        self.mass_def_median=mass_def_median_in

        self.mass_def_inter = mass_def_inter_in
        #self.MCMC_mass_def_centre()
        #centre_sim1,centre_sim2=self.simplified_mcmc()

        self.rads=rads_in
        self.box_length=box_length_in
        self.x_analyse_grid=x_analyse_grid
        self.z_analyse_grid=z_analyse_grid
        self.mass_def=mass_def

        self.distance=distance
        self.mass_dm=mass_dm
        self.OmegaM=OmegaM
        self.OmegaLambda=OmegaLambda
        self.Hubble=Hubble
        self.N_pixels=N_pixels

        self.void_stddev_frac_crit=void_stddev_frac_crit
        self.void_diff_frac_crit=void_diff_frac_crit

        #delta = lambda pos: self.mass_def_sky_plane_void_search(pos)[0]

        #sol = minimize(delta, [0,0])
        mass_def_eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        gamma_eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        edge_i=np.zeros((len(x_analyse_grid),len(z_analyse_grid)), dtype=int)
        mass_lensing_edge_i=np.zeros((len(x_analyse_grid),len(z_analyse_grid)), dtype=int)
        xz_grid = []
        x_grid = []
        z_grid = []

        for i in range(0,len(x_analyse_grid)):
            for j in range(0,len(z_analyse_grid)):
                # step (i) in the paper
                if((self.mass_def_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.mass_def_median)):
                    # Speed up the calculation by sorting out obviously pixel that will not be the centre.
                    mass_def_eta0_grid[i,j] = 2.0*eta0_grid_extreme_high
                    gamma_eta0_grid[i,j] = 2.0*eta0_grid_extreme_high
                else:
                    xz_grid.append([x_analyse_grid[i], z_analyse_grid[j]])
                    x_grid.append(i)
                    z_grid.append(j)

        pool = Pool(10)

        # steps (ii)-(v) in the paper
        res = pool.starmap_async(self.mass_def_starter, xz_grid)
        pool.close()
        pool.join()
        average_res = np.array(res.get())

        mass_def_eta0_grid[( np.array(x_grid), np.array(z_grid) )]=average_res[:,0]
        edge_i[( np.array(x_grid), np.array(z_grid) )]=average_res[:,1]
        gamma_eta0_grid[( np.array(x_grid), np.array(z_grid) )]=average_res[:,2]
        mass_lensing_edge_i[( np.array(x_grid), np.array(z_grid) )]=average_res[:,3]


        #################################################################################################################
        # Quick analyser...
        plot_mesh1,plot_mesh2=np.meshgrid(x_analyse_grid, z_analyse_grid, indexing = 'ij', sparse = False)
        # step (vi) of the paper begins here: ++++++++++++
        # "sort_out" means "apply a label for future rejection"
        pixels_to_be_removed=np.where(mass_def_eta0_grid >= eta0_grid_extreme_high)
        print("pixels_to_be_removed: ", pixels_to_be_removed)

        print("Before indexing with pixels_to_be_removed, np.shape(mass_def_eta0_grid) = ",np.shape(mass_def_eta0_grid))
        mass_def_eta0_grid[pixels_to_be_removed]=-1
        mass_def_eta0_grid[pixels_to_be_removed]=np.max(mass_def_eta0_grid)
        print("After indexing with pixels_to_be_removed, np.shape(mass_def_eta0_grid) = ",np.shape(mass_def_eta0_grid))

        lowest_mean=9999999
        mean_map=np.zeros((N_pixels,N_pixels))
        mean_map[:]=np.max(mass_def_eta0_grid)
        #for i in range(5,len(x_analyse_grid)-5):
        #    for l in range(5,len(z_analyse_grid)-5):
        #        neighbours = mass_def_eta0_grid[(i-int(N_average/2)):(i+int(N_average/2)+1),(l-int(N_average/2)+1):(l+int(N_average/2)+1)]
        i_debug1=0
        for i in range(0,len(x_analyse_grid)):
            for l in range(0,len(z_analyse_grid)):
                width = int(N_average/2)
                full_width_plus1 = 1+2*width
                #neighbours = mass_def_eta0_grid[i_select,l_select] # T^2 square centred on [i,l]
                neighbours = np.roll(mass_def_eta0_grid,
                                     (-(i-width),-(l-width)),axis=(0,1))[0:full_width_plus1,0:full_width_plus1]
                i_debug1 += 1
                if(i_debug1<3):
                    print("mass_def_eta0_grid:: neighbours=", neighbours)
                    print("mass_def_eta0_grid:: np.shape(neighbours)=", np.shape(neighbours))
                pixel_median = np.median(neighbours)
                #pixel_mean = np.mean(neighbours)
                pixel_mean = lens_calc_ut.lowpass_filter(neighbours) # use a lowpass filter rather than a top-hat filter
                mean_map[i,l] = pixel_mean
                if(mass_def_eta0_grid[i,l]==np.max(mass_def_eta0_grid)):
                    # sort out pixels that are previousy identified to be false
                    mean_map[i,l] = np.max(mass_def_eta0_grid)
                if(lowest_mean>pixel_mean):
                    lowest_mean_coord=[np.array([x_analyse_grid[i]]),np.array([z_analyse_grid[l]])]
                    lowest_mean=pixel_mean

        # step (vi) of the paper ends here: ++++++++++++
        print("lowest_mean_coord: ", lowest_mean_coord)

        np.save("mass_def_eta0_grid", mass_def_eta0_grid)
        np.save("mean_map", mean_map)

        # step (vii) of the paper, identify the local minima:
        local_minima=lens_calc_ut.local_minima(mean_map, np.mean(mass_def_eta0_grid), np.std(mass_def_eta0_grid), N_pixels, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean)
        local_minima=np.where(local_minima==True)
        local_minima_x=x_analyse_grid[local_minima[0]]
        local_minima_z=z_analyse_grid[local_minima[1]]
        pre_centre=np.stack((local_minima_x, local_minima_z), axis=1)

        mass_def_radius = self.rads[edge_i[local_minima[0], local_minima[1]] ]
        print("mass_def_radius: ", mass_def_radius)

        # step (viii) of the paper, merge close (in distance) local minima:
        local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, mass_def_radius, merge_distance, N_pixels)

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'z')
        plt.xlabel(r'x')
        c = ax.pcolormesh(plot_mesh1,plot_mesh2, mass_def_eta0_grid, vmin=np.percentile(mass_def_eta0_grid,5), vmax=np.percentile(mass_def_eta0_grid,95))
        cbar=plt.colorbar(c,ax=ax)
        plt.scatter(lowest_mean_coord[0][0], lowest_mean_coord[1][0], marker='x', color='red', label='centre')
        plt.scatter(local_centres[:,0], local_centres[:,1], marker='x', color="tomato")
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value10, cmap='GnBu')
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value30, cmap='GnBu')
        cbar.set_label(r'$P(\Sigma)$', rotation = 270, labelpad=15)
        figname=("mass_deficit_likelihood_" + "gals_%i" + ".eps") % void_count
        plt.savefig(figname)
        plt.close()
        ###################################################################################################################

        ###### GAMMA ######
        #################################################################################################################
        # Quick analyser...
        pixels_to_be_removed=np.where(gamma_eta0_grid >= eta0_grid_extreme_high)

        gamma_eta0_grid[pixels_to_be_removed]=-1
        gamma_eta0_grid[pixels_to_be_removed]=np.max(gamma_eta0_grid)

        lowest_mean=9999999
        mean_map=np.zeros((N_pixels,N_pixels))
        mean_map[:]=np.max(gamma_eta0_grid)
        for i in range(0,len(x_analyse_grid)):
            for l in range(0,len(z_analyse_grid)):
                width = int(N_average/2)
                #neighbours = gamma_eta0_grid[i_select,l_select] # T^2 square centred on [i,l]
                full_width_plus1 = 1+2*width
                neighbours = np.roll(gamma_eta0_grid,
                                     (-(i-width),-(l-width)),axis=(0,1))[0:full_width_plus1,0:full_width_plus1]

                pixel_median = np.median(neighbours)
                #pixel_mean = np.mean(neighbours)
                pixel_mean = lens_calc_ut.lowpass_filter(neighbours) # use a lowpass filter rather than a top-hat filter
                mean_map[i,l] = pixel_mean
                if(gamma_eta0_grid[i,l]==np.max(gamma_eta0_grid)):
                    # sort out pixels that are previousy identified to be false
                    mean_map[i,l] = np.max(gamma_eta0_grid)
                if(lowest_mean>pixel_mean):
                    lowest_mean_coord=[np.array([x_analyse_grid[i]]),np.array([z_analyse_grid[l]])]
                    lowest_mean=pixel_mean

        print("lowest_mean_coord: ", lowest_mean_coord)

        np.save("gamma_eta0_grid", gamma_eta0_grid)
        np.save("gamma_mean_map", mean_map)

        local_minima_gamma=lens_calc_ut.local_minima(mean_map, np.mean(gamma_eta0_grid), np.std(gamma_eta0_grid), N_pixels, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean)
        local_minima_gamma=np.where(local_minima_gamma==True)
        local_minima_x_gamma=x_analyse_grid[local_minima_gamma[0]]
        local_minima_z_gamma=z_analyse_grid[local_minima_gamma[1]]
        pre_centre_gamma=np.stack((local_minima_x_gamma, local_minima_z_gamma), axis=1)

        print("pre_centre_gamma: ", pre_centre_gamma)

        gamma_radius = self.rads[mass_lensing_edge_i[local_minima_gamma[0], local_minima_gamma[1]] ]

        local_centres_gamma, local_radii_gamma = lens_calc_ut.merge_centres(pre_centre_gamma, gamma_radius, merge_distance, N_pixels)

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'z')
        plt.xlabel(r'x')
        c = ax.pcolormesh(plot_mesh1,plot_mesh2, gamma_eta0_grid, vmin=np.percentile(gamma_eta0_grid,5), vmax=np.percentile(gamma_eta0_grid,95))
        cbar=plt.colorbar(c,ax=ax)
        plt.scatter(lowest_mean_coord[0][0], lowest_mean_coord[1][0], marker='x', color='red', label='centre')
        plt.scatter(local_centres_gamma[:,0], local_centres_gamma[:,1], marker='x', color="tomato")
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value10, cmap='GnBu')
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value30, cmap='GnBu')
        cbar.set_label(r'$P(\gamma)$', rotation = 270, labelpad=15)
        figname=("gamma_likelihood_" + "gals_%i" + ".eps") % void_count
        plt.savefig(figname)
        plt.close()
        ###################################################################################################################


        #return centre, centre_radii, average_mass_def_dens, average_mass_def, average_mass_def_gamma_centre, average_mass_def_combined, average_mass_def_lensing_dens, average_mass_def_lensing, average_mass_def_gamma_centre_lensing, average_mass_def_combined_lensing, kappa_grid, gamma_grid

        average_mass_def_lensing_dens=self.average_mass_def_lensing([local_centres[2,0], local_centres[2,1]], self.rads, self.box_length, self.mass_def_inter)
        print("local_radii[0]: ", local_radii[2])
        np.save("average_mass_def_lensing_dens",average_mass_def_lensing_dens)

        return local_centres, local_radii, local_centres_gamma, local_radii_gamma


    def MCMC_mass_def_centre(self):

        # Not used yet!
        # Not finished yet!

        import emcee

        ndim = 2

        nwalkers = 32
        p0 = np.random.rand(nwalkers, ndim)*40

        sampler = emcee.EnsembleSampler(nwalkers, ndim, self.mass_def_sky_plane_void_search)

        state = sampler.run_mcmc(p0, 100, skip_initial_state_check=True)
        sampler.reset()

        sampler.run_mcmc(state, 100);

        samples = sampler.get_chain(flat=True)
        print("samples: ", samples)
        with open("mcmc.dat","w+") as F:
            for i in range(0,len(samples)):
                F.write("%f  %f\n" % (samples[i,0], samples[i,1]))
        #plt.hist(samples[:, 0], 100, color="k", histtype="step")


    def simplified_mcmc(self):
        N_steps=500
        centre=[-20,-20]
        pos=[0,0]
        delta_old = self.mass_def_sky_plane_void_search(pos)

        for i in range(0,N_steps):

            pos[0]=centre[0]+(np.random.normal(0, 5, 1))[0]
            pos[1]=centre[1]+(np.random.normal(0, 5, 1))[0]
            delta = self.mass_def_sky_plane_void_search(pos)

            print("pos_test:",pos)
            print("delta:",delta)
            print("delta_old:",delta_old)
            if(delta < delta_old):
                # did the position leaked outside?
                print("accepted")
                if(pos[0] >= 40):
                    pos[0]=-80+pos[0]
                if(pos[0] <= -40):
                    pos[0]=80+pos[0]
                if(pos[1] >= 40):
                    pos[1]=-80+pos[1]
                if(pos[1] <= -40):
                    pos[1]=80+pos[1]
                centre=pos
                delta_old=delta
            else:
                # reject
                pos=centre
            print("pos_new:",centre)

        centre1= centre
        centre=[20,20]
        pos=[0,0]
        delta_old = self.mass_def_sky_plane_void_search(pos)

        for i in range(0,N_steps):

            pos[0]=centre[0]+(np.random.normal(0, 5, 1))[0]
            pos[1]=centre[1]+(np.random.normal(0, 5, 1))[0]
            delta = self.mass_def_sky_plane_void_search(pos)

            if(delta < delta_old):
                centre=pos
                delta_old=delta
            else:
                # reject
                pos=centre
        centre2= centre

        print("simplified centre1: ", centre1)
        print("simplified centre2: ", centre2)

        return centre1, centre2


    def average_expansion(self, pos, rads, box_length, expansion_inter):

        # Can we merge this with the other average scripts?
        # Could leak out of box...
        coord=coordinate_calc()

        average_expansion=np.zeros((len(rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(rads)):

            if(rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            expansion_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                expansion_ring[j]=expansion_inter(coord.cylinder_coord(rads[i], phi, pos, box_length)[0], coord.cylinder_coord(rads[i], phi, pos, box_length)[1] )
                j=j+1

            expansion_ring=savgol_filter(expansion_ring,15,3)

            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, expansion_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, len(angles))
            integral=float(X)

            average_expansion[i]=integral/(2*np.pi)

        return average_expansion

    def expansion_sky_plane_void_search(self, x):

        average_exp = self.average_expansion([x[0], x[1]], self.rads, self.box_length, self.expansion_inter)
        slopes=np.gradient(average_exp,self.rads,edge_order=2)
        continue_crit=0

        i_initial=5
        i=i_initial
        this_slope=slopes[i]

        diff=average_exp[i]-average_exp[i-5]
        #while(diff <= 0.3e-5 ):
        while(continue_crit < 4):
            i=i+1
            this_slope=slopes[i]
            diff=average_exp[i]-average_exp[i-5]
            if((this_slope > np.mean(np.abs(slopes))) and (average_exp[i]-average_exp[i-1] > self.void_stddev_frac_crit_theta*np.std(average_exp[:i])) ):
                continue_crit = continue_crit + 1
            if(average_exp[i] < self.void_diff_frac_crit_theta * np.mean(average_exp[:i])):
                # increase is withing error margin, do not break here!
                diff=0
                continue_crit = 0
            if(i==100):
                break
        if(i<100):
            while(average_exp[i+1]>average_exp[i]):
                i=i+1
                if(i>=100):
                    break

        #or average_exp[i] < 200
        if(i>=95):
            delta_expansion=0.001/eta0_grid_extreme_high
            delta_rads=0.001/eta0_grid_extreme_high
        else:
            delta_expansion = np.max(average_exp[i_initial:i+5]) - average_exp[0]
            delta_rads = self.rads[np.where(average_exp==np.max(average_exp[i_initial:i+5]))[0]]-self.rads[0]


        #if(any(average_exp[:10] > 9.e-6) ):
        #    # remove pixel where initial values are too high
        #    delta_deficit=0.001/eta0_grid_extreme_high
        #    delta_rads=0.001/eta0_grid_extreme_high

        if(hasattr(delta_expansion, '__len__')):
            delta_expansion=delta_expansion[0]
        if(hasattr(delta_rads, '__len__')):
            delta_rads=delta_rads[0]

        return 1/(delta_rads), i

    def expansion_starter(self, x, z):

        eta0_grid, edge_i = self.expansion_sky_plane_void_search([x, z])
        return eta0_grid, edge_i

    def expansion_centre(self, expansion_inter_in, x_analyse_grid, z_analyse_grid, rads_in, median_expansion_in, box_length_in, N_pixels, void_stddev_frac_crit_theta, void_diff_frac_crit_theta, N_average, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean, merge_distance, void_count):
        from scipy.optimize import minimize

        self.expansion_inter = expansion_inter_in
        self.median_expansion=median_expansion_in
        self.box_length=box_length_in
        self.rads=rads_in
        self.void_stddev_frac_crit_theta = void_stddev_frac_crit_theta
        self.void_diff_frac_crit_theta = void_diff_frac_crit_theta

        eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        edge_i=np.zeros((len(x_analyse_grid),len(z_analyse_grid)), dtype=int)
        xz_grid=[]
        x_grid=[]
        z_grid=[]

        for i in range(0,len(x_analyse_grid)):
            for j in range(0,len(z_analyse_grid)):
                #if( (self.expansion_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.median_expansion)  or (np.sqrt(x_analyse_grid[i]**2 + z_analyse_grid[j]**2) >= void_radius) ):
                #if( (self.expansion_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.median_expansion)  or (np.abs(x_analyse_grid[i]) > self.box_length/4) or (np.abs(z_analyse_grid[j]) > self.box_length/4) ):
                if( (self.expansion_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.median_expansion)):
                    # Speed up the calculation by sorting out obviously pixel that will not be the centre.
                    eta0_grid[i,j]=2.0*eta0_grid_extreme_high
                else:
                    xz_grid.append([x_analyse_grid[i], z_analyse_grid[j]])
                    x_grid.append(i)
                    z_grid.append(j)

        pool = Pool(10)

        res = pool.starmap_async(self.expansion_starter, xz_grid)
        pool.close()
        pool.join()
        average_res = np.array(res.get())

        eta0_grid[( np.array(x_grid), np.array(z_grid) )]=average_res[:,0]
        print("expansion centre")
        print("max(eta0_grid): ", np.max(eta0_grid))
        print("min(eta0_grid): ", np.min(eta0_grid))
        edge_i[( np.array(x_grid), np.array(z_grid) )]=average_res[:,1]

        #################################################################################################################
        # Quick analyser...
        plot_mesh1,plot_mesh2=np.meshgrid(x_analyse_grid, z_analyse_grid, indexing = 'ij', sparse = False)
        pixels_to_be_removed=np.where(eta0_grid >= eta0_grid_extreme_high)

        eta0_grid[pixels_to_be_removed]=-1
        eta0_grid[pixels_to_be_removed]=np.max(eta0_grid)

        lowest_mean=9999999
        mean_map=np.zeros((N_pixels,N_pixels))
        mean_map[:]=np.max(eta0_grid)
        for i in range(0,len(x_analyse_grid)):
            for l in range(0,len(z_analyse_grid)):
                width = int(N_average/2)
                full_width_plus1 = 1+2*width
                #neighbours = eta0_grid[i_select,l_select] # T^2 square centred on [i,l]
                neighbours = np.roll(eta0_grid,
                                     (-(i-width),-(l-width)),axis=(0,1))[0:full_width_plus1,0:full_width_plus1]

                pixel_median = np.median(neighbours)
                #pixel_mean = np.mean(neighbours)
                pixel_mean = lens_calc_ut.lowpass_filter(neighbours) # use a lowpass filter rather than a top-hat filter
                mean_map[i,l] = pixel_mean
                if(eta0_grid[i,l]==np.max(eta0_grid)):
                    # sort out pixels that are previousy identified to be false
                    mean_map[i,l] = np.max(eta0_grid)
                if(lowest_mean>pixel_mean):
                    lowest_mean_coord=[np.array([x_analyse_grid[i]]),np.array([z_analyse_grid[l]])]
                    lowest_mean=pixel_mean


        print("max(eta0_grid): ", np.max(eta0_grid))
        print("min(eta0_grid): ", np.min(eta0_grid))
        print("np.percentile(eta0_grid,95): ", np.percentile(eta0_grid,95))
        np.save("expansion_eta0_grid", eta0_grid)
        np.save("expansion_map", mean_map)

        local_minima=lens_calc_ut.local_minima(mean_map, np.mean(eta0_grid), np.std(eta0_grid), N_pixels, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean)
        local_minima=np.where(local_minima==True)
        local_minima_x=x_analyse_grid[local_minima[0]]
        local_minima_z=z_analyse_grid[local_minima[1]]
        pre_centre=np.stack((local_minima_x, local_minima_z), axis=1)

        expansion_radius = self.rads[edge_i[local_minima[0], local_minima[1]] ]

        local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, expansion_radius, merge_distance, N_pixels)

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'z')
        plt.xlabel(r'x')
        c = ax.pcolormesh(plot_mesh1,plot_mesh2, eta0_grid, vmin=np.percentile(eta0_grid,5), vmax=np.percentile(eta0_grid,95))
        cbar=plt.colorbar(c,ax=ax)
        plt.scatter(lowest_mean_coord[0][0], lowest_mean_coord[1][0], marker='x', color='red', label='centre')
        plt.scatter(local_centres[:,0], local_centres[:,1], marker='x', color="tomato")
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value10, cmap='GnBu')
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value30, cmap='GnBu')
        cbar.set_label(r'$P(\theta)$', rotation = 270, labelpad=15)
        figname=("expansion_likelihood_" + "gals_%i" + ".eps") % void_count
        plt.savefig(figname)
        plt.close()
        ###################################################################################################################

        return local_centres, local_radii


    def average_sigma(self, pos, rads, box_length, sigma_inter):

        # Can we merge this with the other average scripts?
        # Could leak out of box...
        coord=coordinate_calc()

        average_sigma=np.zeros((len(rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(rads)):

            if(rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            sigma_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                sigma_ring[j]=sigma_inter(coord.cylinder_coord(rads[i], phi, pos, box_length)[0], coord.cylinder_coord(rads[i], phi, pos, box_length)[1] )
                j=j+1

            sigma_ring=savgol_filter(sigma_ring,15,3)

            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, sigma_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, len(angles))
            integral=float(X)

            average_sigma[i]=integral/(2*np.pi)

        return average_sigma

    def sigma_sky_plane_void_search(self, x):

        average_sig = self.average_sigma([x[0], x[1]], self.rads, self.box_length, self.sigma_inter)
        slopes=np.gradient(average_sig,self.rads,edge_order=2)

        i_initial=5
        i=i_initial
        this_slope=slopes[i]
        continue_crit=0

        diff=average_sig[i]-average_sig[i-5]
        #while(diff <= 0.6e-5 ):
        while(continue_crit < 4):
            i=i+1
            this_slope=slopes[i]
            diff=average_sig[i]-average_sig[i-5]
            if((this_slope > np.mean(np.abs(slopes))) and (average_sig[i]-average_sig[i-1] > self.void_stddev_frac_crit_sigma*np.std(average_sig[:i]))):
                continue_crit = continue_crit + 1
            if(average_sig[i] < self.void_diff_frac_crit_sigma * np.mean(average_sig[:i])):
                # increase is within error margin, do not break here!
                diff=0
                continue_crit = 0
            if(i==100):
                break
        if(i<100):
            while(average_sig[i+1]>average_sig[i]):
                i=i+1
                if(i>=100):
                    break

        #or average_sig[i] < 200
        if(i>=95):
            delta_sigma=0.001/eta0_grid_extreme_high
            delta_rads=0.001/eta0_grid_extreme_high
        else:
            delta_sigma = np.max(average_sig[i_initial:i+5]) - average_sig[0]
            delta_rads = self.rads[np.where(average_sig==np.max(average_sig[i_initial:i+5]))[0]]-self.rads[0]


        #if(any(average_sig[:10] > 2.e-5) ):
        #    # remove pixel where initial values are too high
        #    delta_deficit=0.001/eta0_grid_extreme_high
        #    delta_rads=0.001/eta0_grid_extreme_high

        if(hasattr(delta_sigma, '__len__')):
            delta_sigma=delta_sigma[0]
        if(hasattr(delta_rads, '__len__')):
            delta_rads=delta_rads[0]

        return 1/(delta_rads), i

    def sigma_starter(self, x, z):

        eta0_grid, edge_i = self.sigma_sky_plane_void_search([x, z])
        return eta0_grid, edge_i

    def sigma_centre(self, sigma_inter_in, x_analyse_grid, z_analyse_grid, rads_in, median_sigma_in, box_length_in, N_pixels, void_stddev_frac_crit_sigma, void_diff_frac_crit_sigma, N_average, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean, merge_distance, void_count):
        from scipy.optimize import minimize

        self.sigma_inter = sigma_inter_in
        self.median_sigma=median_sigma_in
        self.box_length=box_length_in
        self.rads=rads_in
        self.void_stddev_frac_crit_sigma = void_stddev_frac_crit_sigma
        self.void_diff_frac_crit_sigma = void_diff_frac_crit_sigma

        print("self.void_stddev_frac_crit_sigma: ", self.void_stddev_frac_crit_sigma)

        eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        edge_i=np.zeros((len(x_analyse_grid),len(z_analyse_grid)), dtype=int)
        xz_grid=[]
        x_grid=[]
        z_grid=[]

        for i in range(0,len(x_analyse_grid)):
            for j in range(0,len(z_analyse_grid)):
                #if( (self.sigma_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.median_sigma)  or (np.sqrt(x_analyse_grid[i]**2 + z_analyse_grid[j]**2) >= void_radius) ):
                #if( (self.sigma_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.median_sigma)  or (np.abs(x_analyse_grid[i]) > self.box_length/4) or (np.abs(z_analyse_grid[j]) > self.box_length/4) ):
                if( (self.sigma_inter(x_analyse_grid[i],z_analyse_grid[j]) > self.median_sigma) ):
                    # Speed up the calculation by sorting out obviously pixel that will not be the centre.
                    eta0_grid[i,j]=2.0*eta0_grid_extreme_high
                else:
                    xz_grid.append([x_analyse_grid[i], z_analyse_grid[j]])
                    x_grid.append(i)
                    z_grid.append(j)

        pool = Pool(10)

        res = pool.starmap_async(self.sigma_starter, xz_grid)
        pool.close()
        pool.join()
        average_res = np.array(res.get())

        eta0_grid[( np.array(x_grid), np.array(z_grid) )]=average_res[:,0]
        edge_i[( np.array(x_grid), np.array(z_grid) )]=average_res[:,1]
        print("sigma centre")
        print("max(eta0_grid): ", np.max(eta0_grid))
        print("min(eta0_grid): ", np.min(eta0_grid))

        #################################################################################################################
        # Quick analyser...
        plot_mesh1,plot_mesh2=np.meshgrid(x_analyse_grid, z_analyse_grid, indexing = 'ij', sparse = False)
        pixels_to_be_removed=np.where(eta0_grid >= eta0_grid_extreme_high)

        eta0_grid[pixels_to_be_removed]=-1
        eta0_grid[pixels_to_be_removed]=np.max(eta0_grid)

        lowest_mean=9999999
        mean_map=np.zeros((N_pixels,N_pixels))
        mean_map[:]=np.max(eta0_grid)
        for i in range(0,len(x_analyse_grid)):
            for l in range(0,len(z_analyse_grid)):
                width = int(N_average/2)
                full_width_plus1 = 1+2*width
                #neighbours = eta0_grid[i_select,l_select] # T^2 square centred on [i,l]
                neighbours = np.roll(eta0_grid,
                                     (-(i-width),-(l-width)),axis=(0,1))[0:full_width_plus1,0:full_width_plus1]

                pixel_median = np.median(neighbours)
                #pixel_mean = np.mean(neighbours)
                pixel_mean = lens_calc_ut.lowpass_filter(neighbours) # use a lowpass filter rather than a top-hat filter
                mean_map[i,l] = pixel_mean
                if(eta0_grid[i,l]==np.max(eta0_grid)):
                    # sort out pixels that are previousy identified to be false
                    mean_map[i,l] = np.max(eta0_grid)
                if(lowest_mean>pixel_mean):
                    lowest_mean_coord=[np.array([x_analyse_grid[i]]),np.array([z_analyse_grid[l]])]
                    lowest_mean=pixel_mean

        np.save("sigma_eta0_grid", eta0_grid)
        np.save("sigma_map", mean_map)
        print("max(eta0_grid): ", np.max(eta0_grid))
        print("min(eta0_grid): ", np.min(eta0_grid))
        print("np.percentile(eta0_grid,95): ", np.percentile(eta0_grid,95))

        local_minima=lens_calc_ut.local_minima(mean_map, np.median(eta0_grid), np.std(eta0_grid), N_pixels, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean )
        local_minima=np.where(local_minima==True)
        local_minima_x=x_analyse_grid[local_minima[0]]
        local_minima_z=z_analyse_grid[local_minima[1]]
        pre_centre=np.stack((local_minima_x, local_minima_z), axis=1)

        sigma_radius = self.rads[edge_i[local_minima[0], local_minima[1]] ]

        local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, sigma_radius, merge_distance, N_pixels)

        fig, ax = plt.subplots(1, 1)
        plt.ylabel(r'z')
        plt.xlabel(r'x')
        c = ax.pcolormesh(plot_mesh1,plot_mesh2, eta0_grid, vmin=np.percentile(eta0_grid,5), vmax=np.percentile(eta0_grid,95))
        cbar=plt.colorbar(c,ax=ax)
        plt.scatter(lowest_mean_coord[0][0], lowest_mean_coord[1][0], marker='x', color='red', label='centre')
        plt.scatter(local_centres[:,0], local_centres[:,1], marker='x', color="tomato")
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value10, cmap='GnBu')
        #equi_pixel = ax.pcolormesh(azimuth, polar, random_pixel_value30, cmap='GnBu')
        cbar.set_label(r'$P(\sigma)$', rotation = 270, labelpad=15)
        figname=("sigma_likelihood_" + "gals_%i" + ".eps") % void_count
        plt.savefig(figname)
        plt.close()
        ###################################################################################################################

        return local_centres, local_radii


    def average_los_velocity(self, pos):

        # Can we merge this with the other average scripts?
        # Could leak out of box...
        coord=coordinate_calc()

        average_los_velocity=np.zeros((len(self.rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(self.rads)):

            if(self.rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*self.rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            los_velocity_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                los_velocity_ring[j]=np.abs(self.los_velocity_inter(coord.cylinder_coord(self.rads[i], phi, pos, self.box_length)[0], coord.cylinder_coord(self.rads[i], phi, pos, self.box_length)[1] ) )
                j=j+1
            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, los_velocity_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, len(angles))
            integral=float(X)

            average_los_velocity[i]=integral/(2*np.pi)

        return average_los_velocity

    def los_velocity_sky_plane_void_search(self, x):

        average_los_vel = self.average_los_velocity([x[0], x[1]])
        delta_los_velocity = np.max(average_los_vel) - average_los_vel[0]
        slopes=np.gradient(average_los_vel,self.rads,edge_order=2)


        i=0
        this_slope=slopes[i]

        if(np.abs(average_los_vel[i]-average_los_vel[i-1]) < 0.5*np.std(average_los_vel[:i]) ):
            # increase is withing error margin, do not break here!
            this_slope=1000

        while(this_slope/np.abs(np.mean(np.abs(slopes))) < 0.5 ):
            i=i+1

            this_slope=slopes[i]

            if(np.abs(average_los_vel[i]-average_los_vel[i-1]) < 0.5*np.std(average_los_vel[:i]) ):
                # increase is withing error margin, do not break here!
                this_slope=1000
            if(i==100):
                break
        if(i>=95):
            delta_los_velocity=0.001
            delta_rads=0.001
        else:
            delta_los_velocity = np.max(average_los_vel[i+5]) - average_los_vel[0]
            delta_rads=self.rads[np.where(average_los_vel==np.max(average_los_vel[i+5]))[0]]-self.rads[0]

        if(any(average_los_vel[:15] > 100) ):
            # remove pixel where initial values are too high
            delta_los_velocity=0.0001
            delta_rads=0.0001
        if(np.isnan(delta_los_velocity)):
            delta_los_velocity=0.001
            delta_rads=0.001
        if(self.rads[i] > self.box_length/2):
            delta_los_velocity=0.001
            delta_rads=0.001
        if(np.isnan(average_los_vel[0])):
            delta_los_velocity=0.001
            delta_rads=0.001

        if(hasattr(delta_los_velocity, '__len__')):
            delta_los_velocity=delta_los_velocity[0]
        if(hasattr(delta_rads, '__len__')):
            delta_rads=delta_rads[0]

        return 1/(delta_los_velocity+10*delta_rads)


    def los_velocity_centre(self, los_velocity_inter_in, x_analyse_grid, z_analyse_grid, mean_los_vel, box_length_in):
        from scipy.optimize import minimize

        self.los_velocity_inter = los_velocity_inter_in
        self.box_length=box_length_in
        self.rads=np.linspace(1, 1.5*(self.box_length/2), 101)

        eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        for i in range(0,len(x_analyse_grid)):
            for j in range(0,len(z_analyse_grid)):
                if(self.los_velocity_inter(x_analyse_grid[i],z_analyse_grid[j]) > mean_los_vel ):
                    # Speed up the calculation by sorting out obviously pixel that will not be the centre.
                    eta0_grid[i,j]=9999
                else:
                    eta0_grid[i,j]=self.los_velocity_sky_plane_void_search([x_analyse_grid[i],z_analyse_grid[j]])

        centre_grid=[x_analyse_grid[(np.where(eta0_grid==np.min(eta0_grid))[0])], x_analyse_grid[(np.where(eta0_grid==np.min(eta0_grid))[1])] ]

        average_los_vel = self.average_los_velocity([centre_grid[0], centre_grid[1]])
        average_los_vel_dens = self.average_los_velocity([0,0])

        plt.plot(self.rads, average_los_vel, 'k-', label='los velocity centre')
        plt.plot(self.rads, average_los_vel_dens, 'r-', label='Revolver centre')
        plt.grid()
        plt.legend(loc='best')
        plt.xlabel('r')
        plt.ylabel(r'$\bar{v_{\mathrm{los}}} (km/s)$')
        figname="average_los_velocity" + ".eps"
        plt.savefig(figname, format = 'eps')
        plt.close()

        centre=centre_grid

        return centre


    def average_radial_velocity(self, pos):

        # Can we merge this with the other average scripts?
        # Could leak out of box...
        coord=coordinate_calc()

        average_radial_velocity=np.zeros((len(self.rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(self.rads)):

            if(self.rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*self.rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            radial_velocity_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                radial_velocity_ring[j]=np.abs(self.radial_velocity_inter(coord.cylinder_coord(self.rads[i], phi, pos, self.box_length)[0], coord.cylinder_coord(self.rads[i], phi, pos, self.box_length)[1] ) )
                j=j+1
            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, radial_velocity_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, len(angles))
            integral=float(X)

            average_radial_velocity[i]=integral/(2*np.pi)

        return average_radial_velocity

    def radial_velocity_sky_plane_void_search(self, x):

        average_radial_vel = self.average_radial_velocity([x[0], x[1]])
        delta_radial_velocity = np.max(average_radial_vel) - average_radial_vel[0]
        #print("average_radial_vel: ",average_radial_vel)
        #print("np.max(average_radial_vel): ", np.max(average_radial_vel))
        #print("self.rads: ",self.rads)
        delta_rads = self.rads[np.where(average_radial_vel==np.max(average_radial_vel))[0]]-self.rads[0]

        if(np.max(average_radial_vel) == average_radial_vel[0]):
            # retrun high value as this is obviously not the correct place
            delta_radial_velocity=0.001
        if(np.isnan(delta_radial_velocity)):
            delta_radial_velocity=0.001
        if(np.isnan(average_radial_vel[0])):
            average_radial_vel[0]=9999

        if(hasattr(delta_radial_velocity, '__len__')):
            delta_radial_velocity=delta_radial_velocity[0]
        if(hasattr(delta_rads, '__len__')):
            delta_rads=delta_rads[0]

        return 1/(delta_radial_velocity+10*delta_rads)


    def radial_velocity_centre(self, radial_velocity_inter_in, x_analyse_grid, z_analyse_grid, mean_radial_vel, box_length_in):
        from scipy.optimize import minimize

        self.radial_velocity_inter = radial_velocity_inter_in
        self.box_length=box_length_in
        self.rads=np.linspace(1, 1.5*(self.box_length/2), 101)

        eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        for i in range(0,len(x_analyse_grid)):
            for j in range(0,len(z_analyse_grid)):
                if(self.radial_velocity_inter(x_analyse_grid[i],z_analyse_grid[j]) > mean_radial_vel ):
                    # Speed up the calculation by sorting out obviously pixel that will not be the centre.
                    eta0_grid[i,j]=9999
                else:
                    eta0_grid[i,j]=self.radial_velocity_sky_plane_void_search([x_analyse_grid[i],z_analyse_grid[j]])

        centre_grid=[x_analyse_grid[(np.where(eta0_grid==np.min(eta0_grid))[0])], x_analyse_grid[(np.where(eta0_grid==np.min(eta0_grid))[1])] ]

        average_radial_vel = self.average_radial_velocity([centre_grid[0], centre_grid[1]])
        average_radial_vel_dens = self.average_radial_velocity([0,0])

        plt.plot(self.rads, average_radial_vel, 'k-', label='radial velocity centre')
        plt.plot(self.rads, average_radial_vel_dens, 'r-', label='Revolver centre')
        plt.grid()
        plt.legend(loc='best')
        plt.xlabel('r')
        plt.ylabel(r'$\bar{v_{\mathrm{radial}}} (km/s)$')
        figname="average_radial_velocity" + ".eps"
        plt.savefig(figname, format = 'eps')
        plt.close()

        centre=centre_grid

        return centre


    def bisect_func(self, redshift, OmegaM, OmegaLambda, Hubble, distance_to_source):
        coord=coordinate_calc()

        # return distance in Mpc/h
        return distance_to_source-coord.calc_distance(redshift, OmegaM, OmegaLambda, Hubble)*(Hubble/100)


    def average_mass_def_lensing(self, pos, rads, box_length, mass_def_lensing_inter):

        # Could leak out of box...
        coord=coordinate_calc()

        average_mass_def_lensing=np.zeros((len(rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(rads)):

            #if(self.rads[i] < 5):
            #    angles=np.linspace(0,2*np.pi,11)
            if(rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            mass_def_lensing_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                mass_def_lensing_ring[j]=mass_def_lensing_inter(coord.cylinder_coord(rads[i], phi, pos, box_length)[0], coord.cylinder_coord(rads[i], phi, pos, box_length)[1] )
                j=j+1
            mass_def_lensing_ring=savgol_filter(mass_def_lensing_ring,15,3)

            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, mass_def_lensing_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, int(len(angles)))
            integral=float(X)

            average_mass_def_lensing[i]=integral/(2*np.pi)

        return average_mass_def_lensing


    def weak_lensing_shear_sky_plane_void_search(self, x):

        #average_mass_def_lensing = self.average_mass_def_lensing([x[0], x[1]], self.rads, self.box_length, self.mass_def_lensing_inter)
        average_mass_def = self.average_mass_deficit([x[0], x[1]], self.rads, self.box_length, self.mass_def_inter)
        # Use eq. 47 of https://arxiv.org/abs/1411.0115
        gamma_mean = (self.Sigma_bar - average_mass_def)/self.Sigma_crit
        print("gamma_mean: ", gamma_mean)
        print("len(gamma_mean): ", len(gamma_mean))
        print("weak_lensing_shear_sky_plane_void_search: 100.0*void_gamma_frac_slope_max_percentile : ",100.0*self.void_gamma_frac_slope_max_percentile) # debug only

        slopes=np.gradient(gamma_mean,self.rads,edge_order=2) #slope = d_gamma/d_rads
        local_avg=np.zeros(len(self.rads))
        local_slope=np.zeros(len(self.rads))

        # Calculate smoothed values of gamma and its slope.
        for i in range(0,len(self.rads)):
            if(i==0):
                local_avg[i]=np.mean(gamma_mean[i:i+3])
                local_slope[i]=np.median(slopes[i:i+3])
            elif(i==1):
                local_avg[i]=np.mean(gamma_mean[i-1:i+3])
                local_slope[i]=np.median(slopes[i-1:i+3])
            elif(i==2):
                local_avg[i]=np.mean(gamma_mean[i-2:i+3])
                local_slope[i]=np.median(slopes[i-2:i+3])
            elif(i==len(self.rads)-1):
                local_avg[i]=np.mean(gamma_mean[i-3:i])
                local_slope[i]=np.median(slopes[i-3:i])
            elif(i==len(self.rads)-2):
                local_avg[i]=np.mean(gamma_mean[i-3:i+1])
                local_slope[i]=np.median(slopes[i-3:i+1])
            else:
                local_avg[i]=np.mean(gamma_mean[i-3:i+3])
                local_slope[i]=np.median(slopes[i-3:i+3])

        # Calculate the low percentile - which should lie among the slightly negative slopes.
        neg_slope_cut = np.percentile(local_avg, 100.0*self.void_gamma_frac_slope_max_percentile)
        print("weak_lensing_shear_sky_plane_void_search: neg_slope_cut = ",neg_slope_cut)

        i=0
        # First phase: increment through the radii where gamma is above the
        # global mean and gamma' is at least slightly negative.
        #
        # Caveat: the "global" mean depends on how the range of 'rads' is
        # assigned in lensing.py.
        while(local_avg[i] > np.mean(local_avg) and local_slope[i] < neg_slope_cut):
            i=i+1
            if(i>=len(self.rads)):
                break
        i_begin=i
        # Second phase: increment through the radii where gamma has not yet
        # risen above the global mean.
        while(local_avg[i] < np.mean(local_avg)):
            i=i+1
            if(i>=len(self.rads)):
                break
        i_end=i

        if(i_end!=i_begin):
            i_centre=int(np.where(local_avg==np.min(local_avg[i_begin:i_end]))[0])
        else:
            i_centre=99

        delta_deficit = np.abs(gamma_mean[i_centre] - gamma_mean[0])
        delta_rads=self.rads[i_centre]-self.rads[0]

        if(i_end==len(self.rads) or i_begin==len(self.rads) or i_end==i_begin or i_begin==0):
            # The dump never returned to 0
            # This could be possible though
            delta_deficit=0.001/eta0_grid_extreme_high
            delta_rads=0.001/eta0_grid_extreme_high
        if(abs(np.mean(local_avg))/np.max(abs(local_avg))<0.1):
            # Remove cases without a clear structure and only noise.
            delta_deficit=0.001/eta0_grid_extreme_high
            delta_rads=0.001/eta0_grid_extreme_high
        if(np.isnan(delta_deficit)):
            delta_deficit=0.001/eta0_grid_extreme_high
            delta_rads=0.001/eta0_grid_extreme_high
        if(np.isnan(gamma_mean[0])):
            delta_deficit=0.001/eta0_grid_extreme_high
            delta_rads=0.001/eta0_grid_extreme_high
        #if(i==99):
        #    delta_deficit=0.001
        #    delta_rads=0.001

        return 1/(delta_rads), i


    # Not used!
    def mass_def_lensing_centre(self, mass_def_lensing_inter_in, x_analyse_grid, z_analyse_grid): #box_length

        from scipy.optimize import minimize

        self.mass_def_lensing_inter = mass_def_lensing_inter_in
        #self.MCMC_mass_def_centre()
        #centre_sim1,centre_sim2=self.simplified_mcmc()

        #self.rads=np.linspace(0.1, 1.5*box_length, 101)
        #delta = lambda pos: self.mass_def_sky_plane_void_search(pos)[0]

        #sol = minimize(delta, [0,0])


        eta0_grid=np.zeros((len(x_analyse_grid),len(z_analyse_grid)))
        for i in range(0,len(x_analyse_grid)):
            for j in range(0,len(z_analyse_grid)):
                eta0_grid[i,j], mass_lensing_edge_i=self.weak_lensing_shear_sky_plane_void_search([x_analyse_grid[i],z_analyse_grid[j]])
                print("eta0_grid[i,j] :", eta0_grid[i,j])

        return eta0_grid


    def mass_def_lensing(self, mass_def, pos, distance, mass_dm, OmegaM, OmegaLambda, Hubble, N_pixels, x_analyse_grid, z_analyse_grid, rads, box_length):

        from scipy.optimize import bisect

        # rads from mass_def_centre
        #self.rads=np.linspace(0.1, 60, 101)
        kappa=np.zeros((len(rads)))
        # Warning: Calculating gamma for a pixel like this would require a symmetric projected surface density Sigma!
        # In general such a Sigma is not given in a real matter distribution.
        # The calculation of gamma_sym_model might be interesting to check but it is not a real estimate for the
        # weak lensing shear gamma at a given position!
        gamma_sym_model=np.zeros((len(rads)))
        Sigma_bar=np.zeros((len(rads)))

        average_mass_def = self.average_mass_deficit([pos[0],pos[1]], rads, box_length, self.mass_def_inter) # unit: m_DM/(Mpc/h)^2

        kappa_grid=np.zeros((N_pixels,N_pixels))
        gamma_sym_model_grid=np.zeros((N_pixels,N_pixels))
        delta_sig_grid=np.zeros((N_pixels,N_pixels))

        # TODO: convert to angular diameter distance!
        distance_to_source = 2*distance
        redshift_to_source=bisect(self.bisect_func, 0, 5, args=(OmegaM, OmegaLambda, Hubble, distance_to_source))
        redshift_to_lens=bisect(self.bisect_func, 0, 5, args=(OmegaM, OmegaLambda, Hubble, distance))
        print("redshift_to_lens: ", redshift_to_lens)

        speed_of_light_in_km_over_s=299792.458
        km_in_Mpc_over_h =1/(3.086e19/(Hubble/100))
        # Sigma_crit unit:(km/s)^2/((km)^3/kg s) ((Mpc/h)/(Mpc/h)^2) = 1/km kg 1/(Mpc/h)
        # = 1/km_in_Mpc_over_h * 1/(mass_dm * 1.e10 * (M_sun_in_kg/(Hubble/100))) * m_DM/(Mpc/h)^2
        Sigma_crit = (speed_of_light_in_km_over_s)**2/(4*np.pi*G)*2*distance/(distance+distance)/(1+redshift_to_source)*(1+redshift_to_lens)**2/(mass_dm * 1.e10 * (M_sun_in_kg/(Hubble/100)))/km_in_Mpc_over_h

        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double
        call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

        for i in range(0, len(rads)):

            if(i>=3):
                X = call_gsl_integration.call_gsl_integration(rads[:(i+1)], (average_mass_def[:(i+1)]*rads[:(i+1)]),
                                                              rads[0], rads[int(len(rads[:(i+1)])) -1],
                                                              1.e-30, rel_err_gsl, len(rads[:(i+1)]))
                # unit: m_DM
                integral=np.double(X)
            elif(i==2):
                # use trapezium rule for too few rad values
                #unit: m_DM
                integral=( (rads[1]-rads[0]) * (average_mass_def[0]*rads[0]+average_mass_def[1]*rads[1])/2  +
                           (rads[2]-rads[1]) * (average_mass_def[1]*rads[1]+average_mass_def[2]*rads[2])/2 )
            elif(i==1):
                integral=( (rads[1]-rads[0]) * (average_mass_def[0]*rads[0]+average_mass_def[1]*rads[1])/2  )
            elif(i==0):
                # single trapezium integration over the interval from 0 to rads[0]
                #unit: m_DM
                integral=( rads[0] * (average_mass_def[0] * rads[0])/2 )
                # unit: m_DM/(Mpc/h)^2
                Sigma_bar[0]=average_mass_def[0] # limiting value; more robust than integrating and dividing
            else:
                print("Error: mass_def_lensing: programming error; i = %d < 0\n" % (i))

            if(i>=1):
                # unit: m_DM/(Mpc/h)^2
                Sigma_bar[i]=2*integral/(rads[i]**2)

            gamma_sym_model[i]=integral/Sigma_crit
            kappa[i]=average_mass_def[i]/Sigma_crit


        rads_inter=rads
        rads_inter[int(len(rads_inter)-1)]=rads[int(len(rads_inter)-1)]+0.1 # cheat increasing it by a small value to avoid stepping out of the interpolation
        Sigma_bar_inter=interp1d(rads_inter, Sigma_bar, kind='linear', bounds_error=False, fill_value=Sigma_bar[int(len(Sigma_bar)-1)])
        rads_inter[int(len(rads_inter)-1)]=rads[int(len(rads_inter)-1)]-0.1 # Does the command overwrite self.rads?
        # TODO: find a better place for this calculation as it's rather heavy"
        for i in range(0,len(x_analyse_grid)):
            for l in range(0,len(z_analyse_grid)):
                this_x=(x_analyse_grid[i]-pos[0])
                this_z=(z_analyse_grid[l]-pos[1])
                if((x_analyse_grid[i]-pos[0]) >= (box_length/2)):
                    this_x = -box_length+(x_analyse_grid[i]-pos[0])
                if((x_analyse_grid[i]-pos[0]) <= -(box_length/2)):
                    this_x = box_length+(x_analyse_grid[i]-pos[0])
                if((z_analyse_grid[l]-pos[1]) >= (box_length/2)):
                    this_z = -box_length+(z_analyse_grid[l]-pos[1])
                if((z_analyse_grid[l]-pos[1]) <= -(box_length/2)):
                    this_z = box_length+(z_analyse_grid[l]-pos[1])
                this_rad=np.sqrt(this_x**2 + this_z**2)
                if(this_rad < 1.e-3):
                    delta_sig_grid[i,l]=0
                    kappa_grid[i,l]=mass_def[i,l]/Sigma_crit
                    gamma_sym_model_grid[i,l]=delta_sig_grid[i,l]/Sigma_crit
                else:
                    delta_sig_grid[i,l]=Sigma_bar_inter(this_rad)-mass_def[i,l]
                    kappa_grid[i,l]=mass_def[i,l]/Sigma_crit
                    gamma_sym_model_grid[i,l]=delta_sig_grid[i,l]/Sigma_crit

        return kappa, gamma_sym_model, kappa_grid, gamma_sym_model_grid, Sigma_bar, Sigma_crit


    #As of 2023-01-14, this function is not currently used.
    def average_shear_deviation(self, pos, inter, rads, box_length):

        # Could leak out of box...
        coord=coordinate_calc()

        average_deviation=np.zeros((len(rads)))
        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")
        call_gsl_integration.call_gsl_integration.restype = ctypes.c_double

        for i in range(0,len(rads)):

            #if(self.rads[i] < 5):
            #    angles=np.linspace(0,2*np.pi,11)
            if(rads[i]>5):
                angles=np.linspace(0,2*np.pi,int(2*np.pi*rads[i]))
            else:
                angles=np.linspace(0,2*np.pi,20)
            average_deviation_ring = np.zeros((len(angles)))

            j=0
            for phi in angles:
                average_deviation_ring[j]=inter(coord.cylinder_coord(rads[i], phi, pos, box_length)[0], coord.cylinder_coord(rads[i], phi, pos, box_length)[1] )
                j=j+1
            average_deviation_ring=savgol_filter(average_deviation_ring,15,3)

            call_gsl_integration.call_gsl_integration.argtypes = [c_input_type, c_input_type, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_size_t]

            X = call_gsl_integration.call_gsl_integration(angles, average_deviation_ring, angles[0], angles[int(len(angles)) -1], 1.e-30, rel_err_gsl, int(len(angles)))
            integral=float(X)

            average_deviation[i]=integral/(2*np.pi)

        return average_deviation


    def calc_Christoffel_Symbols_cart(self, component, pot, scale_factor, scale_factor_dot, slope_dt, slope_dx, slope_dy, slope_dz):
        """
        This function calculates the Christoffel symbols in cartesian coordinates.

        Parameter
        ------------

        returns
        ------------

        """

        Hubble = scale_factor_dot/scale_factor # 1/Myrs avoid problems with dimensions
        #print("Hubble: ", Hubble)

        #r=r*convert_Mpc_km
        if(component[0] != component[1] != component[2] != component[0]):

            return 0.0

        ## Assume perturbations are small
        elif(component[0] == 0):
            if(component[1] == component[2] == 0):
                return slope_dt
            elif((component[1] == 1 and component[2] == 0) or (component[1] == 0 and component[2] == 1)):
                return slope_dx
            elif((component[1] == 2 and component[2] == 0) or (component[1] == 0 and component[2] == 2)):
                return slope_dy
            elif((component[1] == 3 and component[2] == 0) or (component[1] == 0 and component[2] == 3)):
                return slope_dz
            elif(component[1] == component[2] == 1):
                return speed_of_light_Mpc_over_Myr**(-2) * scale_factor**2 * (Hubble - 4*Hubble*pot - slope_dt)
            elif(component[1] == component[2] == 2):
                return speed_of_light_Mpc_over_Myr**(-2) * scale_factor**2 * (Hubble - 4*Hubble*pot - slope_dt)
            elif(component[1] == component[2] == 3):
                return speed_of_light_Mpc_over_Myr**(-2) * scale_factor**2 * (Hubble - 4*Hubble*pot - slope_dt)

        elif(component[0] == 1):
            if(component[1] == component[2] == 0):
                return speed_of_light_Mpc_over_Myr**2 * 1/(scale_factor**2) * slope_dx
            elif((component[1] == 1 and component[2] == 0) or (component[1] == 0 and component[2] == 1)):
                return Hubble - slope_dt
            elif((component[1] == 1 and component[2] == 2) or (component[1] == 2 and component[2] == 1)):
                return -slope_dy
            elif((component[1] == 1 and component[2] == 3) or (component[1] == 3 and component[2] == 1)):
                return -slope_dz
            elif(component[1] == component[2] == 1):
                return -slope_dx
            elif(component[1] == component[2] == 2):
                return slope_dx
            elif(component[1] == component[2] == 3):
                return slope_dx

        elif(component[0] == 2):
            if(component[1] == component[2] == 0):
                return speed_of_light_Mpc_over_Myr**2 * 1/(scale_factor**2) * slope_dy
            elif((component[1] == 2 and component[2] == 0) or (component[1] == 0 and component[2] == 2)):
                return Hubble - slope_dt
            elif((component[1] == 1 and component[2] == 2) or (component[1] == 2 and component[2] == 1)):
                return -slope_dx
            elif((component[1] == 2 and component[2] == 3) or (component[1] == 3 and component[2] == 2)):
                return -slope_dz
            elif(component[1] == component[2] == 1):
                return slope_dy
            elif(component[1] == component[2] == 2):
                return -slope_dy
            elif(component[1] == component[2] == 3):
                return slope_dy

        elif(component[0] == 3):
            if(component[1] == component[2] == 0):
                return speed_of_light_Mpc_over_Myr**2 * 1/(scale_factor**2) * slope_dz
            elif((component[1] == 3 and component[2] == 0) or (component[1] == 0 and component[2] == 3)):
                return Hubble - slope_dt
            elif((component[1] == 3 and component[2] == 1) or (component[1] == 1 and component[2] == 3)):
                return -slope_dx
            elif((component[1] == 3 and component[2] == 2) or (component[1] == 2 and component[2] == 3)):
                return -slope_dy
            elif(component[1] == component[2] == 1):
                return slope_dz
            elif(component[1] == component[2] == 2):
                return slope_dz
            elif(component[1] == component[2] == 3):
                return -slope_dz


    def calc_weyl_cart(self, component, pot, slope_dt, slope_dx, slope_dy, slope_dz, pot_time_curve, pot_x_curve, pot_y_curve, pot_z_curve, dxdy, dxdz, dydx, dydz, dzdx, dzdy, dtdx, dtdy, dtdz, scale_factor_dot, dyad_basis1, dyad_basis2, redshift, scale_factor, distance):
        """
        This function calculates the Weyl tensor following eq 3.21 of
        Sasaki 1993 "Cosmological Gravitational Lens Equation".

        Parameter
        ------------

        returns
        ------------

        """
        # TODO: calculate the proper redshift to the position based on the radius r
        # Until now we use the redhsift from the observer to the centre of the void.
        # The default setting is z = 1000 thus the current approximation should be
        # sufficient for the moment.
        scale = 1/(1+redshift)

        diag = 0

        #print("Calculate Weyl:")
        #print("Components: ", component[0], component[1], component[2], component[3])

        first_pot_derivs = np.array([slope_dt, slope_dx, slope_dy, slope_dz])

        second_pot_derivs = np.empty((4,4))
        second_pot_derivs[0,0] = pot_time_curve
        second_pot_derivs[1,1] = pot_x_curve
        second_pot_derivs[2,2] = pot_y_curve
        second_pot_derivs[3,3] = pot_z_curve
        second_pot_derivs[0,1] = dtdx
        second_pot_derivs[0,2] = dtdy
        second_pot_derivs[0,3] = dtdz
        second_pot_derivs[1,0] = dtdx
        second_pot_derivs[1,2] = dxdy
        second_pot_derivs[1,3] = dxdz
        second_pot_derivs[2,0] = dtdy
        second_pot_derivs[2,1] = dydx
        second_pot_derivs[2,3] = dydz
        second_pot_derivs[3,0] = dtdz
        second_pot_derivs[3,1] = dzdx
        second_pot_derivs[3,2] = dzdy

        basis_vecs = np.array((dyad_basis1,dyad_basis2))

        #print("first_pot_derivs=",first_pot_derivs)
        #print("second_pot_derivs=",second_pot_derivs)
        #print("basis_vecs=",basis_vecs)


        if( component[1] != component[3] != 0 ):
            print("Error: Missing formula")
            return 0.0
        else:
            Christoffel_symbols = np.zeros((4,4,4))

            for i in range(0,4):
                for l in range(0,4):
                    for m in range(0,4):
                        #Christoffel_symbols[i,l,m] = self.calc_Christoffel_Symbols_cart([i,l,m], pot, scale_factor[int(Nsnaps_analyse-1)], scale_factor_dot, slope_dt, slope_dx, slope_dy, slope_dz)
                        Christoffel_symbols[i,l,m] = self.calc_Christoffel_Symbols_cart([i,l,m], pot, scale_factor[3], scale_factor_dot, slope_dt, slope_dx, slope_dy, slope_dz)
                        #print("i l m", i, l, m, "  Christoffel_symbols[i,l,m]: ",Christoffel_symbols[i,l,m])

            off_diag = 0
            diag = 0

            for mu in range(0,4):
                for nu in range(0,4):
                    # Note that the dimensions of the Christoffel symbols vary for the different elements.
                    # They are always multiplied with a entry of the first_pot_derivs vector that will
                    # result in 1/Mpc^2 as dimension
                    sum_christoffel = np.sum(Christoffel_symbols[:,mu,nu] * first_pot_derivs[:])
                    off_diag += 2 * (second_pot_derivs[mu,nu] - sum_christoffel) * basis_vecs[(component[0]-1),mu] * basis_vecs[(component[2]-1),nu]
                    if(component[0] == component[2]):
                        diag += ( (second_pot_derivs[mu,nu] - sum_christoffel)
                                       * (basis_vecs[0,mu] * basis_vecs[0,nu]
                                           + basis_vecs[1,mu] * basis_vecs[1,nu]) )

            WeylComponent = (off_diag - diag)*(1+redshift)**2

            return WeylComponent, diag


    def calc_ricci(self, local_den, redshift, mass_dm, Hubble_h):
        """
        This function calculates the Ricci tensor following eq 3.22 of
        Sasaki 1993 "Cosmological Gravitational Lens Equation".

        Parameter
        ------------

        returns
        ------------

        """


        #print("G=",G)
        #print("mass_dm=",mass_dm)
        #print("M_sun_in_kg=",M_sun_in_kg)
        #print("local_den=",local_den)

        #density in 1/(Mpc/h)**3
        #Is the 0.7 (h) correct to multiply here?
        local_den=local_den/(convert_Mpc_km**3)*Hubble_h**2 # already take the h from mass_dm into account!

        #print("8*np.pi*G/(mass_dm * 1.e10 * M_sun_in_kg)*local_den*(1+redshift)**2=",8*np.pi*G/(mass_dm * 1.e10 * M_sun_in_kg)*local_den*(1+redshift)**2)

        Myr_in_s = 31557600000000

        a=8*np.pi*G*(mass_dm * 1.e10 * M_sun_in_kg) # km^3/(h s^2)
        #print("a: ",a)
        b=a*local_den # 1/s^2
        #print("b: ",b)
        c=b*(Myr_in_s**2) # 1/Myr^2
        #print("c: ",c)

        #return 8*np.pi*G*(mass_dm * 1.e10 * M_sun_in_kg)*local_den*(1+redshift)**2*(Myr_in_s**2)

        # unit: 1/Myr^2
        return c*(1+redshift)**2


    def solve_diff_cart(self, radii_comoving, axis1, axis2, void_position_from_observer, redshift, mass_dm, Nfiles, snap_times, scale_factors, dyad_basis1, dyad_basis2, mean_dens, pot_time_slope, pot_time_curve, dtdy, dtdx, dtdz, scale_factor_dot, pot_y_slope, pot_y_curve, dydx, dydz, pot_x_slope, pot_x_curve, dxdy, dxdz, pot_z_slope, pot_z_curve, dzdy, dzdx, rho, pot, box_length, Hubble_h):
        """
        This function solves the system of differential equations defined in
        3.16 and 3.17 of Sasaki 1993 "Cosmological Gravitational Lens Equation".

        Parameter
        ------------

        returns
        ------------

        """

        print("-------------------")
        print("solve_diff")

        C1010 = np.zeros(len(radii_comoving))
        C1020 = np.zeros(len(radii_comoving))
        diag1010 = np.zeros(len(radii_comoving))
        diag1020 = np.zeros(len(radii_comoving))
        R44 = np.zeros(len(radii_comoving))

        # all derivative are taken with respect to 1/(Mpc/h) in the coordinates
        dtdy=dtdy*Hubble_h
        dtdx=dtdx*Hubble_h
        dtdz=dtdz*Hubble_h
        pot_y_slope=pot_y_slope*Hubble_h**2
        pot_y_curve=pot_y_curve*Hubble_h**2
        dydx=dydx*Hubble_h**2
        dydz=dydz*Hubble_h**2
        pot_x_slope=pot_x_slope*Hubble_h**2
        pot_x_curve=pot_x_curve*Hubble_h**2
        dxdy=dxdy*Hubble_h**2
        dxdz=dxdz*Hubble_h**2
        pot_z_slope=pot_z_slope*Hubble_h**2
        pot_z_curve=pot_z_curve*Hubble_h**2
        dzdy=dzdy*Hubble_h**2
        dzdx=dzdx*Hubble_h**2

        i=0
        distance=void_position_from_observer[1]
        for t in radii_comoving:

            local_dens = rho[i,axis1,axis2]
            R44[i] = (speed_of_light_Mpc_over_Myr)**(-2) * self.calc_ricci(local_dens, redshift, mass_dm, Hubble_h) # 1/Mpc^2

            # TODO, redshift is a function of radius
            C1010[i], diag1010[i] = self.calc_weyl_cart([1,0,1,0], pot[i,axis1,axis2], pot_time_slope[i,axis1,axis2], pot_x_slope[i,axis1,axis2], pot_y_slope[i,axis1,axis2], pot_z_slope[i,axis1,axis2], pot_time_curve[i,axis1,axis2], pot_x_curve[i,axis1,axis2], pot_y_curve[i,axis1,axis2], pot_z_curve[i,axis1,axis2], dxdy[i,axis1,axis2], dxdz[i,axis1,axis2], dydx[i,axis1,axis2], dydz[i,axis1,axis2], dzdx[i,axis1,axis2], dzdy[i,axis1,axis2], dtdx[i,axis1,axis2], dtdy[i,axis1,axis2], dtdz[i,axis1,axis2], scale_factor_dot, dyad_basis1, dyad_basis2, redshift, scale_factors, distance)
            C1020[i], diag1010[i] = self.calc_weyl_cart([1,0,2,0], pot[i,axis1,axis2], pot_time_slope[i,axis1,axis2], pot_x_slope[i,axis1,axis2], pot_y_slope[i,axis1,axis2], pot_z_slope[i,axis1,axis2], pot_time_curve[i,axis1,axis2], pot_x_curve[i,axis1,axis2], pot_y_curve[i,axis1,axis2], pot_z_curve[i,axis1,axis2], dxdy[i,axis1,axis2], dxdz[i,axis1,axis2], dydx[i,axis1,axis2], dydz[i,axis1,axis2], dzdx[i,axis1,axis2], dzdy[i,axis1,axis2], dtdx[i,axis1,axis2], dtdy[i,axis1,axis2], dtdz[i,axis1,axis2], scale_factor_dot, dyad_basis1, dyad_basis2, redshift, scale_factors, distance)
            # both in 1/Mpc^2

            #print("R44[i]=", R44[i], "C1010[i]=", C1010[i], "C1020[i]=",C1020[i])

            i=i+1

        #print("R44:", R44)
        #print("C1010:", C1010)

        c_input_type = ndpointer(dtype=ctypes.c_double, flags="C_CONTIGUOUS")

        optical_scalar_solver.optical_scalars.argtypes = [c_input_type,
                                                          c_input_type, c_input_type,
                                                          c_input_type, ctypes.c_double, ctypes.c_double]
        #optical_scalar_solver.optical_scalars_check_distance.argtypes = [c_input_type,
        #                                                  c_input_type, c_input_type,
        #                                                  c_input_type, ctypes.c_double, ctypes.c_double]
        #optical_scalar_solver.only_expansion.argtypes = [c_input_type,
        #                                                 c_input_type, c_input_type,
        #                                                 c_input_type, ctypes.c_double, ctypes.c_double]
        optical_scalar_solver.optical_scalars.restype = ndpointer(dtype=ctypes.c_double, shape=(3,1))
        #optical_scalar_solver.optical_scalars_check_distance.restype = ndpointer(dtype=ctypes.c_double, shape=(4,1))
        #optical_scalar_solver.only_expansion.restype = ctypes.c_double

        r_comoving_max = (distance + 0.95*(box_length/2))/Hubble_h # Mpc
        r_comoving_min = (distance - 0.95*(box_length/2))/Hubble_h # Mpc
        # this is in 3 places!!!
        rad_comoving=np.linspace(0,r_comoving_max-r_comoving_min,len(radii_comoving))

        X = optical_scalar_solver.optical_scalars(rad_comoving, R44, C1010, C1020, rad_comoving[int(len(radii_comoving)-1)], rad_comoving[0])

        #X = optical_scalar_solver.optical_scalars(rad, R44, C1010, C1020, rad[0], rad[19])
        #X = optical_scalar_solver.optical_scalars_check_distance(rad, R44, C1010, C1020, rad[0], rad[19])
        #X = optical_scalar_solver.only_expansion(rad, R44, C1010, C1020, rad[0], rad[19])
        #print("X:", X)

#        diff_sol = np.loadtxt('diff_solution.dat')
#
#        plt.plot(radii_comoving, C1010, 'r-', label='C1010')
#        plt.plot(radii_comoving, C1020  , 'b-', label='C1020')
#        plt.grid()
#        plt.legend(loc='best')
#        plt.xlabel('radius')
#        plt.ylabel('C1010 and C1020')
#        figname="Weyl_tensor_line" + ".eps"
#        plt.savefig(figname, format = 'eps')
#        plt.close()
#
#        plt.plot(radii_comoving, R44, 'r-', label='R44')
#        plt.grid()
#        plt.legend(loc='best')
#        plt.xlabel('radius')
#        plt.ylabel('R44')
#        figname="R44" + ".eps"
#        plt.savefig(figname, format = 'eps')
#        plt.close()
#
#        plt.plot(radii_comoving, R44, 'r-', label='R44')
#        plt.plot(radii_comoving, C1010, 'g-', label='C1010')
#        plt.plot(radii_comoving, C1020  , 'b-', label='C1020')
#        plt.grid()
#        plt.legend(loc='best')
#        plt.xlabel('radius')
#        plt.ylabel('R44')
#        figname="Weyl_vs_R44" + ".eps"
#        plt.savefig(figname, format = 'eps')
#        plt.close()


        return X



    def avg_smoothing(self, array):

        Nelements = array.size
        avg = np.empty(Nelements)

        # calc 5 element median for most of the elements

        for i in range(0,Nelements):
            if(i==0):
                avg[i] = array[i]
            if(i==1):
                avg[i] = (array[i-1] + array[i] + array[i+1])/3
            if( (i >= 2) and (i <= Nelements-3) ):
                avg[i] = (array[i-2] + array[i-1] + array[i] + array[i+1] + array[i+2])/5
            if(i== Nelements-2):
                avg[i] = (array[i-1] + array[i] + array[i+1])/3
            if(i== Nelements-1):
                avg[i] = array[i]


        return avg
