#!/usr/bin/env python

# lensing_calculations_unit_testable - geometrical optics analysis unit testable functions
# (C) 2021, 2022 Marius Peper, Boud Roukema GPL-3+
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

import numpy as np
import scipy

class lensing_calc_unit_testable:

    # Find local minima in array2d and apply one or two other validation
    # criteria (currently: the pixel value must be below
    # tolerance_frac_mean times the mean).
    def local_minima(array2d, mean, std, N_pixels, N_neighbours, N_significance, tolerance_frac_std, tolerance_frac_mean):
        # logical array initialised with all pixels as acceptable
        pre_centres=np.ones((N_pixels,N_pixels),dtype=bool)

        # Do a de facto boolean 'and' operation (multiply by 1) for every pixel
        # where that pixel has a lower value than all (N_neighbours^2-1) of its
        # neighbours, with a T^2 topology.
        for i in range(1, N_neighbours+1):
            for j in range(1, N_neighbours+1):
                pre_centres = (pre_centres*
                               ((array2d < np.roll(array2d,  i, 0)) &
                                (array2d < np.roll(array2d, -i, 0)) &
                                (array2d < np.roll(array2d,  j, 1)) &
                                (array2d < np.roll(array2d, -j, 1)) &
                                (array2d < np.roll(np.roll(array2d,i,0),  j, 1)) &
                                (array2d < np.roll(np.roll(array2d,-i,0), -j, 1)) &
                                (array2d < np.roll(np.roll(array2d,i,0),  -j, 1)) &
                                (array2d < np.roll(np.roll(array2d,-i,0), j, 1))) )

        # Currently not used:
        significant_centres = (((array2d - np.roll(array2d,  N_significance, 0)) < -tolerance_frac_std*std) &
                               ((array2d - np.roll(array2d, -N_significance, 0)) < -tolerance_frac_std*std) &
                               ((array2d - np.roll(array2d,  N_significance, 1)) < -tolerance_frac_std*std) &
                               ((array2d - np.roll(array2d, -N_significance, 1)) < -tolerance_frac_std*std))

        # Evaluate a de facto boolean array that tests if a pixel is lower
        # than tolerance_frac_mean*mean.
        #
        # TODO: Why evaluate this four times? Is this a remnant of a
        # copy/paste from the unused array 'significant_centres'?
        pixel_is_well_below_mean = (array2d < tolerance_frac_mean*mean)

        return pre_centres*pixel_is_well_below_mean #*significant_centres


    # Merge centres of candidate photometrically detected voids if they are
    # too close to each other in the sky plane.
    #
    # pre_centre = np.array of pairs of [x,z] centres of voids
    # rads = np.array of radii of voids
    # merge_distance = minimum distance for a pair of voids to not be merged
    # N_pixels = side length in number of pixels of the fundamental domain
    def merge_centres(pre_centre, rads, merge_distance, N_pixels):

        centres=[]
        new_radii=[]
        merge_index=[]
        print("merge_centres:: pre_centre= ", pre_centre)
        print("merge_centres:: rads= ", rads)
        print("merge_centres:: merge_distance= ",merge_distance)

        # Check for empty or unequal length lists; return trivially if the input is empty.
        if(len(pre_centre) != len(rads)):
            print("merge_centres: WARNING: len(pre_centre) = %d != len(rads) = %d" %
                  len(pre_centre), len(rads))

        if(len(pre_centre) == 0):
            print("merge_centres: WARNING: len(pre_centre) = %d" % len(pre_centre))
            return pre_centre, rads # return the two arrays unchanged

        # Create 'merge_index', a list of pairs of indices where each
        # individual index i is necessarily listed with itself at least
        # once as [i,i], and pairs that should be merged are each listed as
        # [i,j].
        for i in range(0, len(pre_centre)):
            for j in range(i, len(pre_centre)):
                distance=np.sqrt(np.amin([ (pre_centre[i,0]-pre_centre[j,0])**2,
                                           (N_pixels+pre_centre[i,0]-pre_centre[j,0])**2,
                                           (N_pixels+pre_centre[j,0]-pre_centre[i,0])**2 ]) +
                                 np.amin([ (pre_centre[i,1]-pre_centre[j,1])**2,
                                           (N_pixels+pre_centre[i,1]-pre_centre[j,1])**2,
                                           (N_pixels+pre_centre[j,1]-pre_centre[i,1])**2 ]) )

                # assume only 2 centres are close to each other
                if(((distance < rads[i]/4) and (distance < rads[j]/4)) or (distance < merge_distance)):
                    merge_index.append([i,j])

        # Convert 'merge_index' to a numpy array and create an index
        # 'to_merge' to indicate the unequal pairs, i.e. those to be merged.
        merge_index=np.array(merge_index)
        to_merge = np.where(merge_index[:,1]!=merge_index[:,0])[0]
        merge_centre_index = merge_index[to_merge]
        print("Prior to any merging, merge_index: ", merge_index)

        # Create 'non_merge_index', which aims to be a list of pairs of
        # void indices where those voids are *not* to be merged. Start
        # with a copy of the whole list and remove the pairs [i,i] and
        # [j,j] for each mergeable pair [i,j].
        non_merge_index = merge_index.copy()
        for i in range(0,len(to_merge)):
            where_first_index = np.where(non_merge_index[:,0]==merge_index[to_merge[i],0])
            non_merge_index=np.delete(non_merge_index, where_first_index, axis=0)
            where_second_index = np.where(non_merge_index[:,1]==merge_index[to_merge[i],1])
            non_merge_index=np.delete(non_merge_index, where_second_index, axis=0)

        print("After deletion, non_merge_index: ", non_merge_index)

        # Concatenate the two lists so that the non-merged voids [i,i] are
        # at the beginning, and the to-be-merged pairs are at the end.
        merge_index=np.concatenate((non_merge_index, merge_centre_index), axis=0)
        print("After concatenation, merge_index: ", merge_index)

        # From a graph theory point of view, 'merge_index' now contains a
        # list of edges. The following section aims to make a list of all
        # disconnected subgraphs. Each subgraph indicates a list of voids
        # to be merged together.
        multi_merge_lines=[]
        list_indices=np.array([])
        for this_void_index in merge_index:
            print("this_void_index: ", this_void_index)
            if(np.isin(this_void_index[0], list_indices) ):
                j=0
                while( np.where(this_void_index[0]==multi_merge_lines[j])[0].size == 0 ):
                    j=j+1
                where=np.where(this_void_index[0]==multi_merge_lines[j])[0][0]
                multi_merge_lines[j].append(this_void_index[1])
            else:
                if(np.isin(this_void_index[1], list_indices) ):
                    j=0
                    while( np.where(this_void_index[1]==multi_merge_lines[j])[0].size == 0 ):
                        j=j+1
                    where=np.where(this_void_index[1]==multi_merge_lines[j])[0][0]
                    multi_merge_lines[j].append(this_void_index[0])
                else:
                    multi_merge_lines.append(list(this_void_index))
            list_indices=np.append(list_indices,this_void_index)

        # Use the list of disconnected subgraphs to merge the voids'
        # centres and radii.
        print("multi_merge_lines: ", multi_merge_lines)
        for this_void_index in multi_merge_lines:
            new_x=0
            new_z=0
            new_rad=0
            print("this_void_index: ", this_void_index)
            for i in range(0,len(this_void_index)):
                print("pre_centre[this_void_index[i],0]: ", pre_centre[this_void_index[i],0])
                print("pre_centre[this_void_index[i],1]: ", pre_centre[this_void_index[i],1])
                new_x=new_x+pre_centre[this_void_index[i],0]
                new_z=new_z+pre_centre[this_void_index[i],1]
                new_rad=new_rad+rads[this_void_index[i]]
            new_centre=[new_x/len(this_void_index) , new_z/len(this_void_index)]
            new_radii.append(new_rad/len(this_void_index))
            centres.append(new_centre)

        print("centres: ", centres)

        return np.array(centres), np.array(new_radii)


    # Calculate the probability that a 3D void is consistent with a photometric (sky plane) void
    # N_pixels = side length in number of pixels of the fundamental domain
    #
    # This routine does not check for array sizes. It *should* work if either:
    #
    # R_2D, x_2D, z_2D are scalar and R_3D, x_3D, z_3D  are all arrays of a single size; or
    #
    # R_3D, x_3D, z_3D are scalar and R_2D, x_2D, z_2D  are all arrays of a single size.
    #
    # The routine is expected to *fail* if both sets of arrays are non-scalar (longer than one).
    def void_probability(R_3D, x_3D, z_3D, std_log10radius, std_centre, R_2D, x_2D, z_2D, N_pixels):

        use_crude_hack = False
        if(not(use_crude_hack)):
            # calculate T^2 distances in x and z; and normalise
            d_x = np.amin([np.abs(x_3D-x_2D), np.abs(N_pixels + x_3D-x_2D), np.abs(N_pixels + x_2D-x_3D)], axis=0) /std_centre
            d_z = np.amin([np.abs(z_3D-z_2D), np.abs(N_pixels + z_3D-z_2D), np.abs(N_pixels + z_2D-z_3D)], axis=0) /std_centre
            # calculate R disagreement
            d_R = np.abs(np.log10(R_2D/R_3D)) /std_log10radius

            # Two-sided (cumulative) Gaussian probabilities of being d_x or
            # further from the model position.
            #
            # scipy.special.erfc(d_x/np.sqrt(2.0)) is mathematically the
            # same as 1 - 2 * (scipy.stats.norm.cdf(d_x) - 0.5), though
            # the numerical implementations are not quite identical.
            x_prob = scipy.special.erfc(d_x/np.sqrt(2.0))
            z_prob = scipy.special.erfc(d_z/np.sqrt(2.0))
            R_prob = scipy.special.erfc(d_R/np.sqrt(2.0))

            prob = x_prob*z_prob*R_prob
            prob_max = np.amax(prob)
            i_highest_prob = (np.where(prob == prob_max))[0][0]
        
        else:
            # old:
            # normalise to 1
            R_prob=scipy.stats.norm(R_3D,std_radius).pdf(R_2D) * std_radius*(np.sqrt(2*np.pi))
            x_prob=scipy.stats.norm(x_3D,std_centre).pdf(x_2D) * std_centre*(np.sqrt(2*np.pi))
            z_prob=scipy.stats.norm(z_3D,std_centre).pdf(z_2D) * std_centre*(np.sqrt(2*np.pi))
            prob = x_prob*z_prob*R_prob
            
        return prob, i_highest_prob


    # Smooth a square array with a lowpass filter
    # 2023-01-13: only the case for a 5x5 array is currently implemented
    def lowpass_filter(neighbours):
        #lowpass = np.array([[1,2,3,2,1], [2,3,4,3,2], [3,4,5,4,3], [2,3,4,3,2], [1,2,3,2,1]])
        lowpass = np.array([[1,4,6,4,1], [4,16,24,16,4], [6,24,36,24,6], [4,16,24,16,4], [1,4,6,4,1]])
        s = np.sum(lowpass)

        if((np.shape(neighbours)[0]==5) and (np.shape(neighbours)[1]==5)):
            return np.sum(neighbours * lowpass) / s # sum weighted by the lowpass filter
        else:
            return np.mean(neighbours)
