#!/usr/bin/env python

# plot_voidgals_properties - galaxies' relationship with their void
# (C) 2020 Boud Roukema, Marius Peper GPL-3+
#
# Fit and plot dependence of galaxies' infall history properties
# in relation to void properties such as how much of the galaxy's
# host halo consists of void particles and the elaphrocentric distance
# of the host halo.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

import sys
import os

import matplotlib
import matplotlib.pyplot as plt
#import matplotlib.cm as cm
import matplotlib.markers as markers

import numpy as np


debug = False

def void_props_read_data(voidprops_filename):

    # Read matched void properties
    R3D = np.loadtxt(voidprops_filename, usecols=0, dtype=np.double)
    x3D = np.loadtxt(voidprops_filename, usecols=1, dtype=np.double)
    z3D = np.loadtxt(voidprops_filename, usecols=2, dtype=np.double)

    R2D = np.loadtxt(voidprops_filename, usecols=3, dtype=np.double)
    x2D = np.loadtxt(voidprops_filename, usecols=4, dtype=np.double)
    z2D = np.loadtxt(voidprops_filename, usecols=5, dtype=np.double)

    prob_match = np.loadtxt(voidprops_filename, usecols=6, dtype=np.double)


    N_voids = (R3D.shape)[0]
    print('N_voids=', N_voids)

    return (N_voids, R3D, x3D, z3D, R2D, x2D, z2D, prob_match)



def do_plot(detector_name, # filename string for detector parameter
            detector_label, # LaTeX string for detector parameter
            outfile_dir, # directory for output postscript files
            given, # string  'given_2D' or 'given_3D'
            N_voids, R3D, x3D, z3D, R2D, x2D, z2D, prob_match):

    ## Possible TODO: sanitize 'outfile_dir' before writing there.
    ## If only run locally, risk is low.

    ## x,z limits
    xlims = (xz_min, xz_max) # 2-tuple
    ylims = xlims

    # x Plot
    plt.clf()
    plt.rc('font',size=voidmatch_fontsize, family='serif')
    fig, axs = plt.subplots()

    plt.plot(xlims, ylims, '-k', linewidth=1)
    plt.plot(x3D, x2D,
             'o', linewidth=1, markersize=5, label=("$" + detector_label + "$"))

    axs.set_xlim(xlims)
    axs.set_ylim(ylims)
    #axs.axis('equal')
    axs.set_aspect('equal','box')
    plt.xlabel(r'$x_{\mathrm{3D}} (\mathrm{Mpc}/h)$')
    plt.ylabel(r'$x_{\mathrm{2D}} (\mathrm{Mpc}/h)$')

    plt.tight_layout()
    plt.savefig(outfile_dir + '/' + detector_name+'_'+given+'_x_corr.eps', metadata=metadata)
    plt.close()

    # z Plot
    plt.clf()
    plt.rc('font',size=voidmatch_fontsize, family='serif')
    fig, axs = plt.subplots()

    plt.plot(xlims, ylims, '-k', linewidth=1)
    plt.plot(z3D, z2D,
             'o', linewidth=1, markersize=5, label=("$" + detector_label + "$"))

    axs.set_xlim(xlims)
    axs.set_ylim(ylims)
    #axs.axis('equal')
    axs.set_aspect('equal','box')
    plt.xlabel(r'$z_{\mathrm{3D}} (\mathrm{Mpc}/h)$')
    plt.ylabel(r'$z_{\mathrm{2D}} (\mathrm{Mpc}/h)$')

    plt.tight_layout()
    plt.savefig(outfile_dir + '/' + detector_name+'_'+given+'_z_corr.eps', metadata=metadata)
    plt.close()


    # Radii Plot
    xlims = (5.0, 30.0) # 2-tuple
    ylims = xlims

    plt.clf()
    plt.rc('font',size=voidmatch_fontsize, family='serif')
    fig, axs = plt.subplots()

    plt.plot(xlims, ylims, '-k', linewidth=1)
    plt.plot(R3D, R2D,
             'o', linewidth=1, markersize=5, label=("$" + detector_label + "$"))

    axs.set_xlim(xlims)
    axs.set_ylim(ylims)
    #axs.axis('equal')
    axs.set_aspect('equal','box')
    plt.xlabel(r'$R_{\mathrm{3D}} (\mathrm{Mpc}/h)$')
    plt.ylabel(r'$R_{\mathrm{2D}} (\mathrm{Mpc}/h)$')

    plt.tight_layout()
    plt.savefig(outfile_dir + '/' + detector_name+'_'+given+'_R_corr.eps', metadata=metadata)
    plt.close()



if __name__ == '__main__':

    pass_val = 0

    voidmatch_fontsize = 21

    if os.environ.get('METADATA_CREATOR'):
        metadata_creator = os.environ.get('METADATA_CREATOR')
    else:
        metadata_creator = 'Peper, Boljeko, Roukema 2023'
    metadata = {"Creator": metadata_creator}

    if os.environ.get('N_PIXELS'):
        N_pixels = int(os.environ.get('N_PIXELS'))
    else:
        N_pixels = 120

    if os.environ.get('INFILE_DIR'):
        infile_dir = os.environ.get('INFILE_DIR')
    else:
        infile_dir = '.'

    if os.environ.get('OUTFILE_DIR'):
        outfile_dir = os.environ.get('OUTFILE_DIR')
    else:
        outfile_dir = '.'

    xz_min = -N_pixels/2.0
    xz_max = N_pixels/2.0


    detector_names = np.array(["mass_def", "gamma", "exp", "sig"])
    detector_labels = np.array(["\\Sigma", "\\gamma", "\\theta", "\\vert\\sigma\\vert"])

    for i in range(0,4):
        #for given in ["given_2D", "given_3D"]:
        for given in ["given_3D"]:
            voidprops_filename = infile_dir + "/" + "void_matches_" + detector_names[i] + "_" + given + ".dat"
            N_voids, R3D, x3D, z3D, R2D, x2D, z2D, prob_match = void_props_read_data(voidprops_filename)
            do_plot(detector_names[i],
                    detector_labels[i],
                    outfile_dir,
                    given,
                    N_voids, R3D, x3D, z3D, R2D, x2D, z2D, prob_match)



    print("pass_val = ",pass_val)

    sys.exit(pass_val)
