import os
import time
import numpy as np
import ctypes
read_gadget = ctypes.CDLL("io_gadget.so")
from numpy.ctypeslib import ndpointer
from os.path import getsize as getFileSize

speed_of_light = 299792.458 # in km/s

class Read:

    def read_tracers(self, tracer_file):
        """
        The routine reads in the positions of the DM particles (the tracers)
        from the simulation. To read in the data the routine will call a
        function that is written in C. This routine only works on simulated
        data that is provided in the gadget format.

        Parameter
        ------------
        tracer_file : string
        String that contains the path and filename of the simulation file we
        want to use.

        returns
        ------------
        tracers : np.array(3, dtype=float)
        The x,y,z coordinates of all DM particles in our simulation.
        """

        print("read tracers")
        print("tracer_file=",tracer_file)
        read_gadget.get_particles.argtypes = [ctypes.c_char_p]
        read_gadget.get_particles.restype = ctypes.c_int
        num_tracers=read_gadget.get_particles(tracer_file.encode('utf-8'))
        tracers = np.empty((num_tracers, 3))
        returnValue=np.empty((num_tracers, 3))

        read_gadget.read_pos.argtypes = [ctypes.c_char_p, ctypes.c_int]
        read_gadget.read_pos.restype = ndpointer(dtype=ctypes.c_float, shape=(num_tracers,3))

        returnValue=read_gadget.read_pos(tracer_file.encode('utf-8'),num_tracers)

        tracers[:,0]=returnValue[:,0]
        tracers[:,1]=returnValue[:,1]
        tracers[:,2]=returnValue[:,2]

        #positions have Mpc/h
        return tracers

    def read_galpos(self, gal_pos_file):

        galaxy_data =  np.loadtxt(gal_pos_file)
        num_gals=len(galaxy_data)
        gal_pos = np.empty((num_gals,3))
        gal_pos = galaxy_data

        return galaxy_data

        

    def read_vels(self, tracer_file):
        """
        The routine reads in the velocities of the DM particles from the
        simulation. To read in the data the routine will call a function
        that is written in C. This routine only works on simulated
        data that is provided in the gadget format.

        Parameter
        ------------
        tracer_file : string
        String that contains the path and filename of the simulation file
        we want to use.

        returns
        ------------
        vels : np.array(dtype=float)
        The current velocities of the DM particles.
        """

        read_gadget.get_particles.argtypes = [ctypes.c_char_p]
        read_gadget.get_particles.restype = ctypes.c_int

        num_tracers=read_gadget.get_particles(tracer_file.encode('utf-8'))
        print("num_tracers=",num_tracers)

        vels = np.empty((num_tracers, 3))
        returnValue=np.empty((num_tracers, 3))

        read_gadget.read_vel.argtypes = [ctypes.c_char_p, ctypes.c_int]
        read_gadget.read_vel.restype = ndpointer(dtype=ctypes.c_float, shape=(num_tracers,3))

        returnValue=read_gadget.read_vel(tracer_file.encode('utf-8'),num_tracers)

        vels[:,0]=returnValue[:,0]
        vels[:,1]=returnValue[:,1]
        vels[:,2]=returnValue[:,2]

        #velocities have km/s
        return vels

    def read_gals(self, model_name, first_file, last_file):
        """
        Read in the data about all galaxies generated by sage.
        """

        # The input galaxy structure:
        Galdesc_full = [
            ('SnapNum'                      , np.int32),
            ('Type'                         , np.int32),
            ('GalaxyIndex'                  , np.int64),
            ('CentralGalaxyIndex'           , np.int64),
            ('SAGEHaloIndex'                , np.int32),
            ('SAGETreeIndex'                , np.int32),
            ('SimulationHaloIndex'          , np.int64),
            ('mergeType'                    , np.int32),
            ('mergeIntoID'                  , np.int32),
            ('mergeIntoSnapNum'             , np.int32),
            ('dT'                           , np.float32),
            ('Pos'                          , (np.float32, 3)),
            ('Vel'                          , (np.float32, 3)),
            ('Spin'                         , (np.float32, 3)),
            ('Len'                          , np.int32),
            ('Mvir'                         , np.float32),
            ('CentralMvir'                  , np.float32),
            ('Rvir'                         , np.float32),
            ('Vvir'                         , np.float32),
            ('Vmax'                         , np.float32),
            ('VelDisp'                      , np.float32),
            ('ColdGas'                      , np.float32),
            ('StellarMass'                  , np.float32),
            ('BulgeMass'                    , np.float32),
            ('HotGas'                       , np.float32),
            ('EjectedMass'                  , np.float32),
            ('BlackHoleMass'                , np.float32),
            ('IntraClusterStars'            , np.float32),
            ('MetalsColdGas'                , np.float32),
            ('MetalsStellarMass'            , np.float32),
            ('MetalsBulgeMass'              , np.float32),
            ('MetalsHotGas'                 , np.float32),
            ('MetalsEjectedMass'            , np.float32),
            ('MetalsIntraClusterStars'      , np.float32),
            ('SfrDisk'                      , np.float32),
            ('SfrBulge'                     , np.float32),
            ('SfrDiskZ'                     , np.float32),
            ('SfrBulgeZ'                    , np.float32),
            ('DiskRadius'                   , np.float32),
            ('Cooling'                      , np.float32),
            ('Heating'                      , np.float32),
            ('QuasarModeBHaccretionMass'    , np.float32),
            ('TimeOfLastMajorMerger'         , np.float32),
            ('TimeOfLastMinorMerger'         , np.float32),
            ('OutflowRate'                  , np.float32),
            ('infallMvir'                   , np.float32),
            ('infallVvir'                   , np.float32),
            ('infallVmax'                   , np.float32)
            ]
        names = [Galdesc_full[i][0] for i in range(len(Galdesc_full))]
        formats = [Galdesc_full[i][1] for i in range(len(Galdesc_full))]
        Galdesc = np.dtype({'names':names, 'formats':formats}, align=True)


        # Initialize variables.
        TotNTrees = 0
        TotNGals = 0
        FileIndexRanges = []

        print("Determining array storage requirements.")

        # Read each file and determine the total number of galaxies to be read in
        goodfiles = 0
        for fnr in range(first_file,last_file+1):
            fname = model_name+'_'+str(fnr)  # Complete filename

            if not os.path.isfile(fname):
                # print "File\t%s  \tdoes not exist!  Skipping..." % (fname)
                continue

            if getFileSize(fname) == 0:
                print("File\t%s  \tis empty!  Skipping..." % (fname))
                continue

            fin = open(fname, 'rb')  # Open the file
            Ntrees = np.fromfile(fin,np.dtype(np.int32),1)  # Read number of trees in file
            NtotGals = np.fromfile(fin,np.dtype(np.int32),1)[0]  # Read number of gals in file.
            TotNTrees = TotNTrees + Ntrees  # Update total sim trees number
            TotNGals = TotNGals + NtotGals  # Update total sim gals number
            goodfiles = goodfiles + 1  # Update number of files read for volume calculation
            fin.close()

        print()
        print("Input files contain:\t%d trees ;\t%d galaxies ." % (TotNTrees, TotNGals))
        print()

        # Initialize the storage array
        G = np.empty(TotNGals, dtype=Galdesc)

        offset = 0  # Offset index for storage array

        # Open each file in turn and read in the preamble variables and structure.
        print("Reading in files.")
        for fnr in range(first_file,last_file+1):
            fname = model_name+'_'+str(fnr)  # Complete filename

            if not os.path.isfile(fname):
                continue

            if getFileSize(fname) == 0:
                continue

            print("read_gals: Will read: ",fname)
            fin = open(fname, 'rb')  # Open the file
            Ntrees = np.fromfile(fin, np.dtype(np.int32), 1)  # Read number of trees in file
            NtotGals = np.fromfile(fin, np.dtype(np.int32), 1)[0]  # Read number of gals in file.
            np.fromfile(fin, np.dtype((np.int32, Ntrees)),1) # skip over GalsPerTree = number of gals in each tree
            print(":   Reading N=", NtotGals, "   \tgalaxies from file: ", fname)
            GG = np.fromfile(fin, Galdesc, NtotGals)  # Read in the galaxy structures

            FileIndexRanges.append((offset,offset+NtotGals))

            # Slice the file array into the global array
            # The copy() operation is required! Otherwise, we point to
            # the GG data, which changes from file to file, rather
            # than making a new copy of the data.
            # This comes from python's strategies for fast vector operations.
            #
            G[offset:offset+NtotGals]=GG[0:NtotGals].copy()

            del(GG)
            offset = offset + NtotGals  # Update the offset position for the global array

            fin.close()  # Close the file


        print()
        print("Total galaxies considered:", TotNGals)

        # Convert the Galaxy array into a recarray
        G = G.view(np.recarray)

        w = np.where(G.StellarMass > 1.0)[0]
        print("Galaxies more massive than 10^10Msun/h:", len(w))

        print()

        # Calculate the volume given the first_file and last_file
        #self.volume = self.BoxSize**3.0 * goodfiles / self.MaxTreeFiles

        return G

    def read_pots(self, tracer_file):
        """
        The routine reads in the potentials of the DM particles from the
        simulation. To read in the data the routine will call a function
        that is written in C. This routine only works on simulated
        data that is provided in the gadget format.

        Parameter
        ------------
        tracer_file : string
        String that contains the path and filename of the simulation file
        we want to use.

        returns
        ------------
        pots : np.array(dtype=float)
        The gravitational potential at the positions of the DM particles. 
        *The potentials are dimensionless!*
        """

        read_gadget.get_particles.argtypes = [ctypes.c_char_p]
        read_gadget.get_particles.restype = ctypes.c_int

        num_tracers=read_gadget.get_particles(tracer_file.encode('utf-8'))
        pots = np.empty(num_tracers)
        returnValue=np.empty(num_tracers)

        read_gadget.read_potential.argtypes = [ctypes.c_char_p, ctypes.c_int]
        read_gadget.read_potential.restype = ndpointer(dtype=ctypes.c_float, shape=num_tracers)

        returnValue=read_gadget.read_potential(tracer_file.encode('utf-8'),num_tracers)

        pots[:]=returnValue[:]

        scale_factor = self.read_scale_factor(tracer_file)

        # dividing by a**4 should convert the values back to the RAMSES default output
        # Units of the potential are (km/s)^2 - dividing by c^2 gives a dimensionless property
        return pots/(speed_of_light*speed_of_light)/(scale_factor**4)

    def read_mass_dm(self, tracer_file):
        """
        The routine reads in the mass of the DM particles from the
        simulation. To read in the data the routine will call a function
        that is written in C. This routine only works on simulated
        data that is provided in the gadget format.

        Parameter
        ------------
        tracer_file : string
        String that contains the path and filename of the simulation file we want to use.

        returns
        ------------
        mass_dm : float
        The mass of the DM particles.
        """

        read_gadget.get_mass_particle.argtypes = [ctypes.c_char_p]
        read_gadget.get_mass_particle.restype = ctypes.c_float
        mass_dm=read_gadget.get_mass_particle(tracer_file.encode('utf-8'))

        print("read_mass_dm")
        print("mass_dm=",mass_dm)

        # mass of the DM particles is 1.e10*M_sun/h
        return mass_dm

    def read_scale_factor(self, tracer_file):
        """
        The routine reads in the scale factor of the simulation that
        is saved in the tracer file.

        Parameter
        ------------
        tracer_file : string
        String that contains the path and filename of the simulation file we want to use.

        returns
        ------------
        scale_factor : float
        The scale factor of the snapshot.
        """

        print("read scale factor")

        read_gadget.get_scale_factor.argtypes = [ctypes.c_char_p]
        read_gadget.get_scale_factor.restype = ctypes.c_float
        scale_factor = read_gadget.get_scale_factor(tracer_file.encode('utf-8'))

        return scale_factor


    def read_void_radius(self, void_file):
        """
        This function reads in an output file from Revolver and saves the
        effective radii for the voids. We could read it from any of Revolver's
        output files, the radius is the same in all of them.

        Parameter
        ------------
        barycentre_file: string
          The path and filename of the input file.

        returns
        ------------
        eff_void_radius: np.array(dtype=float)
           An array with the effective radii of the voids.
        """

        barycenterdata =  np.loadtxt(void_file, skiprows=2)
        num_voids=len(barycenterdata)
        eff_void_radius = np.empty((num_voids))
        eff_void_radius = barycenterdata[:,4]

        # in Mpc/h
        return eff_void_radius


    def read_circ_center(self, void_file, num_voids):
        """
        This function reads in an output file from Revolver and saves the position
        of the circumcentre (position of lowest density).

        Parameter
        ------------
        void_detail_file: string
           The path and filename of the input file.
        num_voids: int
           The number of identified voids.

        returns
        ------------
        circ_centre: np.array(3, dtype=float)
           The positions of the circumcentres of the voids.
        """

        voiddata = np.loadtxt(void_file, skiprows=2)
        circ_center = np.empty((num_voids, 3))

        circ_center[:,0]=voiddata[:,1]
        circ_center[:,1]=voiddata[:,2]
        circ_center[:,2]=voiddata[:,3]

        return circ_center

    def read_densities(self, densities_file, tracer_file, box_length):
        """
        This function reads in the densities associated with a DM particle
        using the Voronoi Tessalation Field Estimator. It is read from an
        output file from Revolver.

        Parameter
        ------------
        densities_file: string
        The path and filename of the file containing the information about the
        densities assigned to a cell around the DM particle.
        tracer_file: string
        The output file from our N-body simulation provided in the gadget format.

        returns
        ------------
        densities_ np.array(dtype=float)
        The desnities estimated with the VTFE.
        """
        # using the Voronoi Tessellation Field Estimator (VTFE) from Revolver
        mass_dm = self.read_mass_dm(tracer_file)
        with open(densities_file, 'r') as File:
            npart = np.fromfile(File, dtype=np.int32, count=1)[0]
            vols = np.fromfile(File, dtype=np.float64, count=npart)
            vols = vols/npart*(box_length**3)
            #densities = mass_dm*1.e10 * M_sun_in_kg / densities
            densities = 1 / vols

        # densities in m_DM/(Mpc/h)^3
        return densities

    def largest_void(self, eff_void_radius):
        """
        This function will find the largest void using the effective radius
        as  measure.

        Parameter
        ------------
        eff_void_radius: np.array(dtype=float)
        An array holding the effective radii for the various voids.

        returns
        ------------
        Returns the index of the largest void.
        """

        return np.int(np.where(eff_void_radius == np.max(eff_void_radius))[0])


    def dm_in_void(self, Ncroot, id_file):

        dm_in_void_id = [] # *indices* of DM particles in the void, read from our additional output
        id_file_void_id = []
        k=0
        N_particles = (np.int(Ncroot))**3
        invalid_voidID = -1
        void_IDs_of_dm_particles = invalid_voidID * np.ones(N_particles+1,dtype=np.int64)
        for line in open(id_file,'r'):
            if not line.startswith('#'):
                if(k%2 == 0):
                    # add a scalar void identity to the void_id list
                    this_voidID=np.fromstring(line, dtype=np.int, sep=' ')
                    id_file_void_id.append(this_voidID[0])
                else:
                    # associate a list of DM particle identities with the void
                    dm_particles_of_this_void=np.fromstring(line, dtype=np.int64, sep=' ')
                    ## Replace the invalid_voidIDs by of these DM particles
                    ## by their void number =  this_voidID
                    void_IDs_of_dm_particles[dm_particles_of_this_void] = this_voidID

                    dm_in_void_id.append(dm_particles_of_this_void)  # to be deprecated?
                k += 1
        return id_file_void_id, dm_in_void_id


    def read_adjacencies(self, adjacency_file, num_tracers):

        # Copied from Revolver
        # read and assign adjacencies from ZOBOV output
        with open(adjacency_file, 'r') as AdjFile:
            npfromadj = np.fromfile(AdjFile, dtype=np.int32, count=1)[0]
            if not npfromadj == num_tracers:
                sys.exit("npart = %d from adjacency file does not match num_tracers = %d!"
                         % (npfromadj, num_tracers))
            partadjs = [[] for i in range(npfromadj)]  # list of lists to record adjacencies - is there a better way?
            partadjcount = np.zeros(npfromadj, dtype=np.int32)  # counter to monitor adjacencies
            nadj = np.fromfile(AdjFile, dtype=np.int32, count=npfromadj)  # number of adjacencies for each particle
            # load up the adjacencies from ZOBOV output
            for i in range(npfromadj):
                numtomatch = np.fromfile(AdjFile, dtype=np.int32, count=1)[0]
                if numtomatch > 0:
                    # particle numbers of adjacent particles
                    adjpartnumbers = np.fromfile(AdjFile, dtype=np.int32, count=numtomatch)
                    # keep track of how many adjacencies had already been assigned
                    oldcount = partadjcount[i]
                    newcount = oldcount + len(adjpartnumbers)
                    partadjcount[i] = newcount
                    # and now assign new adjacencies
                    partadjs[i][oldcount:newcount] = adjpartnumbers
                    # now also assign the reverse adjacencies
                    # (ZOBOV records only (i adj j) or (j adj i), not both)
                    for index in adjpartnumbers:
                        partadjs[index].append(i)
                    partadjcount[adjpartnumbers] += 1

        return partadjs
