#!/usr/bin/env python

# test_local_minima - unit test the function 'local_minima'
# (C) 2020-2023 Marius Peper, Boud Roukema GPL-3+
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

# Manual test:
#
# ./project shell
# cd reproduce/analysis/python
# PYTHONPATH=${PYTHONPATH}:. ./test_local_minima.py

import numpy as np
from numpy.random import SeedSequence, default_rng
from scipy import stats

#import matplotlib
#matplotlib.use('Agg')
#import matplotlib.pyplot as plt

from lensing_calculations_unit_testable import lensing_calc_unit_testable as lens_calc_ut

import sys

def generate_map(N_pixels, mean_X, std_X):
    seeds = SeedSequence(2039)
    seed = seeds.spawn(2) # one seed is ignored
    rng = default_rng()

    array2d = rng.normal(mean_X, std_X, (N_pixels,N_pixels))

    return array2d


if __name__ == '__main__':

    debug = True

    # parameter to indicate that the test passes (0) or fails (any value other than 0)
    pass_fail_code = 0

    N_pixels = 8 # number of pixels in each direction

    # mean and standard deviation of parameter X
    mean_X = 20.0
    std_X = 0.2 # not used as of 2022-10-18

    # other parameters
    N_neighbours = 2
    N_significance = 7 # not used as of 2022-10-18
    tolerance_frac_std = 0.65 # not used as of 2022-10-18
    tolerance_frac_mean = 0.85

    # generate a Gaussian map
    array2d = generate_map(N_pixels, mean_X, std_X)
    np.set_printoptions(precision=2)


    # Test 1: this should detect only a minimum at [5,5]
    pass_5_5 = 1.0*array2d # copy the array
    pass_5_5[5,5] = 0.0
    if(debug):
        print("pass_5_5= \n",pass_5_5)

    # find the local minima
    local_min = lens_calc_ut.local_minima(pass_5_5, mean_X, std_X, N_pixels,
                 N_neighbours, N_significance,
                 tolerance_frac_std, tolerance_frac_mean)
    if(debug):
        print("local_min = \n",local_min,"\n")
    if(not(local_min[5,5])):
        print("Minimum at [5,5] not found.\n")
        pass_fail_code += 1
    # Count the false pixels - counting rows (axis = 0) is enough
    n_false = np.size(np.where(local_min == False)[0])
    if(not(n_false == N_pixels*N_pixels - 1)):
        print("Should have failed in ",N_pixels*N_pixels-1," pixels, but n_false = ",n_false)
        pass_fail_code += 2


    # Test 2: this should fail to detect a minimum at [0,0]
    fail_0_0 = 1.0*array2d # copy the array
    fail_0_0[0,0] = 0.0
    fail_0_0[N_pixels-1,N_pixels-1] = 0.0 # diagonal neighbour
    if(debug):
        print("fail_0_0=\n",fail_0_0)

    # find the local minima
    local_min = lens_calc_ut.local_minima(fail_0_0, mean_X, std_X, N_pixels,
                 N_neighbours, N_significance,
                 tolerance_frac_std, tolerance_frac_mean)
    if(debug):
        print("local_min = \n",local_min,"\n")
    if(local_min[0,0]):
        print("Minimum at [0,0] wrongly found.\n")
        pass_fail_code += 4
    n_false = np.size(np.where(local_min == False)[0])
    if(not(n_false == N_pixels*N_pixels)):
        print("Should have failed in ",N_pixels*N_pixels," pixels, but n_false = ",n_false)
        pass_fail_code += 8


    # Test 3 this should fail to detect a minimum at [0,0]
    fail_0_0 = 1.0*array2d # copy the array
    fail_0_0[0,0] = 0.0
    fail_0_0[1,2] = 0.0 # slightly off-diagonal neighbour
    if(debug):
        print("fail_0_0=\n",fail_0_0)

    # find the local minima
    local_min = lens_calc_ut.local_minima(fail_0_0, mean_X, std_X, N_pixels,
                 N_neighbours, N_significance,
                 tolerance_frac_std, tolerance_frac_mean)
    if(local_min[0,0]):
        print("Minimum at [0,0] wrongly found.\n")
        pass_fail_code += 16
    if(debug):
        print("local_min = \n",local_min,"\n")
    n_false = np.size(np.where(local_min == False)[0])
    if(not(n_false == N_pixels*N_pixels)):
        print("Should have failed in ",N_pixels*N_pixels," pixels, but n_false = ",n_false)
        pass_fail_code += 32


# Test 4: this should detect minima at [2,2] and [2,6]
    pass_two = 1.0*array2d # copy the array
    pass_two[2,2] = 0.0
    pass_two[2,6] = 0.0
    if(debug):
        print("pass_two= \n",pass_two)

    # find the local minima
    local_min = lens_calc_ut.local_minima(pass_two, mean_X, std_X, N_pixels,
                 N_neighbours, N_significance,
                 tolerance_frac_std, tolerance_frac_mean)
    if(debug):
        print("local_min = \n",local_min,"\n")
    if(not(local_min[2,2]) or not(local_min[2,6])):
        print("Minima at [2,2] and [2,6] not both found.\n")
        pass_fail_code += 64
    # Count the false pixels - counting rows (axis = 0) is enough
    n_false = np.size(np.where(local_min == False)[0])
    if(not(n_false == N_pixels*N_pixels - 2)):
        print("Should have failed in ",N_pixels*N_pixels-2," pixels, but n_false = ",n_false)
        pass_fail_code += 128


    if(0==pass_fail_code):
        print("test_local_minima: check passed. pass_fail_code = ",pass_fail_code)
    else:
        print("test_local_minima: check FAILED. pass_fail_code = ",pass_fail_code)

    sys.exit(pass_fail_code)
