#!/usr/bin/env python

# test_merge_centres - unit test the function 'merge_centres'
# (C) 2020-2023 Marius Peper, Boud Roukema GPL-3+
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This script is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. See <http://www.gnu.org/licenses/>.

# Manual test:
#
# ./project shell
# cd reproduce/analysis/python
# PYTHONPATH=${PYTHONPATH}:. ./test_merge_centres.py

import numpy as np
from numpy.random import SeedSequence, default_rng
from scipy import stats

#import matplotlib
#matplotlib.use('Agg')
#import matplotlib.pyplot as plt

from lensing_calculations_unit_testable import lensing_calc_unit_testable as lens_calc_ut

import sys

# N_pixels = side length of pixel array
# rmin, rmax = minimum and maximum radii of the voids
# N_try = number of centres to generate
def generate_centres(N_pixels, rmin, rmax, N_try):
    seeds = SeedSequence(9227)
    seed = seeds.spawn(2) # one seed is ignored
    rng = default_rng()

    # Set up random centres - the density depends on the values of 'N_pixels' and N_try
    pre_centre = np.zeros((N_try,2))
    pre_centre[:,0] = rng.uniform(0.0,np.double(N_pixels),size=N_try)
    pre_centre[:,1] = rng.uniform(0.0,np.double(N_pixels),size=N_try)

    # Set up random radii - any reasonable distribution should be OK for testing
    rads = rng.uniform(low=rmin, high=rmax, size=N_try)

    return pre_centre, rads


if __name__ == '__main__':

    debug = True

    # parameter to indicate that the test passes (0) or fails (any value other than 0)
    pass_fail_code = 0

    # default parameters for the tests
    N_pixels = 16
    rmin = 3.0
    rmax = 5.0
    merge_distance = 1.0


    print("Test 0: generic test - just to see if the function runs at all")
    N_try = 5
    pre_centre, rads = generate_centres(N_pixels, rmin, rmax, N_try)
    local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, rads, merge_distance, N_pixels)
    print("\n\n")

    print("Test 1: all pairs should merge into 1")
    #merge_distance_all = 1.4143*N_pixels # length of diagonal for the square (not T^2) case
    merge_distance_all = 0.7072*N_pixels # length of diagonal for the T^2 case
    pre_centre, rads = generate_centres(N_pixels, rmin, rmax, N_try)
    local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, rads, merge_distance_all, N_pixels)
    print("All pairs should merge, n = ",len(local_centres))
    if(len(local_centres) != 1):
        pass_fail_code += 1
    print("\n\n")

    print("Test 2: no pairs should merge")
    merge_distance_none = 1e-30*N_pixels # the centres are unlikely to be this close
    pre_centre, rads = generate_centres(N_pixels, rmin, rmax, N_try)
    rads *= 1e-30 # make the radii tiny so that the voids are unlikely to overlap
    local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, rads, merge_distance_none, N_pixels)
    print("No pairs should merge, n = ",len(local_centres))
    if(len(local_centres) != N_try):
        pass_fail_code += 2
    print("\n\n")

    print("Test 3: an empty list is provided")
    pre_centre, rads = generate_centres(N_pixels, rmin, rmax, 0)
    local_centres, local_radii = lens_calc_ut.merge_centres(pre_centre, rads, merge_distance, N_pixels)
    print("There are no voids, so none should be returned, n = ",len(local_centres))
    if(len(local_centres) != 0):
        pass_fail_code += 4
    print("\n\n")

    # Overall test results:
    if(0==pass_fail_code):
        print("test_merge_centres: check passed. pass_fail_code = ",pass_fail_code)
    else:
        print("test_merge_centres: check FAILED. pass_fail_code = ",pass_fail_code)

    sys.exit(pass_fail_code)
