/*
gsl_lensing - solve the differential equations of the void lensing project
Copyright (C) 2022 Marius Peper, Boud Roukema

The purpose of this file is to solve the differential equations of the lensing project.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <gsl/gsl_complex.h>
#include <complex.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_odeiv2.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_fit.h>
#define N_T_SPLINE 150


struct optical_scalar_diff_parameters_s
{
  gsl_interp_accel * accel_R44;
  gsl_spline * spline_R44;
  gsl_interp_accel * accel_C1010;
  gsl_spline * spline_C1010;
  gsl_interp_accel * accel_C1020;
  gsl_spline * spline_C1020;
  double R44[150];
  double C1010[150];
  double C1020[150];
};

struct expansion_splines_s
{
  gsl_interp_accel * accel_expansion;
  gsl_spline * spline_expansion;
  double *expansion;
  double *expansion_radii;
};

struct second_order_splines_s
{
  gsl_interp_accel * accel_real_sig;
  gsl_spline * spline_real_sig;
  gsl_interp_accel * accel_img_sig;
  gsl_spline * spline_img_sig;
  gsl_interp_accel * accel_R44;
  gsl_spline * spline_R44;
};

double R44_model(double r, double shift_x, double shift_y, double amp)
{
  /*
    A simple model function to simulate an input for Ricci tensor R44.
    This function was only used for debugging.
  */
  
  return amp*((r-shift_x)*(r-shift_x)) - shift_y;
}

double C1010_model(double r, double shift_x, double shift_y, double amp)
{
  /*
    A simple model function to simulate an input for Weyl tensor C.
    This function was only used for debugging.
  */

  return amp*((r-shift_x)*(r-shift_x)) - shift_y;
}

int func_only_expansion (double t, const double y[], double f[], void *params)
{
  /*
    A simplified differential equation. This function is defines the simplified set of equations,
    in which the shear (sigma) is set to be zero.
  */

  (void)(t); /* avoid unused parameter warning */
  struct optical_scalar_diff_parameters_s optical_scalar_diff_parameters =
    *(struct optical_scalar_diff_parameters_s *)params;
    
  f[0] = (-gsl_spline_eval(optical_scalar_diff_parameters.spline_R44, t, optical_scalar_diff_parameters.accel_R44)
          - 0.5*(y[0]*y[0]) );

  return GSL_SUCCESS;
}

int jac_only_expansion (double t, const double y[], double *dfdy, double dfdt[], void *params)
{
  /*
    A simplified differential equation. This function is the jacobian for the simplified set of equations,
    in which the shear (sigma) is set to be zero.
  */

  (void)(t); /* avoid unused parameter warning */
  struct optical_scalar_diff_parameters_s optical_scalar_diff_parameters =
    *(struct optical_scalar_diff_parameters_s *)params;
  
  gsl_matrix_view dfdy_mat
    = gsl_matrix_view_array (dfdy, 1, 1);
  gsl_matrix * m = &dfdy_mat.matrix;

  gsl_matrix_set (m, 0, 0, -y[0]);

  dfdt[0] = 0.0;  

  return GSL_SUCCESS;
}

double only_expansion(double *r, double *R44, double *C1010, double *C1020, double t, double t1)
{
  /*
    A simplified differential equation. This function is solving only one of the differential
    equations, the shear (sigma) is set to be zero in this simplification.
  */

  double f_tmp[1];
  double out;
  struct optical_scalar_diff_parameters_s params;

  gsl_interp_accel *accel_R44 = gsl_interp_accel_alloc();
  gsl_spline *spline_R44 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
    /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/
    

  gsl_spline_init( spline_R44,
                   r,
                   R44,
                   N_T_SPLINE );

  gsl_interp_accel *accel_C1010 = gsl_interp_accel_alloc();
  gsl_spline *spline_C1010 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_C1010,
                   r,
                   C1010,
                   N_T_SPLINE );

  gsl_interp_accel *accel_C1020 = gsl_interp_accel_alloc();
  gsl_spline *spline_C1020 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_C1020,
                   r,
                   C1020,
                   N_T_SPLINE );
  /*
  FILE *outfile;
  outfile = fopen("python_params.dat", "w+");
  int j;
  for(j=0;j<=100;j++)
    {
      fprintf(outfile,"%f %e %e %e\n",r[j], R44[j], C1010[j], C1020[j]);
    }
  fclose(outfile);
  */

  params.accel_R44 = accel_R44;
  params.spline_R44 = spline_R44;
  params.accel_C1010 = accel_C1010;
  params.spline_C1010 = spline_C1010;
  params.accel_C1020 = accel_C1020;
  params.spline_C1020 = spline_C1020;
  
  gsl_odeiv2_system sys;

  sys.function = func_only_expansion;
  sys.jacobian = jac_only_expansion;
  sys.dimension = 1;
  sys.params = &params;
  
  const gsl_odeiv2_step_type *T =
    gsl_odeiv2_step_rkf45;

  gsl_odeiv2_step * s
    = gsl_odeiv2_step_alloc (T, 1);
  gsl_odeiv2_control * c
    = gsl_odeiv2_control_y_new (1e-6,1.0);
  gsl_odeiv2_evolve * e
    = gsl_odeiv2_evolve_alloc (1);

  double h = 1e-6;
  // initial conditions
  double y[1] = { 0.0};

  //FILE *diff_sol_file;
  //diff_sol_file = fopen("diff_solution.dat", "w+");

  while (t < t1)
    {
      if(t != 0) h = fmin(0.1 , h);
      int status = gsl_odeiv2_evolve_apply (e, c, s,
                                           &sys,
                                           &t, t1,
                                           &h, y);
      if (status != GSL_SUCCESS)
          break;

      //printf("t = %g t1 %g; h = %g\n",
      //       t,t1, h);
      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      func_only_expansion(t,
           y,
           (double*)&f_tmp,
           &params);

      out=y[0];
      //printf("%.5e %.5e\n", t, y[0]);

      
      //printf("f_tmp[0] = %g   f_tmp[1] = %g\n",f_tmp[0],f_tmp[1]);

      //printf("%.5e %.5e %.5e %.5e %.5e\n", t, y[0], y[1], y[2], y[3]);
      //fprintf(diff_sol_file,"%.5e %.5e\n", t, y[0]);

    }

  //fclose(diff_sol_file);

  gsl_odeiv2_evolve_free (e);
  gsl_odeiv2_control_free (c);
  gsl_odeiv2_step_free (s);

  // TODO: free the splines!

  gsl_spline_free (spline_R44);
  gsl_interp_accel_free (accel_R44);
  gsl_spline_free (spline_C1010);
  gsl_interp_accel_free (accel_C1010);
  gsl_spline_free (spline_C1020);
  gsl_interp_accel_free (accel_C1020);

  return out;
}


int func_optical_scalars (double t, const double y[], double f[], void *params)
{
  /*
    The system of differential equations. This function specifies set of equations following
    Sasaki (1993). The optical scalars define the expansion and shear that a light bundle,
    that transverses through a void, experiences.
  */

  (void)(t); /* avoid unused parameter warning */
  struct optical_scalar_diff_parameters_s optical_scalar_diff_parameters =
    *(struct optical_scalar_diff_parameters_s *)params;
    
  f[0] = ((-1)*gsl_spline_eval(optical_scalar_diff_parameters.spline_R44, t, optical_scalar_diff_parameters.accel_R44)
          - 2*(y[1]*y[1] + y[2]*y[2]) - 0.5*(y[0]*y[0]) );
  f[1] = ((-1)*gsl_spline_eval(optical_scalar_diff_parameters.spline_C1010, t, optical_scalar_diff_parameters.accel_C1010)
          -y[0]*y[1]);
  f[2] = ((-1)*gsl_spline_eval(optical_scalar_diff_parameters.spline_C1020, t, optical_scalar_diff_parameters.accel_C1020)
          -y[0]*y[2]);

  return GSL_SUCCESS;
}



int jac_optical_scalars (double t, const double y[], double *dfdy, double dfdt[], void *params)
{
  /*
    This function is the jacobian for the full set of equations, defined 
    in func_optical_scalars() .
  */
  
  (void)(t); /* avoid unused parameter warning */

  struct optical_scalar_diff_parameters_s optical_scalar_diff_parameters =
    *(struct optical_scalar_diff_parameters_s *)params;
  
  gsl_matrix_view dfdy_mat
    = gsl_matrix_view_array (dfdy, 3, 3);
  gsl_matrix * m = &dfdy_mat.matrix;

  gsl_matrix_set (m, 0, 0, -y[0]);
  gsl_matrix_set (m, 0, 1, -4*y[1]);
  gsl_matrix_set (m, 0, 2, 4*y[2]);

  gsl_matrix_set (m, 1, 0, -y[1]);
  gsl_matrix_set (m, 1, 1, -y[0]);
  gsl_matrix_set (m, 1, 2, 0);

  gsl_matrix_set (m, 2, 0, -y[2]);
  gsl_matrix_set (m, 2, 1, 0);
  gsl_matrix_set (m, 2, 2, -y[0]);


  dfdt[0] = 0.0;
  dfdt[1] = 0.0;
  dfdt[2] = 0.0;
  

  return GSL_SUCCESS;
}

int func_distance (double t, const double y[], double f[], void *params)
{
  /*
    Equation to double check the results we receive for the optical scalars.
    This function defines eq.3 of https://arxiv.org/abs/1204.0909 to verify
    if our results are reasonable. It relates the angular diameter distance
    to the optical scalars.
  */

  (void)(t); /* avoid unused parameter warning */
  struct expansion_splines_s expansion_spline =
    *(struct expansion_splines_s *)params;
    
  f[0] = ( gsl_spline_eval(expansion_spline.spline_expansion, t, expansion_spline.accel_expansion) * y[0] );

  //printf("Inside t: %f  spline: %e  f[0]: %e\n",t,gsl_spline_eval(expansion_spline.spline_expansion, t, expansion_spline.accel_expansion), f[0]);


  return GSL_SUCCESS;
}

int func_second_order (double t, const double y[], double f[], void *params)
{
  /*
    Equation to double check the results we receive for the optical scalars.
    This function defines eq.4 of https://arxiv.org/abs/1204.0909 to verify
    if our results are reasonable. It relates the angular diameter distance
    to the optical scalars.
  */

  (void)(t); /* avoid unused parameter warning */
  struct second_order_splines_s splines =
    *(struct second_order_splines_s *)params;

  //printf("INSIDE: %f %.5e %.5e\n",t, gsl_spline_eval(splines.spline_real_sig, t, splines.accel_real_sig), gsl_spline_eval(splines.spline_R44, t, splines.accel_R44) );
    
  f[0] = y[1];
  f[1] = ((-1)*( ((gsl_spline_eval(splines.spline_real_sig, t, splines.accel_real_sig)*gsl_spline_eval(splines.spline_real_sig, t, splines.accel_real_sig) - gsl_spline_eval(splines.spline_img_sig, t, splines.accel_img_sig)) * gsl_spline_eval(splines.spline_img_sig, t, splines.accel_img_sig) ) + 0.5*(gsl_spline_eval(splines.spline_R44, t, splines.accel_R44)) ) * y[0] );

  //printf("Inside t: %f  spline: %e  f[0]: %e\n",t,gsl_spline_eval(expansion_spline.spline_expansion, t, expansion_spline.accel_expansion), f[0]);


  return GSL_SUCCESS;
}



int jac_distance (double t, const double y[], double *dfdy, double dfdt[], void *params)
{
  /*
    Equation to double check the results we receive for the optical scalars.
    This function is the jacobian for eq.3 of https://arxiv.org/abs/1204.0909 
    defined in func_distance() that we use to verify if our resutls
    are reasonable
  */

  (void)(t); /* avoid unused parameter warning */

  struct expansion_splines_s expansion_spline =
    *(struct expansion_splines_s *)params;
  
  gsl_matrix_view dfdy_mat
    = gsl_matrix_view_array (dfdy, 1, 1);
  gsl_matrix * m = &dfdy_mat.matrix;

  gsl_matrix_set (m, 0, 0, gsl_spline_eval(expansion_spline.spline_expansion, t, expansion_spline.accel_expansion));

  dfdt[0] = 1.0;

  return GSL_SUCCESS;
}

double *optical_scalars_check_distance(double *r, double *R44, double *C1010, double *C1020, double t, double t1)
{
  /*
    Equation to double check the results we receive for the optical scalars.
    This function is a modified version of optical_scalars() . It uses the
    output of the optical scalars for different radii (positions in the
    simulated box) to solve the differential equations defined in func_distance()
    and func_second_order() . These equations are taken from 
    https://arxiv.org/abs/1204.0909 .
  */

  double f_tmp[3];
  double f_distance[1];
  double *tmp;
  double *out = NULL;
  double *expansion = NULL;
  double *real_sig = NULL;
  double *img_sig = NULL;
  double *expansion_radii = NULL;
  int N_diff=1;
  int N_out=0;
  int N_loop=0;
  double tstart=t;
  out=malloc(sizeof(double)*4);
  expansion=malloc(sizeof(double));
  real_sig=malloc(sizeof(double));
  img_sig=malloc(sizeof(double));
  expansion_radii=malloc(sizeof(double));

  struct optical_scalar_diff_parameters_s params;
  struct expansion_splines_s params_distance;

  gsl_interp_accel *accel_R44 = gsl_interp_accel_alloc();
  gsl_spline *spline_R44 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_R44,
                   r,
                   R44,
                   N_T_SPLINE );

  gsl_interp_accel *accel_C1010 = gsl_interp_accel_alloc();
  gsl_spline *spline_C1010 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_C1010,
                   r,
                   C1010,
                   N_T_SPLINE );

  gsl_interp_accel *accel_C1020 = gsl_interp_accel_alloc();
  gsl_spline *spline_C1020 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_C1020,
                   r,
                   C1020,
                   N_T_SPLINE );

  params.accel_R44 = accel_R44;
  params.spline_R44 = spline_R44;
  params.accel_C1010 = accel_C1010;
  params.spline_C1010 = spline_C1010;
  params.accel_C1020 = accel_C1020;
  params.spline_C1020 = spline_C1020;
  
  gsl_odeiv2_system sys;

  sys.function = func_optical_scalars;
  sys.jacobian = jac_optical_scalars;
  sys.dimension = 3;
  sys.params = &params;

  /* 
     The first part of the function is similar to optical_scalar()
     to solve for the optical scalars. It uses a different solver
     scheme though (one that uses the jacobian) to double check
     that the results are not dependent on the chosen algorithm.
     Using a different method requires a driver thus there are
     small derivatios to the previously defined function.
  */
  
  const gsl_odeiv2_step_type *T =
    gsl_odeiv2_step_rk4imp;


  gsl_odeiv2_step * s
    = gsl_odeiv2_step_alloc (T, 3);
  //gsl_odeiv2_control * c
  //  = gsl_odeiv2_control_y_new (1e-6,1.0);
  //gsl_odeiv2_evolve * e
  //  = gsl_odeiv2_evolve_alloc (3);

  gsl_odeiv2_driver * d =
    gsl_odeiv2_driver_alloc_y_new (&sys, T,
                                  1e-6, 1e-6, 0.0);
  gsl_odeiv2_step_set_driver(s, d);

  double h = 1e-6;
  // initial conditions
  double y[3] = { 0.0, 0.0, 0.0};

  //FILE *diff_sol_file;
  //diff_sol_file = fopen("diff_solution.dat", "w+");
  t=t+h;

  /* old structure is just commented out */

  while (t < (t1-0.1))
    {
      if(t != 0) h = fmin(0.1 , 0.2);
      //int status = gsl_odeiv2_evolve_apply (e, c, s,
      //                                    &sys,
      //                                    &t, t1,
      //                                    &h, y);
      int status = gsl_odeiv2_driver_apply (d, &t, (t+h), y);
      if (status != GSL_SUCCESS)
          break;

      //printf("t = %g t1 %g; h = %g\n",
      //       t,t1, h);
      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      //func_optical_scalars(t,
      //     y,
      //     (double*)&f_tmp,
      //     &params);

      out[0]=y[0];
      out[1]=y[1];
      out[2]=y[2];

      // Save the optical scalars for various radii.
      if(N_loop==N_out)
        {    
          if(N_diff>1)
            {
              tmp=realloc(expansion,sizeof(double)*N_diff);
              if(tmp == NULL) {
                printf("\nError! memory not allocated.");
                exit(0);
              }
              expansion=tmp;
              tmp=realloc(real_sig,sizeof(double)*N_diff);
              if(tmp == NULL) {
                printf("\nError! memory not allocated.");
                exit(0);
              }
              real_sig=tmp;
              tmp=realloc(img_sig,sizeof(double)*N_diff);
              if(tmp == NULL) {
                printf("\nError! memory not allocated.");
                exit(0);
              }
              img_sig=tmp;
              tmp=realloc(expansion_radii,sizeof(double)*N_diff);
              if(tmp == NULL) {
                printf("\nError! memory not allocated.");
                exit(0);
              }
              expansion_radii=tmp;
            }
          if(N_diff==1)
            {
              expansion[N_diff-1]=0.0;
              real_sig[N_diff-1]=0.0;
              img_sig[N_diff-1]=0.0;
              expansion_radii[N_diff-1]=0.0;
            }
          else
            {
              expansion[N_diff-1]=y[0];
              real_sig[N_diff-1]=y[1];
              img_sig[N_diff-1]=y[2];
              expansion_radii[N_diff-1]=t;
            }

          //printf("N_diff: %i    expansion_radii[N_diff-1]: %f\n", N_diff, expansion_radii[N_diff-1]);
          N_diff++;
          // We interpolate over the data later so we don't need to remember all steps.
          N_out=N_out+30;
          //printf("%.5e %.5e\n", t, y[0]);
        }
      N_loop++;
      /*
      printf("f_tmp[0] = %g   f_tmp[1] = %g\n",f_tmp[0],f_tmp[1]);

      printf("%.5e %.5e %.5e %.5e %.5e\n", t, y[0], y[1], y[2], y[3]);
      fprintf(diff_sol_file,"%.5e %.5e %.5e %.5e %.5e\n", t, y[0], y[1], y[2], y[3]);
      */
    }

  //fclose(diff_sol_file);
  if(N_diff>1)
    {
      tmp=realloc(expansion,sizeof(double)*N_diff);
      if(tmp == NULL) {
        printf("\nError! memory not allocated.");
        exit(0);
      }
      expansion=tmp;
      tmp=realloc(real_sig,sizeof(double)*N_diff);
      if(tmp == NULL) {
        printf("\nError! memory not allocated.");
        exit(0);
      }
      real_sig=tmp;
      tmp=realloc(img_sig,sizeof(double)*N_diff);
      if(tmp == NULL) {
        printf("\nError! memory not allocated.");
        exit(0);
      }
      img_sig=tmp;
      tmp=realloc(expansion_radii,sizeof(double)*N_diff);
      if(tmp == NULL) {
        printf("\nError! memory not allocated.");
        exit(0);
      }
      expansion_radii=tmp;
    }
  expansion[N_diff-1]=y[0];
  real_sig[N_diff-1]=y[1];
  img_sig[N_diff-1]=y[2];
  expansion_radii[N_diff-1]=t;
  //printf("N_diff: %i    expansion_radii[N_diff-1]: %f\n", N_diff, expansion_radii[N_diff-1]);
  //N_diff++;
  //printf("%.5e %.5e\n", t, y[0]);


  /*
    This part solves eq.3 of https://arxiv.org/abs/1204.0909 .
    It's analogous to the previous code but uses a different
    function and jacobian.
  */
  ///////////////////////////////////////////
  /*

  gsl_interp_accel *accel_expansion = gsl_interp_accel_alloc();
  gsl_spline *spline_expansion =
    gsl_spline_alloc(gsl_interp_cspline, N_diff);

  gsl_spline_init( spline_expansion,
                   expansion_radii,
                   expansion,
                   N_diff );

  //params_distance.expansion = expansion;
  //params_distance.expansion_radii = expansion_radii;
  params_distance.accel_expansion = accel_expansion;
  params_distance.spline_expansion = spline_expansion;

  
  gsl_odeiv2_system sys_distance;

  sys_distance.function = func_distance;
  sys_distance.jacobian = jac_distance;
  sys_distance.dimension = 1;
  sys_distance.params = &params_distance;
  
  gsl_odeiv2_step * s_distance
    = gsl_odeiv2_step_alloc (T, 1);
  //gsl_odeiv2_control * c_distance
  //  = gsl_odeiv2_control_y_new (1e-6,1.0);
  //gsl_odeiv2_evolve * e_distance
  //  = gsl_odeiv2_evolve_alloc (1);

  gsl_odeiv2_driver * d_distance =
    gsl_odeiv2_driver_alloc_y_new (&sys_distance, T,
                                  1e-6, 1e-6, 0.0);
  gsl_odeiv2_step_set_driver(s_distance, d_distance);


  h = 1e-6;
  // initial conditions
  double y_distance[1] = { 3300.0 };
  int i=0;
  double exp;

  t=tstart+h;
  while (t < (t1-0.1))
    {
      if(t != 0) h = fmin(0.1 , 0.2);
      //int status = gsl_odeiv2_evolve_apply (e_distance, c_distance, s_distance,
      //                                    &sys_distance,
      //                                   &t, t1,
      //                                   &h, y_distance);
      int status = gsl_odeiv2_driver_apply (d_distance, &t, (t+h), y_distance);
      if (status != GSL_SUCCESS)
          break;

      //printf("t = %g t1 %g; h = %g\n",
      //       t,t1, h);
      
      printf("t: %f  spline: %f\n",t,gsl_spline_eval(spline_expansion, t, accel_expansion));

      //while(expansion_radii[i] < t)
      //  {
      //    i++;
      //  }
      //if(i>=N_diff)
      //  {
      //    i=i-1;
      //  }
      //exp=(expansion[i]-expansion[i-1])/(expansion_radii[i]-expansion_radii[i-1]);
      //exp=expansion[i-1]+(exp*(t-expansion_radii[i-1]));

      printf("t %f  exp_rad %f  exp_inp %e  exp %e\n",t, expansion_radii[i], expansion[i], exp);

      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      //func_distance(t,
      //    y_distance,
      //   (double*)&f_distance,
      //   &params_distance);

      printf("%.5e %.5e\n", t, y_distance[0]);

      out[3]=y_distance[0];
    }

  //gsl_odeiv2_evolve_free (e_distance);
  //gsl_odeiv2_control_free (c_distance);
  gsl_odeiv2_step_free (s_distance);

  gsl_spline_free (spline_expansion);
  gsl_interp_accel_free (accel_expansion);
  */
  //////////////////////////////////////////////////

  /*
    This part solves eq.4 of https://arxiv.org/abs/1204.0909 .
    It's analogous to the previous code but uses a different
    function.
  */
  ///////////////////////////////////////////

  struct second_order_splines_s params_second_order;

  gsl_interp_accel *accel_real_sig = gsl_interp_accel_alloc();
  gsl_spline *spline_real_sig =
    gsl_spline_alloc(gsl_interp_steffen, N_diff);
  /*gsl_spline_alloc(gsl_interp_cspline, N_diff);*/

  gsl_spline_init( spline_real_sig,
                   expansion_radii,
                   real_sig,
                   N_diff );

  gsl_interp_accel *accel_img_sig = gsl_interp_accel_alloc();
  gsl_spline *spline_img_sig =
    gsl_spline_alloc(gsl_interp_steffen, N_diff);
  /*gsl_spline_alloc(gsl_interp_cspline, N_diff);*/

  gsl_spline_init( spline_img_sig,
                   expansion_radii,
                   img_sig,
                   N_diff );

  params_second_order.accel_R44 = accel_R44;
  params_second_order.spline_R44 = spline_R44;
  params_second_order.accel_real_sig = accel_real_sig;
  params_second_order.spline_real_sig = spline_real_sig;
  params_second_order.accel_img_sig = accel_img_sig;
  params_second_order.spline_img_sig = spline_img_sig;

  const gsl_odeiv2_step_type *T_second_order =
    gsl_odeiv2_step_rkf45;

  gsl_odeiv2_system sys_second_order;

  sys_second_order.function = func_second_order;
  sys_second_order.jacobian = NULL;
  sys_second_order.dimension = 2;
  sys_second_order.params = &params_second_order;
  
  gsl_odeiv2_step * s_second_order
    = gsl_odeiv2_step_alloc (T_second_order, 2);
  //gsl_odeiv2_control * c_distance
  //  = gsl_odeiv2_control_y_new (1e-6,1.0);
  //gsl_odeiv2_evolve * e_distance
  //  = gsl_odeiv2_evolve_alloc (1);

  gsl_odeiv2_driver * d_second_order =
    gsl_odeiv2_driver_alloc_y_new (&sys_second_order, T_second_order,
                                  1e-6, 1e-6, 0.0);
  gsl_odeiv2_step_set_driver(s_second_order, d_second_order);


  h = 1e-6;
  // initial conditions
  double y_second_order[2] = { 3300, 0.0 };

  t=tstart+h;
  while (t < (t1-0.1))
    {
      if(t != 0) h = fmin(0.1 , 0.2);
      //int status = gsl_odeiv2_evolve_apply (e_distance, c_distance, s_distance,
      //                                    &sys_distance,
      //                                   &t, t1,
      //                                   &h, y_distance);
      int status = gsl_odeiv2_driver_apply (d_second_order, &t, (t+h), y_second_order);
      if (status != GSL_SUCCESS)
          break;

      //printf("t = %g t1 %g; h = %g\n",
      //       t,t1, h);
      
      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      //func_distance(t,
      //    y_distance,
      //   (double*)&f_distance,
      //   &params_distance);

      printf("%.5e %.5e %.5e\n", t, y_second_order[0], y_second_order[1]);

      out[3]=y_second_order[0];
    }

  //gsl_odeiv2_evolve_free (e_distance);
  //gsl_odeiv2_control_free (c_distance);
  gsl_odeiv2_step_free (s_second_order);

  gsl_spline_free (spline_real_sig);
  gsl_interp_accel_free (accel_real_sig);
  gsl_spline_free (spline_img_sig);
  gsl_interp_accel_free (accel_img_sig);

  //////////////////////////////////////////////////

  //gsl_odeiv2_evolve_free (e);
  //gsl_odeiv2_control_free (c);
  gsl_odeiv2_step_free (s);

  gsl_spline_free (spline_R44);
  gsl_interp_accel_free (accel_R44);
  gsl_spline_free (spline_C1010);
  gsl_interp_accel_free (accel_C1010);
  gsl_spline_free (spline_C1020);
  gsl_interp_accel_free (accel_C1020);
  
  return out;
}

double *optical_scalars(double *r, double *R44, double *C1010, double *C1020, double t, double t1)
{
  /*
    The system of differential equations. This function solves set of equations following
    Sasaki (1993). The optical scalars define the expansion and shear that a light bundle,
    that transverses through a void, experiences. This routine uses the gsl library to
    solve the set of equations that are defined in func_optical_scalars() .
  */

  double f_tmp[3];
  double *out = NULL;
  out=malloc(sizeof(double)*3);

  struct optical_scalar_diff_parameters_s params;

  gsl_interp_accel *accel_R44 = gsl_interp_accel_alloc();
  gsl_spline *spline_R44 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_R44,
                   r,
                   R44,
                   N_T_SPLINE );

  gsl_interp_accel *accel_C1010 = gsl_interp_accel_alloc();
  gsl_spline *spline_C1010 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_C1010,
                   r,
                   C1010,
                   N_T_SPLINE );

  gsl_interp_accel *accel_C1020 = gsl_interp_accel_alloc();
  gsl_spline *spline_C1020 =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_C1020,
                   r,
                   C1020,
                   N_T_SPLINE );
  /*
  FILE *outfile;
  outfile = fopen("python_params.dat", "w+");
  int j;
  for(j=0;j<=100;j++)
    {
      fprintf(outfile,"%f %e %e %e\n",r[j], R44[j], C1010[j], C1020[j]);
    }
  fclose(outfile);
  */

  params.accel_R44 = accel_R44;
  params.spline_R44 = spline_R44;
  params.accel_C1010 = accel_C1010;
  params.spline_C1010 = spline_C1010;
  params.accel_C1020 = accel_C1020;
  params.spline_C1020 = spline_C1020;
  
  gsl_odeiv2_system sys;

  sys.function = func_optical_scalars;
  sys.jacobian = jac_optical_scalars;
  sys.dimension = 3;
  sys.params = &params;
  
  const gsl_odeiv2_step_type *T =
    gsl_odeiv2_step_rkf45;

  gsl_odeiv2_step * s
    = gsl_odeiv2_step_alloc (T, 3);
  gsl_odeiv2_control * c
    = gsl_odeiv2_control_y_new (1e-6,1.0);
  gsl_odeiv2_evolve * e
    = gsl_odeiv2_evolve_alloc (3);

  double h = -1e-6;
  // initial conditions
  double y[3] = { 0.0, 0.0, 0.0};

  //FILE *diff_sol_file;
  //diff_sol_file = fopen("diff_solution.dat", "w+");

  while (t > t1)
    {
      if(t != 0) h = fmax(-0.1 , h);
      int status = gsl_odeiv2_evolve_apply (e, c, s,
                                           &sys,
                                           &t, t1,
                                           &h, y);
      if (status != GSL_SUCCESS)
          break;

      //printf("t = %g t1 %g; h = %g\n",
      //       t,t1, h);
      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      func_optical_scalars(t,
           y,
           (double*)&f_tmp,
           &params);

      out[0]=y[0];
      out[1]=y[1];
      out[2]=y[2];
      //printf("%.5e %.5e %.5e %.5e\n", t, y[0], y[1], y[2]);

      /*
      printf("f_tmp[0] = %g   f_tmp[1] = %g\n",f_tmp[0],f_tmp[1]);

      printf("%.5e %.5e %.5e %.5e %.5e\n", t, y[0], y[1], y[2], y[3]);
      fprintf(diff_sol_file,"%.5e %.5e %.5e %.5e %.5e\n", t, y[0], y[1], y[2], y[3]);
      */
    }

  //fclose(diff_sol_file);

  gsl_odeiv2_evolve_free (e);
  gsl_odeiv2_control_free (c);
  gsl_odeiv2_step_free (s);

  // TODO: free the splines!

  gsl_spline_free (spline_R44);
  gsl_interp_accel_free (accel_R44);
  gsl_spline_free (spline_C1010);
  gsl_interp_accel_free (accel_C1010);
  gsl_spline_free (spline_C1020);
  gsl_interp_accel_free (accel_C1020);

  return out;
}

struct mass_deficit_diff_parameters_s
{
  gsl_interp_accel * accel_rho;
  gsl_spline * spline_rho;
  double rho[100];
  double rho_mean;
};


int func_mass_deficit (double t, const double y[], double f[], void *rho_inter)
{
  (void)(t); /* avoid unused parameter warning */
  struct mass_deficit_diff_parameters_s mass_deficit_diff_parameters =
    *(struct mass_deficit_diff_parameters_s *)rho_inter;
    
  f[0] = gsl_spline_eval(mass_deficit_diff_parameters.spline_rho, t, mass_deficit_diff_parameters.accel_rho) - mass_deficit_diff_parameters.rho_mean;

  return GSL_SUCCESS;
}


int jac_mass_deficit (double t, const double y[], double *dfdy, double dfdt[], void *rho_inter)
{
  (void)(t); /* avoid unused parameter warning */
  struct mass_deficit_diff_parameters_s mass_deficit_diff_parameters =
    *(struct mass_deficit_diff_parameters_s *)rho_inter;
  
  gsl_matrix_view dfdy_mat
    = gsl_matrix_view_array (dfdy, 1, 1);
  gsl_matrix * m = &dfdy_mat.matrix;

  gsl_matrix_set (m, 0, 0, 1);

  dfdt[0] = 0.0;

  return GSL_SUCCESS;
}


double mass_deficit(double *r, double *rho, double rho_mean, double t, double t1)
{
  double f_tmp[1];
  double *out_radii = malloc(sizeof(double));
  double *out_solutions = malloc(sizeof(double));
  struct mass_deficit_diff_parameters_s rho_inter;

  gsl_interp_accel *accel_rho = gsl_interp_accel_alloc();
  gsl_spline *spline_rho =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_rho,
                   r,
                   rho,
                   N_T_SPLINE );


  rho_inter.accel_rho = accel_rho;
  rho_inter.spline_rho = spline_rho;
  rho_inter.rho_mean = rho_mean;

  gsl_odeiv2_system sys;

  sys.function = func_mass_deficit;
  sys.jacobian = jac_mass_deficit;
  sys.dimension = 1;
  sys.params = &rho_inter;

  FILE *diff_sol_file;
  diff_sol_file = fopen("mass_def_solution.dat", "w+");
  
  const gsl_odeiv2_step_type *T =
    gsl_odeiv2_step_rkf45;

  gsl_odeiv2_step * s
    = gsl_odeiv2_step_alloc (T, 1);
  gsl_odeiv2_control * c
    = gsl_odeiv2_control_y_new (1e-6,1.0);
  gsl_odeiv2_evolve * e
    = gsl_odeiv2_evolve_alloc (1);

  double h = 1e-6;
  // initial conditions
  double y[1] = { 0.0};
  int how_many_iterations=0;

  fprintf(diff_sol_file,"start: %f   end: %f\n", t, t1);

  while (t < t1)
    {
      h = fmin(0.1 , h);
      int status = gsl_odeiv2_evolve_apply (e, c, s,
                                           &sys,
                                           &t, t1,
                                           &h, y);
      if (status != GSL_SUCCESS)
          break;

      printf("t = %g t1 %g; h = %g\n",
             t,t1, h);
      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      func_mass_deficit(t,
           y,
           (double*)&f_tmp,
            &rho_inter);

      printf("%.5e %.5e\n", t, y[0]);
      fprintf(diff_sol_file,"%.5e %.5e\n", t, y[0]);

    }
  fclose(diff_sol_file);

  gsl_odeiv2_evolve_free (e);
  gsl_odeiv2_control_free (c);
  gsl_odeiv2_step_free (s);

  gsl_spline_free (spline_rho);
  gsl_interp_accel_free (accel_rho);
  printf("y[0] = %f:\n",y[0]);

  return y[0]; 

}


struct ISW_diff_parameters_s
{
  gsl_interp_accel * accel_phi_dot;
  gsl_spline * spline_phi_dot;
  double phi_dot[100];
};


int func_ISW (double t, const double y[], double f[], void *phi_dot_inter)
{
  (void)(t); /* avoid unused parameter warning */
  struct ISW_diff_parameters_s ISW_diff_parameters =
    *(struct ISW_diff_parameters_s *)phi_dot_inter;
    
  f[0] = gsl_spline_eval(ISW_diff_parameters.spline_phi_dot, t, ISW_diff_parameters.accel_phi_dot);

  return GSL_SUCCESS;
}


int jac_ISW (double t, const double y[], double *dfdy, double dfdt[], void *phi_dot_inter)
{
  (void)(t); /* avoid unused parameter warning */
  struct ISW_diff_parameters_s ISW_diff_parameters =
    *(struct ISW_diff_parameters_s *)phi_dot_inter;
  
  gsl_matrix_view dfdy_mat
    = gsl_matrix_view_array (dfdy, 1, 1);
  gsl_matrix * m = &dfdy_mat.matrix;

  gsl_matrix_set (m, 0, 0, 1);

  dfdt[0] = 0.0;

  return GSL_SUCCESS;
}


double ISW(double *r, double *phi_dot, double t, double t1)
{
  double f_tmp[1];
  struct ISW_diff_parameters_s phi_dot_inter;

  gsl_interp_accel *accel_phi_dot = gsl_interp_accel_alloc();
  gsl_spline *spline_phi_dot =
    gsl_spline_alloc(gsl_interp_steffen, N_T_SPLINE);
  /*gsl_spline_alloc(gsl_interp_cspline, N_T_SPLINE);*/

  gsl_spline_init( spline_phi_dot,
                   r,
                   phi_dot,
                   N_T_SPLINE );


  phi_dot_inter.accel_phi_dot = accel_phi_dot;
  phi_dot_inter.spline_phi_dot = spline_phi_dot;

  gsl_odeiv2_system sys;

  sys.function = func_ISW;
  sys.jacobian = jac_ISW;
  sys.dimension = 1;
  sys.params = &phi_dot_inter;

  FILE *diff_sol_file;
  diff_sol_file = fopen("ISW_solution.dat", "w+");
  
  const gsl_odeiv2_step_type *T =
    gsl_odeiv2_step_rkf45;

  gsl_odeiv2_step * s
    = gsl_odeiv2_step_alloc (T, 1);
  gsl_odeiv2_control * c
    = gsl_odeiv2_control_y_new (1e-6,1.0);
  gsl_odeiv2_evolve * e
    = gsl_odeiv2_evolve_alloc (1);

  double h = 1e-6;
  // initial conditions
  double y[1] = { 0.0};
  int how_many_iterations=0;

  fprintf(diff_sol_file,"start: %f   end: %f\n", t, t1);

  while (t < t1)
    {
      h = fmin(0.1 , h);
      int status = gsl_odeiv2_evolve_apply (e, c, s,
                                           &sys,
                                           &t, t1,
                                           &h, y);
      if (status != GSL_SUCCESS)
          break;

      printf("t = %g t1 %g; h = %g\n",
             t,t1, h);
      if(status) printf("solver: %s\n",
                             gsl_strerror(status));
      func_mass_deficit(t,
           y,
           (double*)&f_tmp,
            &phi_dot_inter);

      printf("%.5e %.5e\n", t, y[0]);
      fprintf(diff_sol_file,"%.5e %.5e\n", t, y[0]);

    }
  fclose(diff_sol_file);

  gsl_odeiv2_evolve_free (e);
  gsl_odeiv2_control_free (c);
  gsl_odeiv2_step_free (s);

  gsl_spline_free (spline_phi_dot);
  gsl_interp_accel_free (accel_phi_dot);
  printf("y[0] = %f:\n",y[0]);

  return y[0]; 

}
