/*
gsl_lensing_int - simple wrapper for calling GSL integration routines
Copyright (C) 2022 Marius Peper, Boud Roukema

The purpose of this file is to solve the differential equations of the lensing project.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

This file is partly derived from GNU Scientific Library (GSL) examples;
see
 https://sourceware.cygnus.com/gsl
 ftp://alpha.gnu.org/gnu/
 https://sources.redhat.com/gsl
Main authors (C) 1998-2008 include:
 'GNU GSL is being written by Mark Galassi (rosalia@cygnus.com)
  and James Theiler (jt@nis.lanl.gov) with contributions from Jim Davies
  (jimmyd@nis.lanl.gov), Brian Gough (bjg@network-theory.co.uk), Reid
  Priedhorsky (rp@lanl.gov), Gerard Jungman (jungman@lanl.gov) and
  Mike Booth <booth@planck.pha.jhu.edu>.'

*/

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <gsl/gsl_complex.h>
#include <complex.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_integration.h>

struct spline_params
{
  gsl_interp_accel *int_acc;
  gsl_spline *int_spline;
};

double integrand(double r, void *spline)
{
  struct spline_params params = *(struct spline_params *)spline;

  //gsl_interp_accel *func_acc = (spline.int_acc);
  //gsl_spline *func_spline = (spline.int_spline);

  //double f = gsl_spline_eval(func_spline, r, func_acc);
  double f = gsl_spline_eval(params.int_spline, r, params.int_acc);
  return f;
}


double call_gsl_integration(double *r, double *in_array, double a, double b, double abs_error, double rel_error, size_t N_T_SPLINE)
{

  /*
  2022-08-05: The error handler was disabled, initially with the aim of not
  truncating the code when calculating averages. The code leaves the
  simulated box thus the interpolations return NaNs. Some boundary
  conditions were added.

  As advised in https://www.gnu.org/software/gsl/doc/html/err.html, the gsl
  error handler should be set to off status in production runs, and instead
  either GSL_ERROR(reason, gsl_errno) or our own error handler should be used.

  2023-02-03 A generic warning using gsl_error(...) is now implemented.
  */
  int i_gsl_status = 0;
  gsl_set_error_handler_off();

  struct spline_params params;

  gsl_interp_accel *integrant_acc = gsl_interp_accel_alloc();
  /* gsl_spline *integrant_spline = gsl_spline_alloc (gsl_interp_cspline, N_T_SPLINE); */
  /*  gsl_spline *integrant_spline = gsl_spline_alloc (gsl_interp_akima, N_T_SPLINE); */
  gsl_spline *integrant_spline = gsl_spline_alloc (gsl_interp_steffen, N_T_SPLINE);

  gsl_spline_init( integrant_spline,
                   r,
                   in_array,
                   N_T_SPLINE );


  /*
  FILE *input_file;
  input_file = fopen("integration_input.dat", "w+");
  int i;
  for(i=0;i<N_T_SPLINE;i++)
    {
      fprintf(input_file, "%f %f %f\n",r[i], in_array[i],gsl_spline_eval(integrant_spline, r[i], integrant_acc));
    }
  fclose(input_file);
  */
  params.int_acc=integrant_acc;
  params.int_spline=integrant_spline;

  #define WORKSIZE 1000
  gsl_function F;
  gsl_integration_workspace *workspace;
  double result, abserr;

  workspace = gsl_integration_workspace_alloc(WORKSIZE);
  F.function = &integrand;
  F.params = &params;

  i_gsl_status = gsl_integration_qag(&F, a, b, abs_error, rel_error, WORKSIZE, GSL_INTEG_GAUSS21, workspace, &result, &abserr);

  /* printf("i_gsl_status = %d\n",i_gsl_status); */ /* debugging only */
  /* Do not print warnings if the status is OK or if the error is a
     rounding error (since these are expected). */
  if(i_gsl_status != 0 && i_gsl_status != GSL_EROUND){
    printf("Warning: gsl_integration_qag in file %s, line %d: %s.\n", __FILE__, __LINE__, gsl_strerror(i_gsl_status));
  };

  gsl_spline_free(integrant_spline);
  gsl_interp_accel_free(integrant_acc);
  gsl_integration_workspace_free(workspace);

  return result;
}
