/*
test_gsl_lensing_int - test simple wrapper for calling GSL integration routines
Copyright (C) 2022 Marius Peper, Boud Roukema

The purpose of this file is to solve the differential equations of the lensing project.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <math.h>

/* prototypes of functions in lensing_gsl_int.c */
double call_gsl_integration(double *r, double *in_array, double a, double b, double abs_error, double rel_error, int N_T_SPLINE);

int main (void)
{
  int i;

  /* The sampling size N_sample will modify the results of the integration,
     so check_value needs to be changed if N_sample is changed. The value
     N_sample = 100 is appropriate in the version as of 2023-02-02.
  */
#define N_sample 100

  double r[N_sample];
  double in_array[N_sample];
  double test_1, test_2, test_3;

  int pass=0;
  double check_value_akima = 0.139437517306;
  double check_value = 0.139423042617; /* steffen https://ui.adsabs.harvard.edu/abs/1990A%2526A...239..443S */
  /* In practice, as of the version of 2023-02-02, check_tol = 2e-7 should
     be enough to pass the test. GSL is *supposed* to be able to do 
     bitwise reproducible arithmetic. Nevertheless, we set a generous limit
     of 1e-6, five times higher than is needed. */
  /* double check_tol=2e-7; */
  double check_tol=1e-6;

  pass = 0; /* by default everything is OK */

  for(i=0;i<N_sample;i++)
  {
    r[i] = i/(double)N_sample;
    /* smooth but rapidly varying function */
    in_array[i] = (r[i]-0.3)*(r[i]-0.3) + sin(20.0*r[i]);
  }
  /* ignore the absolute error, set the relative error just tight enough to pass the check */
  test_1=call_gsl_integration(r, in_array, r[0], r[N_sample-1], 1.e-30, 1e-5, N_sample);
  printf("test_1 = %20.12f\n", test_1);
  if(fabs(test_1-check_value) > check_tol){
    pass += 1;
  };

  /* ignore the relative error, set the absolute error just tight enough to pass the check */
  test_2=call_gsl_integration(r, in_array, r[0], r[N_sample-1], 1e-6, 1.e-30, N_sample);
  printf("test_2 = %20.12f\n", test_2);
  if(fabs(test_2-check_value) > check_tol){
    pass += 2;
  };

  test_3=call_gsl_integration(r, in_array, r[0], r[N_sample-1], 1e-30, 1e-30, N_sample);
  printf("test_3 = %20.12f\n", test_3);
  if(fabs(test_3-check_value) > check_tol){
    pass += 4;
  };


  printf("pass=%d\n",pass);

  return pass;
}
